#pragma once

#include "../common/common.h"

#include "frame.hpp"

namespace xreg {

class Segment;
class Frame;
class Cell;

enum FeatType { PLANAR, EDGE };

struct Gaussian3d {
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW

  Gaussian3d() = default;

  explicit Gaussian3d(eigen_aligned_std_vector<Vector3d> &pts);

  Gaussian3d(int num, const Vector3d &mu, const Matrix3d &cov);

  Gaussian3d merge(const Gaussian3d &g);

  void addPoint(const Vector3d &pt);

  void decompose();

  // total number
  int num = 0;

  // gaussian parameters
  Vector3d mean = Vector3d::Zero();
  Matrix3d cov = Matrix3d::Zero();

  // decomposition
  SO3 rot;
  Matrix3d rot_mat;
  Vector3d scalar;
};

// struct HomoInfo {
//   EIGEN_MAKE_ALIGNED_OPERATOR_NEW
// }

class Segment {
public:
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW

  using Ptr = std::shared_ptr<Segment>;

  using ConstPtr = std::shared_ptr<const Segment>;

  Segment() = default;

  explicit Segment(const eigen_aligned_std_vector<Vector3d> &pts,
                   bool cache_pts = false);

  ~Segment() = default;

  Matrix4d getQ();

  void clearCache() {
    Tpts_.clear();
    pts_.clear();
  }

  void cachePoint(const Vector3d &Tpt, const Vector3d &pt) {
    Tpts_.push_back(Tpt);
    pts_.push_back(pt);
  }

  void estimateParam();

  void estimateParamHomo();

  int num_;

  Frame *frame_ = nullptr;
  Cell *cell_ = nullptr;

  eigen_aligned_std_vector<Vector3d> pts_, Tpts_;

  // ADD(efllt)
  Matrix4d S_, Q_, L_;

  Gaussian3d gaussian_;
};

class Cell {
public:
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW

public:
  using Ptr = std::shared_ptr<Cell>;
  using ConstPtr = std::shared_ptr<const Cell>;

  size_t idx_;

  // Cell() = default;
  Cell();

  ~Cell() = default;

  int countTotal();

  void solveParam();

  double solveParamHomo();

  void marginalize(int num_frames = 1);

  static Gaussian3d merge(std::vector<Segment *> &segments);

  static Gaussian3d merge(const Gaussian3d &gaussian,
                          const std::vector<Segment *> &segments);

  static Gaussian3d merge2(const Gaussian3d &gaussian,
                           const std::vector<Segment *> &segments);

  void updateInfoAll(bool ignore_back = true);

  double updateInfoAllEigenFactor(bool ignore_back = true);

  void addObservation(Frame *frame, Segment *segment) {
    frames_.push_back(frame);
    segments_.push_back(segment);
  }

  // TODO: better efficiency
  Segment *getSegment(Frame *frame);

  enum ValidResult { SUCCESS, PARTIAL, NOT_ENOUGH };

  /// @note
  bool checkFeature();

  // eigen_aligned_std_vector<Vector3d> pts_cache_;

  // void x();
  bool active_ = true;
  bool optable_ = false;

  // TODO: which frame to marginalize, if not oldest, using list instead?
  std::deque<Frame *> frames_;
  std::deque<Segment *> segments_;
  std::vector<bool> is_outlier_;

  /// @note gaussian info
  Gaussian3d info_map_, info_sw_, info_all_;

public:
  /// @note: efllt
  double solveEF();

  Matrix4d qmat_map_, qmat_sw_, qmat_all_;
  Vector4d pi_map_, pi_;
  double lambda_;
};

} // namespace xreg
