#pragma once

#include "../common/common.h"

#include "cell.h"

namespace xreg {

class Cell;
class Segment;

struct bounds_t {
  double min_x;
  double min_y;
  double min_z;
  double max_x;
  double max_y;
  double max_z;
  int type;

  bounds_t() { min_x = min_y = min_z = max_x = max_y = max_z = 0.0; }
  void inf_x() {
    min_x = -DBL_MAX;
    max_x = DBL_MAX;
  }
  void inf_y() {
    min_y = -DBL_MAX;
    max_y = DBL_MAX;
  }
  void inf_z() {
    min_z = -DBL_MAX;
    max_z = DBL_MAX;
  }
  void inf_xyz() {
    inf_x();
    inf_y();
    inf_z();
  }

  double volume() {
    return (max_x - min_x) * (max_y - min_y) * (max_z - min_z);
  }

  static bounds_t get_intersection_bbx(bounds_t &bbx_1, bounds_t &bbx_2,
                                       //  bounds_t &bbx_intersection,
                                       float bbx_boundary_pad = 0.0);
  double get_intersection_ratio(bounds_t &bbox);
};

class Frame {
 public:
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW

  using Ptr = std::shared_ptr<Frame>;

  using ConstPtr = std::shared_ptr<const Frame>;

  explicit Frame(PointCloudXYZ::Ptr cloud, SE3 Twb = SE3()) {
    cloud_ = cloud;
    setTwb(Twb);
  }

  Frame() = default;

  ~Frame() = default;

  void setTwb(const SE3 &Twb) {
    Twb_ = Twb;
    Tbw_ = Twb.inverse();
  }

  void setTbw(const SE3 &Tbw) {
    Tbw_ = Tbw;
    Twb_ = Tbw.inverse();
  }

  const SE3 &getTwb() const { return Twb_; }

  const SE3 &getTbw() const { return Tbw_; }

  void setBBox(const PointXYZ &pt_min, const PointXYZ &pt_max);

  bounds_t getBBox() const;

  bounds_t getTransformedBBox(const SE3 &Twb) const;

  /// @note: eigen-factor
  Vector6d getJacobian();

  int getNumPoints();

  double countOverlap(Frame *frame);

 public:
  size_t idx_;

  SE3 Twb_, Tbw_, Tgt_;

  PointCloudXYZ::Ptr cloud_ = nullptr, cloud_down_ = nullptr;
  pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr cloud_wn_ = nullptr;
  // CloudFPFH::Ptr cloud_feat_ = nullptr;

  std::vector<Segment *> segments_;
  bounds_t bbox_;

  // ADD(for pairwise registration)
  PointCloudXYZ::Ptr kpts_ = nullptr;
  CloudFPFH::Ptr features_ = nullptr;
  // std::unorder_map<Frame *, std::vector<std::pair<int, int>>> matches_;
};

}  // namespace xreg
