#pragma once

#include "../common/common.h"

#include "cell.h"

// TODO:
// 1. store leaf nodes in a hash table for fast look-up
// 2. xx

namespace xreg {

/** \brief An Oct Tree data structure for storing 3D points
 * \details This is an implementation of a \ref SpatialIndex that splits space
 * using a tree structure. Each node has 8 children that do not necessarily have
 * associated points. Points are stored using \ref NDTCell cells. When inserting
 * points the tree is split until the size BIG_CELL is reached. Further splits
 * occur when the number of points in a leaf goes above the MAX_POINTS threshold
 */
// template <typename PointT>
class OctreeNode {
public:
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW

public:
  using Ptr = std::shared_ptr<OctreeNode>;

  using ConstPtr = std::shared_ptr<const OctreeNode>;

  /// dummy default constructor
  OctreeNode();

  /// creates an oct tree node with known center and size
  //   OctreeNode(PointT center, double xsize, double ysize, double zsize,
  //              NDTCell<PointT> *type, OctreeNode<PointT> *_parent = NULL,
  //              unsigned int _depth = 0);
  OctreeNode(Vector3d center, double xsize, OctreeNode *_parent = nullptr,
             unsigned int _depth = 0);

  virtual ~OctreeNode();

public:
  OctreeNode *parent_ = nullptr;

  OctreeNode *children_[8];

  Cell *cell_ = nullptr;

  double size_;

  // not empty for leaf only
  // eigen_aligned_std_vector<Vector3d> points_;

  unsigned int depth_;
  bool is_leaf_;

  Vector3d center_, color_;

  /// checks in which child node a point would belong
  virtual size_t getIndexForPoint(const Vector3d &pt) const;

  /// fills in the leafs vector when needed
  virtual void computeLeafCells();

public:
  const double &getSize() const { return size_; }

  Vector3d getOrigin() const {
    double half_size = size_ / 2.0;
    // std::cout << size_ << std::endl;
    return center_ - Vector3d(half_size, half_size, half_size);
  }

  // OctreeNode *queryPoint(const Vector3d &pt);

  OctreeNode *expandNode(int idx = -1);

  // bool divideCell();

  // void computeMetrics();

public:
  //--- OctreeNode Parameters ---//
  /// @param maximum depth of the tree, after which no more splits
  static unsigned int MAX_DEPTH;
  /// @param number of points after which to split cell
  static unsigned int MAX_POINTS;
  /// @param at this level do not split any more
  static double SMALL_CELL_SIZE;
  /// @param split tree up to this size before allocating a cell
  static double BIG_CELL_SIZE;

  // bool parametersSet_;

  /// use this to set the parameters for the OctreeNode - will *only* apply to
  /// leafs of the current node.
  static void setParameters(double _BIG_CELL_SIZE = 2,
                            double _SMALL_CELL_SIZE = 0.5, int _MAX_POINTS = 20,
                            int _MAX_DEPTH = 3);

  // Cell Segment
  /// add a point to the index
  std::tuple<OctreeNode *, Cell *, Segment *>
  addPoint(Frame *frame, const Vector3d &Tpt, const Vector3d &pt);

  // virtual OctreeNode *addPoint(const Vector3d &point);

  /// returns a pointer to the cell containing the point or NULL if not found
  virtual OctreeNode *getNodeForPoint(const Vector3d &point, int depth = -1);

  //   inline virtual Cell<PointT> *getMyCell() { return myCell_; }
  virtual OctreeNode *getLeafForPoint(const Vector3d &point,
                                      bool create_new = false);

  /// sets the prototype for a cell
  //   virtual void setCellType(Cell<PointT> *type);

  /// iterator through all cells in index, points at the begining
  //   virtual typename SpatialIndex<PointT>::CellVectorItr begin();
  /// iterator through all cells in index, points at the end
  //   virtual typename SpatialIndex<PointT>::CellVectorItr end();

  /// recursively print the tree
  //   void print();

  /// returns a child at the specified index
  inline OctreeNode *getChild(int idx) {
    if (idx < 8 && idx >= 0) {
      return children_[idx];
    }
    return nullptr;
  }

  void getCells(std::vector<Cell *> &cells);

  inline bool isLeaf() { return is_leaf_; }

  bool isInside(const Vector3d &pt);

  /// cloning method inherited from spatial index
  //   virtual SpatialIndex<PointT> *clone() const;
  //   virtual SpatialIndex<PointT> *copy() const;

  //   virtual void setCenter(const double &cx, const double &cy, const double
  //   &cz); virtual void setSize(const double &sx, const double &sy, const
  //   double &sz);

  //   virtual void getNeighbors(const PointT &point, const double &radius,
  //                             std::vector<Cell<PointT> *> &cells);
  //   virtual bool intersectSphere(const PointT center, const double &radius)
  //   const;

  //   virtual Cell<PointT> *getClosestLeafCell(const PointT &pt);
  //   virtual NDTCell<PointT> *getClosestNDTCell(const PointT &pt);

  //   friend class AdaptiveOctTree<PointT>;
};

} // namespace xreg

// #include "./impl/oc_tree.hpp"
