#pragma once

#include "../common/common.h"

#include "octree.h"

namespace xreg {

using VoxelIndex = Eigen::Vector3i;

struct AnyIndexHash {
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW

  /// number was arbitrarily chosen with no good justification
  static constexpr size_t sl = 17191;
  static constexpr size_t sl2 = sl * sl;

  std::size_t operator()(const VoxelIndex &index) const {
    return static_cast<unsigned int>(index.x() + index.y() * sl +
                                     index.z() * sl2);
  }
};

constexpr double kEpsilon = 1e-6; /**< Used for coordinates. */

/**
 * NOTE: Due the limited accuracy of the FloatingPoint type, this
 * function doesn't always compute the correct grid index for coordinates
 * near the grid cell boundaries. Use the safer `getGridIndexFromOriginPoint` if
 * the origin point is available.
 */
template <typename IndexType>
inline IndexType getGridIndexFromPoint(const Vector3d &point,
                                       const double grid_size_inv) {
  return IndexType(std::floor(point.x() * grid_size_inv + kEpsilon),
                   std::floor(point.y() * grid_size_inv + kEpsilon),
                   std::floor(point.z() * grid_size_inv + kEpsilon));
}

/**
 * NOTE: Due the limited accuracy of the FloatingPoint type, this
 * function doesn't always compute the correct grid index for coordinates
 * near the grid cell boundaries.
 */
template <typename IndexType>
inline IndexType getGridIndexFromPoint(const Vector3d &scaled_point) {
  return IndexType(std::floor(scaled_point.x() + kEpsilon),
                   std::floor(scaled_point.y() + kEpsilon),
                   std::floor(scaled_point.z() + kEpsilon));
}

class Block {
 public:
  using Ptr = std::shared_ptr<Block>;

  using ConstPtr = std::shared_ptr<const Block>;

 public:
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW

  Block() = delete;

  Block(double block_size, Vector3d origin);
  //     : block_size_(block_size), block_size_inv_(1.0 / block_size),
  //       origin_(origin) {

  //   VoxelIndex voxel_idx =
  //       getGridIndexFromPoint<VoxelIndex>(origin * block_size_inv_);
  //   idx_ = AnyIndexHash()(voxel_idx);

  //   initOctree();
  // }

  ~Block() = default;

  void initOctree();

  //   size_t countComponents() { return components_.size(); }

 public:
  size_t idx_ = 0;  // hash index

  double block_size_, block_size_inv_;

  Vector3d origin_;

  OctreeNode::Ptr octree_ = nullptr;

  // list<Cell *> cell_cache_;
};

class VoxelMapOctree {
 public:
  using Ptr = std::shared_ptr<VoxelMapOctree>;

  using ConstPtr = std::shared_ptr<const VoxelMapOctree>;

 public:
  VoxelMapOctree(/* args */) = default;

  explicit VoxelMapOctree(double block_size)
      : block_size_(block_size), block_size_inv_(1.0 / block_size_) {}

  ~VoxelMapOctree() = default;

  //   void rayCasting(const Vector3d &origin, const Vector3d &ray);

  Block::Ptr getVoxel(size_t idx) {
    if (!voxels_.count(idx)) {
      return nullptr;
    }

    return voxels_[idx];
  }

  // void insertFrame();

  void initialize();

  // void insertFrame(Frame *frame);

  struct InsertionOptions {
    bool b_check_min_num = false;
    double th_min_num = 5;

    bool b_check_min_cov = false;
    double th_min_cov = 0.4;
  };

  void insertFrame(PointCloudXYZ::Ptr feats, Frame *frame,
                   std::vector<Segment *> &segments, bool b_add_obs = true);

  void insertFrameEF(PointCloudXYZ::Ptr feats, Frame *frame,
                     std::vector<Segment *> &segments, bool b_add_obs = true);

  void querySegments(PointCloudXYZ::Ptr feats, Frame *frame,
                     std::vector<Segment *> &segments);

  OctreeNode *queryPoints(const Vector3d &pt3d);

  void clearInactiveVoxels(const Vector3d &pt_max, const Vector3d &pt_min);

 public:
  double block_size_, block_size_inv_;

  std::unordered_map<size_t, Block::Ptr> voxels_;

  std::list<Block::Ptr> active_voxels_;
  // std::vector<>
};

}  // namespace xreg
