#include "xreg/estimation/registration_ba.h"

// #include "pcl/registration/icp.h"

#include "xreg/types/cell.h"

#include "xreg/optimization/factors_pr.h"
#include "xreg/optimization/se3_param.hpp"

#include "xreg/common/config.h"

#include <pcl/filters/voxel_grid.h>

#include "xreg/utils/timing.h"

namespace xreg {
using namespace std;

RegistrationBA::RegistrationBA(/* args */) : RegistrationBase() {}

void RegistrationBA::initialize() {
  INFOLOG("initialize");

  voxel_map_ =
      std::unique_ptr<VoxelMapOctree>(new VoxelMapOctree(common::resolution));

  Tlb_.push_back(SE3());

  for (size_t i = 1; i < frames_.size(); i++) {
    auto &&f = frames_[i];
    auto &&fl = frames_[i - 1];

    // frame
    SE3 Tlb = fl->getTbw() * f->getTwb();
    Tlb_.push_back(Tlb);
  }

  if (reg::b_downsample) {
    INFOLOG("downsample");
    pcl::VoxelGrid<pcl::PointXYZ> voxel_filer;
    voxel_filer.setLeafSize(reg::resolution_downsample,
                            reg::resolution_downsample,
                            reg::resolution_downsample);
    for (auto &&frame : frames_) {
      frame->cloud_down_ = PointCloudXYZ::Ptr(new PointCloudXYZ);

      voxel_filer.setInputCloud(frame->cloud_);
      voxel_filer.filter(*frame->cloud_down_);
    }
  } else {
    for (auto &&frame : frames_) {
      frame->cloud_down_ = PointCloudXYZ::Ptr(new PointCloudXYZ);
      *frame->cloud_down_ += *frame->cloud_;
    }
  }
}

void RegistrationBA::registration() {
  // insert first frame
  {
    timing::Timer timer_insertion("insert_frame");
    {
      auto f0 = frames_[0];
      voxel_map_->insertFrame(f0->cloud_down_, f0.get(), f0->segments_);
      // for (auto &&segment : f0->segments_) {
      //   segment->cell_->marginalize();
      // }
    }
    timer_insertion.Stop();
  }

  for (size_t i = 1; i < frames_.size(); i++) {
    auto curr_frame = frames_[i];
    auto last_frame = frames_[i - 1];
    curr_frame->setTwb(last_frame->getTwb() * Tlb_[i]);

    if (reg::b_scan2map) {
      scanToMap(curr_frame);
    }

    // getchar();

    timing::Timer timer_insertion("insert_frame");
    voxel_map_->insertFrame(curr_frame->cloud_down_, curr_frame.get(),
                            curr_frame->segments_);
    timer_insertion.Stop();
    INFOLOG("done insertion");

    if (i > ba::idx_ba_init) {
      mapBA(i);
      INFOLOG("done map ba.");
    }
    timing::Timing::Print(std::cout);
  }
}

void RegistrationBA::mapBA(size_t curr_idx) {
  unordered_set<Cell *> active_cells;

  // for (size_t i = 0; i <= curr_idx; i++) {
  //   auto &&frame = frames_[i];
  //   for (auto &&seg : frame->segments_) {
  //     active_cells.insert(seg->cell_);
  //   }
  // }
  // for (size_t i = 0; i <= curr_idx; i++)
  unordered_set<Frame *> set_frames;
  {
    auto &&frame = frames_[curr_idx];
    for (auto &&seg : frame->segments_) {
      if (seg->cell_->frames_.size() < 2) continue;
      active_cells.insert(seg->cell_);
      for (auto &&f : seg->cell_->frames_) {
        set_frames.insert(f);
      }
    }

    for (size_t i = 0; i < curr_idx; i++) {
      auto &&f = frames_[i].get();
      // for (auto &&f : set_frames)
      if (set_frames.count(f)) {
        double ratio = frame->countOverlap(f);
        if (ratio < 0.25) {
          set_frames.erase(f);
        }
      }
    }

    for (auto &&f : set_frames) {
      for (auto &&seg : f->segments_) {
        active_cells.insert(seg->cell_);
      }
    }
  }

  // define optimization problem
  ceres::Problem problem;
  eigen_aligned_std_vector<Vector3d> rot_params(curr_idx + 1);
  eigen_aligned_std_vector<Vector3d> trans_params(curr_idx + 1);

  unordered_map<Frame *, size_t> frame_pose;
  auto eval_cb = new SurfRegXXEvalCallBack;

  eigen_aligned_std_vector<SE3> poses;

  for (size_t i = 0; i <= curr_idx; i++) {
    // CHECK_NOTNULL(frames_[i]);
    auto &&frame = frames_[i];
    size_t param_idx = i;
    ceres::LocalParameterization *parametrization = new ParamSO3();

    rot_params[param_idx] = frame->getTwb().so3().log();
    trans_params[param_idx] = frame->getTwb().translation();
    poses.push_back(frame->getTwb());

    problem.AddParameterBlock(rot_params[param_idx].data(), 3, parametrization);
    problem.AddParameterBlock(trans_params[param_idx].data(), 3);

    frame_pose[frame.get()] = param_idx;

    eval_cb->addFrame(frame.get(), rot_params[param_idx].data(),
                      trans_params[param_idx].data());

    if (!set_frames.count(frame.get())) {
      problem.SetParameterBlockConstant(rot_params[param_idx].data());
      problem.SetParameterBlockConstant(trans_params[param_idx].data());
    }
  }

  WARNLOG("#constant block: {}", set_frames.size());
  if (reg::b_fix_first) {
    problem.SetParameterBlockConstant(rot_params[0].data());
    problem.SetParameterBlockConstant(trans_params[0].data());
  }

  //   cout << "a" << endl;

  // STEP. plannar
  for (auto &&cell : active_cells) {
    // TODO: check optable
    CHECK_NOTNULL(cell);
    CHECK_EQ(cell->frames_.size(), cell->segments_.size());
    CHECK(!cell->frames_.empty());

    vector<Frame *> frames(cell->frames_.begin(), cell->frames_.end());
    vector<Segment *> segments(cell->segments_.begin(), cell->segments_.end());

    vector<bool> &outliers = cell->is_outlier_;
    vector<bool> discard(frames.size(), false);
    cell->is_outlier_ = discard;
    // outliers.resize(frames.size(), false);
    // cout << "c" << endl;

    // CHECK optable
    // vector<bool> discard(frames.size(), true);

    // TODO: insert segment check
    // int num_non_trivial = 0;
    // for (auto &&segment : segments)
    // for (size_t i = 0; i < segments.size(); i++) {
    //   auto &&segment = segments[i];

    //   bool c1 = segment->gaussian_.num >= 8;
    //   bool c2 = segment->gaussian_.scalar.x() < 0.1;
    //   // bool c2 = true;

    //   if (c1 && c2) {
    //     num_non_trivial++;
    //     discard[i] = false;
    //   }
    // }
    // if (num_non_trivial < 2) {
    //   continue;
    // }
    if (frames.size() < 2) {
      continue;
    }
    // cout << "d" << endl;

    // TODO: necessary?
    // cell->is_outlier_ = discard;
    cell->updateInfoAll(false);
    // cout << "e" << endl;

    // TODO:
    // if (!cell->optable_) {
    //   continue;
    // }
    Vector3d nhat = cell->info_all_.rot_mat.col(0);
    // cout << nhat << endl;
    // SE3 pose_tgt(cell->rot_, cell->mean_);

    vector<EigenFactorRotAxis *> rfs;
    vector<EigenFactorRotAxis *> rfs1;
    vector<EigenFactorTrans2 *> tfs;

    // for (auto &&frame : frames_)
    for (size_t i = 0; i < frames.size(); i++) {
      auto frame = frames[i];
      if (!frame || discard[i]) {
        continue;
      }

      auto param_idx = frame_pose[frame];
      Vector3d &param_rot = rot_params[param_idx];
      Vector3d &param_t = trans_params[param_idx];

      auto segment = segments[i];
      const Gaussian3d &info_map = segment->cell_->info_all_;
      const Gaussian3d &info_local = segment->gaussian_;

      {
        double scalar = 1.0;
        if (ba::ba_weight_type == BA_WEIGHT::LOCAL) {
          scalar = sqrt(info_local.num * 1.0);
        } else if (ba::ba_weight_type == BA_WEIGHT::LOCAL_GLOBAL) {
          scalar = sqrt(info_local.num * 1.0 / info_map.num);
        }

        ceres::LossFunction *loss = nullptr;
        if (ba::ba_loss_type == LCONSTANT) {
          loss = new ceres::HuberLoss(ba::ba_loss_ratio_trans);
        } else if (ba::ba_loss_type == SCALAR_DEPEND) {
          // TODO: chi_squared
          // chi_squared dist(info_local.num - 1);
          // double chi2 = quantile(dist, alpha);
          // loss =
          //     new ceres::HuberLoss(sqrt(chi2) *
          //     test_ba::ba_loss_ratio_trans);
          loss = new ceres::HuberLoss(ba::ba_loss_ratio_trans * scalar);
        }

        ceres::CostFunction *ft =
            new EigenFactorTrans2(&info_map, info_local.mean, scalar);
        problem.AddResidualBlock(ft, loss, param_rot.data(), param_t.data());
      }

      {
        double scalar = 1.0;
        double rs0 = scalar, rs1 = scalar, rs2 = scalar;
        if (ba::ba_weight_type == BA_WEIGHT::LOCAL) {
          rs0 *= sqrt(info_local.scalar.y() * info_local.num);
          rs1 *= sqrt(info_local.scalar.z() * info_local.num);
          // rs2 = sqrt(info_local.scalar.x() * info_local.num);
        } else if (ba::ba_weight_type == BA_WEIGHT::LOCAL_GLOBAL) {
          rs0 *= sqrt(info_local.scalar.y() * info_local.num / info_map.num);
          rs1 *= sqrt(info_local.scalar.z() * info_local.num / info_map.num);
          // rs2 = sqrt(info_local.scalar.x() * info_local.num / info_map.num);
        } else {
          rs0 *= sqrt(info_local.scalar.y());
          rs1 *= sqrt(info_local.scalar.z());
          // rs2 = sqrt(info_local.scalar.x());
        }

        ceres::LossFunction *l0 = nullptr, *l1 = nullptr, *l2 = nullptr;
        if (ba::ba_loss_type == LCONSTANT) {
          l0 = new ceres::HuberLoss(ba::ba_loss_ratio_rot);
          l1 = new ceres::HuberLoss(ba::ba_loss_ratio_rot);
          // l2 = new ceres::HuberLoss(test_ba::ba_loss_ratio_rot);
        } else if (ba::ba_loss_type == SCALAR_DEPEND) {
          l0 = new ceres::HuberLoss(rs0 * ba::ba_loss_ratio_rot);
          l1 = new ceres::HuberLoss(rs1 * ba::ba_loss_ratio_rot);
          // l2 = new ceres::HuberLoss(rs2 * test_ba::ba_loss_ratio_rot);
        }

        ceres::CostFunction *fr0 = new EigenFactorRotAxis(
            &segment->cell_->info_all_, info_local.rot_mat.col(1), rs0);
        ceres::CostFunction *fr1 = new EigenFactorRotAxis(
            &segment->cell_->info_all_, info_local.rot_mat.col(2), rs1);
        // ceres::CostFunction *fr2 = new EigenFactorRotAxis(
        //     &segment->cell_->info_all_, info_local.rot_mat.col(0), rs2);
        problem.AddResidualBlock(fr0, l0, param_rot.data());
        // problem.AddResidualBlock(fr1, l1, param_rot.data());
      }
    }
    // cout << "e" << endl;
    // cout << "a" << endl;

    // eval_cb->addPlane(cell, rfs, rfs1, tfs);
    // TODO:
    eval_cb->addPlane(cell);
  }

  ceres::Solver::Options options;
  // options.initial_trust_region_radius = 1e16;
  // options.max_trust_region_radius = 2e16;
  // options.min_trust_region_radius = 0.5e16;
  // options.min_relative_decrease = 0.0;
  options.use_nonmonotonic_steps = true;
  options.max_consecutive_nonmonotonic_steps = 5;
  options.update_state_every_iteration = true;
  options.linear_solver_type = ceres::DENSE_SCHUR;
  options.minimizer_progress_to_stdout = false;
  // options.num_threads = 2;
  options.trust_region_strategy_type = ceres::LEVENBERG_MARQUARDT;
  // options.trust_region_strategy_type = ceres::DOGLEG;
  // options.minimizer_type = ceres::LINE_SEARCH;
  options.max_num_iterations = 1000;
  options.num_threads = common::num_threads;
  // options.function_tolerance = 1e-8;
  // options.check_gradients = true;
  options.evaluation_callback = eval_cb;

  ceres::Solver::Summary summary;
  ceres::Solve(options, &problem, &summary);
  WARNLOG("#unsuccessful steps: {}", summary.num_unsuccessful_steps);
  cout << summary.BriefReport() << endl;

  SE3 Tb0w = frames_[0]->getTbw();
  double avg_err = 0;
  for (size_t i = 0; i <= curr_idx; i++) {
    // frames_[i]->setTwc(
    //     SE3(SO3::exp(poses_param[i].topRows(3)),
    auto &&frame = frames_[i];

    // INFOLOG("frame {}, before opt: {}", i,
    // poses[i].translation().transpose());
    //     poses_param[i].bottomRows(3)));
    SE3 Twb(SO3::exp(rot_params[i]), trans_params[i]);
    frame->setTwb(Tb0w * Twb);
    // INFOLOG("frame {}, after opt: {}", i,
    //         frame->getTwb().translation().transpose());
    SE3 Terr_before = frame->Tgt_.inverse() * poses[i];
    SE3 Terr = frame->Tgt_.inverse() * frame->getTwb();
    // WARNLOG("err before opt: {}, after opt: {}",
    //         Terr_before.translation().norm(), Terr.translation().norm());
    // SE3 Terr = poses[i].inverse() * frame->getTwb();
    avg_err += Terr.translation().norm();
    // poses[i]);
  }
  avg_err /= frames_.size();
  // cout << "?" << endl;
  // INFOLOG("err: {}", avg_err);
}

void RegistrationBA::scanToScan(Frame::Ptr frame_src, Frame::Ptr frame_tgt) {}

void RegistrationBA::scanToMap(Frame::Ptr frame_src) {
  INFOLOG("perform scan2map.");
  bool b_early_break = true;
  double delta_trans = 1e-6;
  double delta_rot = 1e-7;

  size_t max_iter_ = 1;
  for (size_t iter_ = 0; iter_ < max_iter_; iter_++) {
    {
      std::vector<Segment *> segments;
      voxel_map_->querySegments(frame_src->cloud_down_, frame_src.get(),
                                segments);

      frame_src->segments_ = segments;
    }

    INFOLOG("add param block");

    // define optimization problem
    ceres::Problem problem;
    ceres::LocalParameterization *parametrization = new ParamSO3();

    Vector3d rot_param = frame_src->getTwb().so3().log();
    Vector3d trans_param = frame_src->getTwb().translation();

    problem.AddParameterBlock(rot_param.data(), 3, parametrization);
    problem.AddParameterBlock(trans_param.data(), 3);

    auto &segments = frame_src->segments_;
    for (size_t i = 0; i < segments.size(); i++) {
      auto &&segment = segments[i];
      CHECK_NOTNULL(segment);
      auto &&cell = segment->cell_;
      CHECK_NOTNULL(cell);
      cell->is_outlier_.resize(cell->frames_.size(), false);
      if (iter_ == 0) {
        cell->updateInfoAll(false);
      }

      const auto &info_map = cell->info_all_;
      const auto &info_local = segment->gaussian_;

      if (info_map.num < 8 || info_local.num < 8) {
        // counta++;
        continue;
      }

      {
        // double scalar = sqrt(info_local.num * 1.0 / info_map.num);
        // double scalar = sqrt(info_local.num * 1.0);
        double scalar = 1.0;
        ceres::CostFunction *ft =
            new EigenFactorTrans2(&info_map, info_local.mean, scalar);
        problem.AddResidualBlock(ft,
                                 // nullptr,
                                 new ceres::HuberLoss(0.05 * scalar),
                                 rot_param.data(), trans_param.data());

        double rs0 = 1.0;
        double rs1 = 1.0;

        if (isnan(rs0) || isnan(rs1)) {
          INFOLOG("t: {}", info_local.scalar.transpose());
          INFOLOG("num: {}", info_local.num);
        }
        ceres::CostFunction *fr0 =
            new EigenFactorRotAxis(&info_map, info_local.rot_mat.col(1), rs0);
        ceres::CostFunction *fr1 =
            new EigenFactorRotAxis(&info_map, info_local.rot_mat.col(2), rs1);
        problem.AddResidualBlock(fr0,
                                 // nullptr,
                                 new ceres::HuberLoss(0.2 * rs0),
                                 rot_param.data());
        problem.AddResidualBlock(fr1,
                                 // nullptr,
                                 new ceres::HuberLoss(0.2 * rs1),
                                 rot_param.data());
      }
    }

    ceres::Solver::Options options;
    options.linear_solver_type = ceres::DENSE_QR;
    options.max_num_iterations = 100;
    // options.minimizer_progress_to_stdout = true;
    ceres::Solver::Summary summary;
    ceres::Solve(options, &problem, &summary);

    SE3 Twb_last = frame_src->getTwb();
    SE3 Twb_est(SO3::exp(rot_param), trans_param);

    SE3 Tdelta = Twb_est.inverse() * Twb_last;

    SE3 Tgt = frame_src->Tgt_;
    // INFOLOG("pose before opt: {}", Twb_last.translation().transpose());
    // INFOLOG("pose scan2map: {}", Twb_est.translation().transpose());
    frame_src->setTwb(Twb_est);
    // INFOLOG("pose gt: {}", Tgt.translation().transpose());
    SE3 Terr = Tgt.inverse() * Twb_est;
    // WARNLOG("translation error: {}", Terr.translation().norm());

    frame_src->setTwb(Twb_est);
    for (size_t i = 0; i < frame_src->segments_.size(); i++) {
      delete frame_src->segments_[i];
      frame_src->segments_[i] = nullptr;
    }
    frame_src->segments_.clear();

    double dtrans = Tdelta.translation().norm();
    double drot = Tdelta.so3().log().norm();
    if (b_early_break && dtrans < delta_trans && drot < delta_rot) {
      break;
    }
    // INFOLOG("final: ");
  }
}

}  // namespace xreg