#include "xreg/estimation/registration_ba_ef.h"

// #include "pcl/registration/icp.h"

#include "xreg/types/cell.h"

#include "xreg/optimization/factors_ef.h"
#include "xreg/optimization/factors_pr.h"
#include "xreg/optimization/se3_param.hpp"

#include "xreg/common/config.h"

namespace xreg {
using namespace std;

RegistrationEigenFactor::RegistrationEigenFactor(/* args */)
    : RegistrationBase() {}

void RegistrationEigenFactor::initialize() {
  INFOLOG("initialize");

  voxel_map_ =
      std::unique_ptr<VoxelMapOctree>(new VoxelMapOctree(common::resolution));

  Tlb_.push_back(SE3());

  for (size_t i = 1; i < frames_.size(); i++) {
    auto &&f = frames_[i];
    auto &&fl = frames_[i - 1];

    // frame
    SE3 Tlb = fl->getTbw() * f->getTwb();
    Tlb_.push_back(Tlb);
  }
}

void RegistrationEigenFactor::registration() {
  // insert first frame
  {
    auto f0 = frames_[0];
    voxel_map_->insertFrame(f0->cloud_, f0.get(), f0->segments_);
  }

  for (size_t i = 1; i < frames_.size(); i++) {
    auto curr_frame = frames_[i];
    auto last_frame = frames_[i - 1];
    curr_frame->setTwb(last_frame->getTwb() * Tlb_[i]);
    if (reg::b_scan2map) {
      scanToMap(curr_frame);
    }

    voxel_map_->insertFrame(curr_frame->cloud_, curr_frame.get(),
                            curr_frame->segments_);

    if (i > 1) {
      mapBA(i);
      // if (i == 4)
      // break;
    }
  }
}

void RegistrationEigenFactor::mapBA(size_t curr_idx) {

  unordered_set<Cell *> active_cells;

  // for (auto &&frame : frames_)
  for (size_t i = 0; i <= curr_idx; i++) {
    // CHECK_NOTNULL(frame);
    auto &&frame = frames_[i];
    for (auto &&seg : frame->segments_) {
      active_cells.insert(seg->cell_);
    }
  }
  //   cout << "aaa" << endl;

  // define optimization problem
  ceres::Problem problem;
  eigen_aligned_std_vector<Vector6d> params(curr_idx + 1);

  unordered_map<Frame *, size_t> frame_pose;
  auto eval_cb = new EigenFactorSQREvalCallBack;

  eigen_aligned_std_vector<SE3> poses;

  for (size_t i = 0; i <= curr_idx; i++) {
    // CHECK_NOTNULL(frames_[i]);
    auto &&frame = frames_[i];
    size_t param_idx = i;
    ceres::LocalParameterization *parametrization = new ParamSE3();

    params[param_idx] = frame->getTwb().log();
    // trans_params[param_idx] = frame->getTwb().translation();
    poses.push_back(frame->getTwb());

    problem.AddParameterBlock(params[param_idx].data(), 6, parametrization);
    // problem.AddParameterBlock(trans_params[param_idx].data(), 3);

    frame_pose[frame.get()] = param_idx;

    eval_cb->addFrame(frame.get(), params[param_idx].data());
  }

  if (reg::b_fix_first) {
    problem.SetParameterBlockConstant(params[0].data());
  }

  // STEP. plannar
  for (auto &&cell : active_cells) {
    // TODO: check optable
    CHECK_NOTNULL(cell);
    CHECK_EQ(cell->frames_.size(), cell->segments_.size());
    CHECK(!cell->frames_.empty());

    vector<Frame *> frames(cell->frames_.begin(), cell->frames_.end());
    vector<Segment *> segments(cell->segments_.begin(), cell->segments_.end());
    vector<bool> discard(frames.size(), true);

    // STEP. check optable
    int num_non_trivial = 0;
    for (size_t i = 0; i < segments.size(); i++) {
      auto &&segment = segments[i];

      bool c1 = segment->gaussian_.num >= 8;
      bool c2 = segment->gaussian_.scalar.x() < 0.1;
      // bool c2 = true;

      if (c1 && c2) {
        num_non_trivial++;
        discard[i] = false;
      }
    }
    if (num_non_trivial < 2) {
      continue;
    }

    // TODO: necessary?
    cell->is_outlier_ = discard;
    cell->updateInfoAllEigenFactor(false);
    // INFOLOG("pi: {}", cell->pi_.transpose());

    vector<EigenFactorSQR *> factors;

    for (size_t i = 0; i <= curr_idx; i++) {
      auto frame = frames[i];
      if (!frame || discard[i]) {
        continue;
      }

      auto param_idx = frame_pose[frame];
      Vector6d &param = params[param_idx];
      // Vector3d &param_t = trans_params[param_idx];

      auto segment = segments[i];

      // scal
      double scalar = sqrt(segment->num_ * 1.0);
      EigenFactorSQR *factor = new EigenFactorSQR(segment->L_, cell);
      problem.AddResidualBlock(factor, 
      // nullptr,
                                new ceres::HuberLoss(0.1 * scalar),
                               param.data());
    }

    eval_cb->addPlane(cell);
  }

  ceres::Solver::Options options;
  options.update_state_every_iteration = true;
  options.linear_solver_type = ceres::DENSE_SCHUR;
  options.minimizer_progress_to_stdout = true;
  options.trust_region_strategy_type = ceres::LEVENBERG_MARQUARDT;
  // options.trust_region_strategy_type = ceres::DOGLEG;
  // options.minimizer_type = ceres::LINE_SEARCH;
  options.max_num_iterations = reg::num_iteration;
  options.num_threads = common::num_threads;
  // options.function_tolerance = 1e-8;
  // options.check_gradients = true;
  options.evaluation_callback = eval_cb;

  ceres::Solver::Summary summary;
  ceres::Solve(options, &problem, &summary);
  cout << summary.FullReport() << endl;

  SE3 Tb0w = frames_[0]->getTbw();
  double avg_err = 0;
  for (size_t i = 0; i <= curr_idx; i++) {
    auto &&frame = frames_[i];

    SE3 Twb(SE3::exp(params[i]));
    frame->setTwb(Tb0w * Twb);
    SE3 Terr_before = frame->Tgt_.inverse() * poses[i];
    SE3 Terr = frame->Tgt_.inverse() * frame->getTwb();
    WARNLOG("err before opt: {}, after opt: {}",
            Terr_before.translation().norm(), Terr.translation().norm());
    avg_err += Terr.translation().norm();
  }
  avg_err /= frames_.size();
  // INFOLOG("err: {}", avg_err);
}

void RegistrationEigenFactor::scanToMap(Frame::Ptr frame_src) {
  INFOLOG("perform scan2map.");
  bool b_early_break = true;
  double delta_trans = 1e-6;
  double delta_rot = 1e-7;

  size_t max_iter_ = 1;
  for (size_t iter_ = 0; iter_ < max_iter_; iter_++) {
    {
      std::vector<Segment *> segments;
      voxel_map_->querySegments(frame_src->cloud_, frame_src.get(), segments);

      frame_src->segments_ = segments;
    }

    INFOLOG("add param block");

    // define optimization problem
    ceres::Problem problem;
    ceres::LocalParameterization *parametrization = new ParamSO3();

    Vector3d rot_param = frame_src->getTwb().so3().log();
    Vector3d trans_param = frame_src->getTwb().translation();

    problem.AddParameterBlock(rot_param.data(), 3, parametrization);
    problem.AddParameterBlock(trans_param.data(), 3);

    auto &segments = frame_src->segments_;
    for (size_t i = 0; i < segments.size(); i++) {
      auto &&segment = segments[i];
      CHECK_NOTNULL(segment);
      auto &&cell = segment->cell_;
      CHECK_NOTNULL(cell);
      cell->is_outlier_.resize(cell->frames_.size(), false);
      if (iter_ == 0) {
        cell->updateInfoAll(false);
      }
      // INFOLOG("b");

      const auto &info_map = cell->info_all_;
      const auto &info_local = segment->gaussian_;

      if (info_map.num < 8 || info_local.num < 8) {
        // counta++;
        continue;
      }

      // if (info_map.scalar.x() > mapping::scan2map::eigen_thresh_surf) {
      //   continue;
      // }

      // TODO: check lambda ?
      // if (info_map.scalar.y() / info_map.scalar.x() <
      //     mapping::scan2map::eigen_ratio_surf) {
      //   // cout << info_map.scalar.transpose() << endl;
      //   // countb++;
      //   continue;
      // }

      // if (mapping::scan2map::method == mapping::scan2map::PR)
      {
        // double scalar = sqrt(info_local.num * 1.0 / info_map.num);
        // double scalar = sqrt(info_local.num * 1.0);
        double scalar = 1.0;
        ceres::CostFunction *ft =
            new EigenFactorTrans2(&info_map, info_local.mean, scalar);
        problem.AddResidualBlock(ft,
                                 // nullptr,
                                 new ceres::HuberLoss(0.05 * scalar),
                                 rot_param.data(), trans_param.data());

        // if (info_local.scalar.y() / info_local.scalar.x() >=
        //     mapping::scan2map::eigen_ratio_surf) {
        // if (info_local.scalar.y() / info_local.scalar.x() >= 5.0)
        // {
        // TODO: regulation method
        // double rs0 =
        //     sqrt(info_local.scalar.y() * info_local.num / info_map.num);
        // double rs1 =
        //     sqrt(info_local.scalar.z() * info_local.num / info_map.num);
        // double rs0 = sqrt(info_local.scalar.y() * info_local.num);
        // double rs1 = sqrt(info_local.scalar.z() * info_local.num);
        // double rs0 = sqrt(info_local.num);
        // double rs1 = sqrt(info_local.num);
        double rs0 = 1.0;
        double rs1 = 1.0;

        if (isnan(rs0) || isnan(rs1)) {
          // cout <<
          INFOLOG("t: {}", info_local.scalar.transpose());
          INFOLOG("num: {}", info_local.num);
        }
        // double rs0 = 1.0;
        // double rs1 = 1.0;
        ceres::CostFunction *fr0 =
            new EigenFactorRotAxis(&info_map, info_local.rot_mat.col(1), rs0);
        ceres::CostFunction *fr1 =
            new EigenFactorRotAxis(&info_map, info_local.rot_mat.col(2), rs1);
        // TODO: kernel function
        problem.AddResidualBlock(fr0,
                                 // nullptr,
                                 new ceres::HuberLoss(0.2 * rs0),
                                 rot_param.data());
        problem.AddResidualBlock(fr1,
                                 // nullptr,
                                 new ceres::HuberLoss(0.2 * rs1),
                                 rot_param.data());
        // }
      }
    }
    // INFOLOG("residual block");

    ceres::Solver::Options options;
    options.linear_solver_type = ceres::DENSE_QR;
    options.max_num_iterations = 100;
    options.minimizer_progress_to_stdout = true;
    ceres::Solver::Summary summary;
    ceres::Solve(options, &problem, &summary);

    SE3 Twb_last = frame_src->getTwb();
    SE3 Twb_est(SO3::exp(rot_param), trans_param);

    // Twb_last *
    SE3 Tdelta = Twb_est.inverse() * Twb_last;

    SE3 Tgt = frame_src->Tgt_;
    INFOLOG("pose before opt: {}", Twb_last.translation().transpose());
    INFOLOG("pose scan2map: {}", Twb_est.translation().transpose());
    frame_src->setTwb(Twb_est);
    INFOLOG("pose gt: {}", Tgt.translation().transpose());
    SE3 Terr = Tgt.inverse() * Twb_est;
    WARNLOG("translation error: {}", Terr.translation().norm());

    frame_src->setTwb(Twb_est);
    for (size_t i = 0; i < frame_src->segments_.size(); i++) {
      delete frame_src->segments_[i];
      frame_src->segments_[i] = nullptr;
    }

    double dtrans = Tdelta.translation().norm();
    double drot = Tdelta.so3().log().norm();
    if (b_early_break && dtrans < delta_trans && drot < delta_rot) {
      break;
    }

    INFOLOG("final: ");
  }
}

} // namespace xreg