#include "xreg/estimation/registration_ba.h"

// #include "pcl/registration/icp.h"

#include "xreg/types/cell.h"

#include "xreg/optimization/factors_pr.h"
#include "xreg/optimization/se3_param.hpp"

#include "xreg/common/config.h"

#include <pcl/filters/voxel_grid.h>

namespace xreg {
using namespace std;

RegistrationBA::RegistrationBA(/* args */) : RegistrationBase() {}

void RegistrationBA::initialize() {
  INFOLOG("initialize");

  voxel_map_ =
      std::unique_ptr<VoxelMapOctree>(new VoxelMapOctree(common::resolution));

  Tlb_.push_back(SE3());

  for (size_t i = 1; i < frames_.size(); i++) {
    auto &&f = frames_[i];
    auto &&fl = frames_[i - 1];

    // frame
    SE3 Tlb = fl->getTbw() * f->getTwb();
    Tlb_.push_back(Tlb);
  }

  if (reg::b_downsample) {
    INFOLOG("downsample");
    pcl::VoxelGrid<pcl::PointXYZ> voxel_filer;
    voxel_filer.setLeafSize(reg::resolution_downsample,
                            reg::resolution_downsample,
                            reg::resolution_downsample);
    for (auto &&frame : frames_) {
      frame->cloud_down_ = PointCloudXYZ::Ptr(new PointCloudXYZ);

      voxel_filer.setInputCloud(frame->cloud_);
      voxel_filer.filter(*frame->cloud_down_);
    }
  } else {
    for (auto &&frame : frames_) {
      frame->cloud_down_ = PointCloudXYZ::Ptr(new PointCloudXYZ);
      *frame->cloud_down_ += *frame->cloud_;
    }
  }
}

void RegistrationBA::registration() {
  // insert first frame
  {
    auto f0 = frames_[0];
    voxel_map_->insertFrame(f0->cloud_down_, f0.get(), f0->segments_);
    // for (auto &&segment : f0->segments_) {
    //   segment->cell_->marginalize();
    // }
  }

  for (size_t i = 1; i < frames_.size(); i++) {
    auto curr_frame = frames_[i];
    auto last_frame = frames_[i - 1];
    curr_frame->setTwb(last_frame->getTwb() * Tlb_[i]);

    if (reg::b_scan2map) {
      scanToMap(curr_frame);
    }

    // getchar();

    voxel_map_->insertFrame(curr_frame->cloud_down_, curr_frame.get(),
                            curr_frame->segments_);

    if (i > 1) {
      mapBA(i);
      // if (i == 15)
      //   break;
    }
  }
}

void RegistrationBA::scanToScan(Frame::Ptr frame_src, Frame::Ptr frame_tgt) {}

void RegistrationBA::scanToMap(Frame::Ptr frame_src) {
  INFOLOG("perform scan2map.");
  bool b_early_break = true;
  double delta_trans = 1e-6;
  double delta_rot = 1e-7;

  size_t max_iter_ = 1;
  for (size_t iter_ = 0; iter_ < max_iter_; iter_++) {
    {
      std::vector<Segment *> segments;
      voxel_map_->querySegments(frame_src->cloud_down_, frame_src.get(),
                                segments);

      frame_src->segments_ = segments;
    }

    INFOLOG("add param block");

    // define optimization problem
    ceres::Problem problem;
    ceres::LocalParameterization *parametrization = new ParamSO3();

    Vector3d rot_param = frame_src->getTwb().so3().log();
    Vector3d trans_param = frame_src->getTwb().translation();

    problem.AddParameterBlock(rot_param.data(), 3, parametrization);
    problem.AddParameterBlock(trans_param.data(), 3);

    auto &segments = frame_src->segments_;
    for (size_t i = 0; i < segments.size(); i++) {
      auto &&segment = segments[i];
      CHECK_NOTNULL(segment);
      auto &&cell = segment->cell_;
      CHECK_NOTNULL(cell);
      // INFOLOG("a");
      // auto& outliers = cell->is_outlier_;
      // outliers.resize(cel, false);
      cell->is_outlier_.resize(cell->frames_.size(), false);
      if (iter_ == 0) {
        cell->updateInfoAll(false);
      }
      // INFOLOG("b");

      const auto &info_map = cell->info_all_;
      const auto &info_local = segment->gaussian_;

      if (info_map.num < 8 || info_local.num < 8) {
        // counta++;
        continue;
      }

      // if (info_map.scalar.x() > mapping::scan2map::eigen_thresh_surf) {
      //   continue;
      // }

      // TODO: check lambda ?
      // if (info_map.scalar.y() / info_map.scalar.x() <
      //     mapping::scan2map::eigen_ratio_surf) {
      //   // cout << info_map.scalar.transpose() << endl;
      //   // countb++;
      //   continue;
      // }

      // if (mapping::scan2map::method == mapping::scan2map::PR)
      {
        // double scalar = sqrt(info_local.num * 1.0 / info_map.num);
        // double scalar = sqrt(info_local.num * 1.0);
        double scalar = 1.0;
        ceres::CostFunction *ft =
            new EigenFactorTrans2(&info_map, info_local.mean, scalar);
        problem.AddResidualBlock(ft,
                                 // nullptr,
                                 new ceres::HuberLoss(0.05 * scalar),
                                 rot_param.data(), trans_param.data());

        // if (info_local.scalar.y() / info_local.scalar.x() >=
        //     mapping::scan2map::eigen_ratio_surf) {
        // if (info_local.scalar.y() / info_local.scalar.x() >= 5.0)
        // {
        // TODO: regulation method
        // double rs0 =
        //     sqrt(info_local.scalar.y() * info_local.num / info_map.num);
        // double rs1 =
        //     sqrt(info_local.scalar.z() * info_local.num / info_map.num);
        // double rs0 = sqrt(info_local.scalar.y() * info_local.num);
        // double rs1 = sqrt(info_local.scalar.z() * info_local.num);
        // double rs0 = sqrt(info_local.num);
        // double rs1 = sqrt(info_local.num);
        double rs0 = 1.0;
        double rs1 = 1.0;

        if (isnan(rs0) || isnan(rs1)) {
          // cout <<
          INFOLOG("t: {}", info_local.scalar.transpose());
          INFOLOG("num: {}", info_local.num);
        }
        // double rs0 = 1.0;
        // double rs1 = 1.0;
        ceres::CostFunction *fr0 =
            new EigenFactorRotAxis(&info_map, info_local.rot_mat.col(1), rs0);
        ceres::CostFunction *fr1 =
            new EigenFactorRotAxis(&info_map, info_local.rot_mat.col(2), rs1);
        // TODO: kernel function
        problem.AddResidualBlock(fr0,
                                 // nullptr,
                                 new ceres::HuberLoss(0.2 * rs0),
                                 rot_param.data());
        problem.AddResidualBlock(fr1,
                                 // nullptr,
                                 new ceres::HuberLoss(0.2 * rs1),
                                 rot_param.data());
        // }
      }
    }
    // INFOLOG("residual block");

    ceres::Solver::Options options;
    options.linear_solver_type = ceres::DENSE_QR;
    options.max_num_iterations = 100;
    options.minimizer_progress_to_stdout = true;
    ceres::Solver::Summary summary;
    ceres::Solve(options, &problem, &summary);

    SE3 Twb_last = frame_src->getTwb();
    SE3 Twb_est(SO3::exp(rot_param), trans_param);

    // Twb_last *
    SE3 Tdelta = Twb_est.inverse() * Twb_last;

    SE3 Tgt = frame_src->Tgt_;
    INFOLOG("pose before opt: {}", Twb_last.translation().transpose());
    INFOLOG("pose scan2map: {}", Twb_est.translation().transpose());
    frame_src->setTwb(Twb_est);
    INFOLOG("pose gt: {}", Tgt.translation().transpose());
    SE3 Terr = Tgt.inverse() * Twb_est;
    WARNLOG("translation error: {}", Terr.translation().norm());

    frame_src->setTwb(Twb_est);
    for (size_t i = 0; i < frame_src->segments_.size(); i++) {
      delete frame_src->segments_[i];
      frame_src->segments_[i] = nullptr;
    }

    double dtrans = Tdelta.translation().norm();
    double drot = Tdelta.so3().log().norm();
    if (b_early_break && dtrans < delta_trans && drot < delta_rot) {
      break;
    }

    INFOLOG("final: ");
  }
}

void RegistrationBA::mapBA(size_t curr_idx) {
  unordered_set<Cell *> active_cells;

  // for (auto &&frame : frames_)
  for (size_t i = 0; i <= curr_idx; i++) {
    // CHECK_NOTNULL(frame);
    auto &&frame = frames_[i];
    for (auto &&seg : frame->segments_) {
      active_cells.insert(seg->cell_);
    }
  }
  //   cout << "aaa" << endl;

  // define optimization problem
  ceres::Problem problem;
  eigen_aligned_std_vector<Vector3d> rot_params(curr_idx + 1);
  eigen_aligned_std_vector<Vector3d> trans_params(curr_idx + 1);

  unordered_map<Frame *, size_t> frame_pose;
  auto eval_cb = new SurfRegXXEvalCallBack;

  eigen_aligned_std_vector<SE3> poses;

  for (size_t i = 0; i <= curr_idx; i++) {
    // CHECK_NOTNULL(frames_[i]);
    auto &&frame = frames_[i];
    size_t param_idx = i;
    ceres::LocalParameterization *parametrization = new ParamSO3();

    rot_params[param_idx] = frame->getTwb().so3().log();
    trans_params[param_idx] = frame->getTwb().translation();
    poses.push_back(frame->getTwb());

    problem.AddParameterBlock(rot_params[param_idx].data(), 3, parametrization);
    problem.AddParameterBlock(trans_params[param_idx].data(), 3);

    frame_pose[frame.get()] = param_idx;

    eval_cb->addFrame(frame.get(), rot_params[param_idx].data(),
                      trans_params[param_idx].data());
  }
  // problem.SetParameterBlockConstant(rot_params[0].data());
  // problem.SetParameterBlockConstant(trans_params[0].data());
  //   cout << "a" << endl;

  // STEP. plannar
  for (auto &&cell : active_cells) {
    // TODO: check optable
    CHECK_NOTNULL(cell);
    CHECK_EQ(cell->frames_.size(), cell->segments_.size());
    CHECK(!cell->frames_.empty());

    vector<Frame *> frames(cell->frames_.begin(), cell->frames_.end());
    vector<Segment *> segments(cell->segments_.begin(), cell->segments_.end());
    vector<bool> discard(frames.size(), true);

    // vector<bool> &outliers = cell->is_outlier_;
    // outliers.resize(frames.size(), false);
    // cout << "c" << endl;

    // CHECK optable
    // TODO: insert segment check
    int num_non_trivial = 0;
    // for (auto &&segment : segments)
    for (size_t i = 0; i < segments.size(); i++) {
      auto &&segment = segments[i];

      bool c1 = segment->gaussian_.num >= 8;
      bool c2 = segment->gaussian_.scalar.x() < 0.1;
      // bool c2 = true;

      if (c1 && c2) {
        num_non_trivial++;
        discard[i] = false;
      }
    }
    if (num_non_trivial < 2) {
      continue;
    }
    // cout << "d" << endl;

    // TODO: necessary?
    cell->is_outlier_ = discard;
    cell->updateInfoAll(false);
    // cout << "e" << endl;

    // TODO:
    // if (!cell->optable_) {
    //   continue;
    // }
    Vector3d nhat = cell->info_all_.rot_mat.col(0);
    // cout << nhat << endl;
    // SE3 pose_tgt(cell->rot_, cell->mean_);

    vector<EigenFactorRotAxis *> rfs;
    vector<EigenFactorRotAxis *> rfs1;
    vector<EigenFactorTrans2 *> tfs;

    // for (auto &&frame : frames_)
    for (size_t i = 0; i <= curr_idx; i++) {
      auto frame = frames[i];
      if (!frame || discard[i]) {
        continue;
      }

      auto param_idx = frame_pose[frame];
      Vector3d &param_rot = rot_params[param_idx];
      Vector3d &param_t = trans_params[param_idx];

      auto segment = segments[i];
      const Gaussian3d &info_map = segment->cell_->info_all_;
      const Gaussian3d &info_local = segment->gaussian_;

      double scalar = sqrt(info_local.num * 1.0 / info_map.num);
      // double scalar = sqrt(info_local.num * 1.0);
      // double scalar = 1.0;
      ceres::CostFunction *ft =
          new EigenFactorTrans2(&info_map, info_local.mean, scalar);
      problem.AddResidualBlock(ft, nullptr,
                               //  new ceres::HuberLoss(0.1 * scalar),
                               param_rot.data(), param_t.data());

      // if (info_local.scalar.y() / info_local.scalar.x() >=
      //     mapping::scan2map::eigen_ratio_surf) {
      //   if (info_local.scalar.y() / info_local.scalar.x() >= 5.0)
      {
        double rs0 =
            sqrt(info_local.scalar.y() * info_local.num / info_map.num);
        double rs1 =
            sqrt(info_local.scalar.z() * info_local.num / info_map.num);
        // double rs2 =
        //     sqrt(info_local.scalar.x() * info_local.num / info_map.num);
        // if (isnan(rs0)) {
        //   WARNLOG("fuck");
        //   WARNLOG("{}, {}, {}", info_local.scalar.y(), info_local.num,
        //           info_map.num);
        // }
        // double rs0 = sqrt(info_local.scalar.y() * info_local.num);
        // double rs1 = sqrt(info_local.scalar.z() * info_local.num);
        // double rs0 = sqrt(info_local.scalar.y());
        // double rs1 = sqrt(info_local.scalar.z());
        // double rs2 = sqrt(info_local.scalar.z());
        // double rs0 = 1.0;
        // double rs1 = 1.0;
        // double rs2 = 1.0;
        // if (rs0 <= 1e-8) {
        //   cout << rs0 << ' ' << info_local.scalar.y() << ' ' <<
        //   info_local.num
        //        << endl;
        //   exit(0);
        // }
        // cout << rs0 << ' ' << rs1 << endl;
        ceres::CostFunction *fr0 =
            new EigenFactorRotAxis(&info_map, info_local.rot_mat.col(1), rs0);
        ceres::CostFunction *fr1 =
            new EigenFactorRotAxis(&info_map, info_local.rot_mat.col(2), rs1);
        // ceres::CostFunction *fr3 =
        //     new EigenFactorRotAxis(&info_map, info_local.rot_mat.col(0),
        //     rs2);
        problem.AddResidualBlock(fr0, nullptr,
                                 //  new ceres::HuberLoss(0.1 * rs0),
                                 param_rot.data());
        problem.AddResidualBlock(fr1, nullptr,
                                 //  new ceres::HuberLoss(0.1 * rs1),
                                 param_rot.data());
        // problem.AddResidualBlock(fr3, nullptr,
        //                          //  new ceres::HuberLoss(0.1 * rs1),
        //                          param_rot.data());
      }
    }
    // cout << "e" << endl;

    // eval_cb->addPlane(cell, rfs, rfs1, tfs);
    // TODO:
    eval_cb->addPlane(cell);
  }

  ceres::Solver::Options options;
  // options.update_state_every_iteration = true;
  options.linear_solver_type = ceres::DENSE_SCHUR;
  options.minimizer_progress_to_stdout = false;
  // options.num_threads = 2;
  options.trust_region_strategy_type = ceres::LEVENBERG_MARQUARDT;
  // options.trust_region_strategy_type = ceres::DOGLEG;
  // options.minimizer_type = ceres::LINE_SEARCH;
  options.max_num_iterations = 1000;
  options.num_threads = common::num_threads;
  // options.function_tolerance = 1e-8;
  // options.check_gradients = true;
  options.evaluation_callback = eval_cb;

  ceres::Solver::Summary summary;
  ceres::Solve(options, &problem, &summary);
  cout << summary.BriefReport() << endl;

  SE3 Tb0w = frames_[0]->getTbw();
  double avg_err = 0;
  for (size_t i = 0; i <= curr_idx; i++) {
    // frames_[i]->setTwc(
    //     SE3(SO3::exp(poses_param[i].topRows(3)),
    auto &&frame = frames_[i];

    // INFOLOG("frame {}, before opt: {}", i,
    // poses[i].translation().transpose());
    //     poses_param[i].bottomRows(3)));
    SE3 Twb(SO3::exp(rot_params[i]), trans_params[i]);
    frame->setTwb(Tb0w * Twb);
    // INFOLOG("frame {}, after opt: {}", i,
    //         frame->getTwb().translation().transpose());
    SE3 Terr_before = frame->Tgt_.inverse() * poses[i];
    SE3 Terr = frame->Tgt_.inverse() * frame->getTwb();
    // WARNLOG("err before opt: {}, after opt: {}",
    //         Terr_before.translation().norm(), Terr.translation().norm());
    // SE3 Terr = poses[i].inverse() * frame->getTwb();
    avg_err += Terr.translation().norm();
    // poses[i]);
  }
  avg_err /= frames_.size();
  // cout << "?" << endl;
  // INFOLOG("err: {}", avg_err);
}

// void RegistrationBA::scanToScan(Frame::Ptr frame_tgt, Frame::Ptr frame_src)
// {

//   INFOLOG("perform scan2map.");

//   voxel_map_ =
//       std::unique_ptr<VoxelMapOctree>(new
//       VoxelMapOctree(common::resolution));

//   // for (size_t i = 0; i < frames_.size(); i++) {
//   //   auto frame = frames_[i];

//   //   vector<Segment *> segments;
//   //   voxel_map_->insertFrame(frame->cloud_, frame.get(), segments);

//   //   frame->segments_ = segments;
//   // }
//   // Frame::Ptr frame_tgt, frame_src;
//   bool b_early_break = true;
//   double delta_trans = 1e-6;
//   double delta_rot = 1e-7;

//   {
//     vector<Segment *> segments;
//     // frame_tgt = frames_[0];
//     // frame_tgt->setTwb(SE3());
//     voxel_map_->insertFrame(frame_tgt->cloud_, frame_tgt.get(), segments);
//   }

//   // unordered_set<Cell *> active_cells;
//   // frame_src = frames_[1];
//   frame_src->setTwb(frame_tgt->getTwb());

//   size_t max_iter_ = 100;
//   for (size_t iter_ = 0; iter_ < max_iter_; iter_++) {
//     {
//       std::vector<Segment *> segments;

//       // frame_src->setTwb(SE3());
//       voxel_map_->querySegments(frame_src->cloud_, frame_src.get(),
//       segments);

//       frame_src->segments_ = segments;
//       // active_cells
//       // for (auto &&seg : segments) {
//       //   active_cells.insert(seg->cell_);
//       // }
//     }

//     INFOLOG("add param block");

//     // define optimization problem
//     ceres::Problem problem;

//     ceres::LocalParameterization *parametrization = new ParamSO3();

//     Vector3d rot_param = frame_src->getTwb().so3().log();
//     Vector3d trans_param = frame_src->getTwb().translation();

//     problem.AddParameterBlock(rot_param.data(), 3, parametrization);
//     problem.AddParameterBlock(trans_param.data(), 3);
//     //   cout << "a" << endl;

//     auto &segments = frame_src->segments_;
//     for (size_t i = 0; i < segments.size(); i++) {

//       auto &&segment = segments[i];
//       CHECK_NOTNULL(segment);
//       auto &&cell = segment->cell_;
//       CHECK_NOTNULL(cell);
//       // INFOLOG("a");
//       // auto& outliers = cell->is_outlier_;
//       // outliers.resize(cel, false);
//       cell->is_outlier_.resize(cell->frames_.size(), false);
//       if (iter_ == 0) {
//         cell->updateInfoAll(false);
//       }
//       // INFOLOG("b");

//       const auto &info_map = cell->info_all_;
//       const auto &info_local = segment->gaussian_;

//       if (info_map.num < 8 || info_local.num < 8) {
//         // counta++;
//         continue;
//       }

//       // if (info_map.scalar.x() > mapping::scan2map::eigen_thresh_surf) {
//       //   continue;
//       // }

//       // TODO: check lambda ?
//       // if (info_map.scalar.y() / info_map.scalar.x() <
//       //     mapping::scan2map::eigen_ratio_surf) {
//       //   // cout << info_map.scalar.transpose() << endl;
//       //   // countb++;
//       //   continue;
//       // }

//       // if (mapping::scan2map::method == mapping::scan2map::PR)
//       {
//         // double scalar = sqrt(info_local.num * 1.0 / info_map.num);
//         // double scalar = sqrt(info_local.num * 1.0);
//         double scalar = 1.0;
//         ceres::CostFunction *ft =
//             new EigenFactorTrans2(&info_map, info_local.mean, scalar);
//         problem.AddResidualBlock(ft,
//                                  // nullptr,
//                                  new ceres::HuberLoss(0.05 * scalar),
//                                  rot_param.data(), trans_param.data());

//         // if (info_local.scalar.y() / info_local.scalar.x() >=
//         //     mapping::scan2map::eigen_ratio_surf) {
//         // if (info_local.scalar.y() / info_local.scalar.x() >= 5.0)
//         {
//           // double rs0 =
//           //     sqrt(info_local.scalar.y() * info_local.num /
//           info_map.num);
//           // double rs1 =
//           //     sqrt(info_local.scalar.z() * info_local.num /
//           info_map.num);
//           // double rs0 = sqrt(info_local.scalar.y() * info_local.num);
//           // double rs1 = sqrt(info_local.scalar.z() * info_local.num);
//           // double rs0 = sqrt(info_local.num);
//           // double rs1 = sqrt(info_local.num);
//           double rs0 = 1.0;
//           double rs1 = 1.0;

//           if (isnan(rs0) || isnan(rs1)) {
//             // cout <<
//             INFOLOG("t: {}", info_local.scalar.transpose());
//             INFOLOG("num: {}", info_local.num);
//           }
//           // double rs0 = 1.0;
//           // double rs1 = 1.0;
//           ceres::CostFunction *fr0 =
//               new EigenFactorRotAxis(&info_map, info_local.rot_mat.col(1),
//               rs0);
//           ceres::CostFunction *fr1 =
//               new EigenFactorRotAxis(&info_map, info_local.rot_mat.col(2),
//               rs1);
//           problem.AddResidualBlock(fr0,
//                                    // nullptr,
//                                    new ceres::HuberLoss(0.2 * rs0),
//                                    rot_param.data());
//           problem.AddResidualBlock(fr1,
//                                    // nullptr,
//                                    new ceres::HuberLoss(0.2 * rs1),
//                                    rot_param.data());
//         }
//       }
//     }
//     // INFOLOG("residual block");

//     ceres::Solver::Options options;
//     options.linear_solver_type = ceres::DENSE_QR;
//     options.max_num_iterations = 100;
//     options.minimizer_progress_to_stdout = true;
//     ceres::Solver::Summary summary;
//     ceres::Solve(options, &problem, &summary);

//     SE3 Twb_last = frame_src->getTwb();
//     SE3 Twb_est(SO3::exp(rot_param), trans_param);

//     // Twb_last *
//     SE3 Tdelta = Twb_est.inverse() * Twb_last;

//     SE3 Tgt = frame_tgt->Tgt_.inverse() * frame_src->Tgt_;
//     INFOLOG("pose before opt: {}",
//             frame_src->getTwb().translation().transpose());
//     INFOLOG("pose scan2map: {}", Twb_est.translation().transpose());
//     frame_src->setTwb(Twb_est);
//     INFOLOG("pose gt: {}", Tgt.translation().transpose());
//     // frame_src->setTwb(Twb_est);
//     // for ()
//     for (size_t i = 0; i < frame_src->segments_.size(); i++) {
//       delete frame_src->segments_[i];
//       frame_src->segments_[i] = nullptr;
//     }

//     double dtrans = Tdelta.translation().norm();
//     double drot = Tdelta.so3().log().norm();
//     if (b_early_break && dtrans < delta_trans && drot < delta_rot) {
//       break;
//     }
//   }
//   // INFOLOG("pose scan2map: {}", Twb_est.translation().transpose());
//   // frame_src->setTwb(Twb_est);
//   // INFOLOG("pose gt: {}", Tgt.translation().transpose());
//   // SE3 Tgt = frame_tgt->Tgt_.inverse() * frame_src->Tgt_;
//   INFOLOG("final: ");

//   // for (auto &seg : frame_tgt->segments_) {
//   //   delete seg;
//   //   seg = nullptr;
//   // }
// }

// void RegistrationBA::registration() {

//   INFOLOG("perform scan2map.");

//   voxel_map_ =
//       std::unique_ptr<VoxelMapOctree>(new
//       VoxelMapOctree(common::resolution));

//   // for (size_t i = 0; i < frames_.size(); i++) {
//   //   auto frame = frames_[i];

//   //   vector<Segment *> segments;
//   //   voxel_map_->insertFrame(frame->cloud_, frame.get(), segments);

//   //   frame->segments_ = segments;
//   // }
//   Frame::Ptr frame_tgt, frame_src;

//   {
//     vector<Segment *> segments;
//     frame_tgt = frames_[0];
//     // frame_tgt->setTwb(SE3());
//     voxel_map_->insertFrame(frame_tgt->cloud_, frame_tgt.get(), segments);
//   }

//   // unordered_set<Cell *> active_cells;
//   frame_src = frames_[1];
//   frame_src->setTwb(frame_tgt->getTwb());

//   size_t max_iter_ = 10;
//   for (size_t iter_ = 0; iter_ < max_iter_; iter_++) {
//     {
//       std::vector<Segment *> segments;

//       // frame_src->setTwb(SE3());
//       voxel_map_->querySegments(frame_src->cloud_, frame_src.get(),
//       segments);

//       frame_src->segments_ = segments;
//       // active_cells
//       // for (auto &&seg : segments) {
//       //   active_cells.insert(seg->cell_);
//       // }
//     }

//     INFOLOG("add param block");

//     // define optimization problem
//     ceres::Problem problem;

//     ceres::LocalParameterization *parametrization = new ParamSO3();

//     Vector3d rot_param = frame_src->getTwb().so3().log();
//     Vector3d trans_param = frame_src->getTwb().translation();

//     problem.AddParameterBlock(rot_param.data(), 3, parametrization);
//     problem.AddParameterBlock(trans_param.data(), 3);
//     //   cout << "a" << endl;

//     auto &segments = frame_src->segments_;
//     for (size_t i = 0; i < segments.size(); i++) {

//       auto &&segment = segments[i];
//       CHECK_NOTNULL(segment);
//       auto &&cell = segment->cell_;
//       CHECK_NOTNULL(cell);
//       // INFOLOG("a");
//       // auto& outliers = cell->is_outlier_;
//       // outliers.resize(cel, false);
//       cell->is_outlier_.resize(cell->frames_.size(), false);
//       if (iter_ == 0) {
//         cell->updateInfoAll(false);
//       }
//       // INFOLOG("b");

//       const auto &info_map = cell->info_all_;
//       const auto &info_local = segment->gaussian_;

//       if (info_map.num < 8 || info_local.num < 8) {
//         // counta++;
//         continue;
//       }

//       // if (info_map.scalar.x() > mapping::scan2map::eigen_thresh_surf) {
//       //   continue;
//       // }

//       // TODO: check lambda ?
//       // if (info_map.scalar.y() / info_map.scalar.x() <
//       //     mapping::scan2map::eigen_ratio_surf) {
//       //   // cout << info_map.scalar.transpose() << endl;
//       //   // countb++;
//       //   continue;
//       // }

//       // if (mapping::scan2map::method == mapping::scan2map::PR)
//       {
//         // double scalar = sqrt(info_local.num * 1.0 / info_map.num);
//         // double scalar = sqrt(info_local.num * 1.0);
//         double scalar = 1.0;
//         ceres::CostFunction *ft =
//             new EigenFactorTrans2(&info_map, info_local.mean, scalar);
//         problem.AddResidualBlock(ft, nullptr,
//                                  //  new ceres::HuberLoss(0.1 * scalar),
//                                  rot_param.data(), trans_param.data());

//         // if (info_local.scalar.y() / info_local.scalar.x() >=
//         //     mapping::scan2map::eigen_ratio_surf) {
//         // if (info_local.scalar.y() / info_local.scalar.x() >= 5.0)
//         {
//           // double rs0 =
//           //     sqrt(info_local.scalar.y() * info_local.num /
//           info_map.num);
//           // double rs1 =
//           //     sqrt(info_local.scalar.z() * info_local.num /
//           info_map.num);
//           // double rs0 = sqrt(info_local.scalar.y() * info_local.num);
//           // double rs1 = sqrt(info_local.scalar.z() * info_local.num);
//           // double rs0 = sqrt(info_local.num);
//           // double rs1 = sqrt(info_local.num);
//           double rs0 = 1.0;
//           double rs1 = 1.0;

//           if (isnan(rs0) || isnan(rs1)) {
//             // cout <<
//             INFOLOG("t: {}", info_local.scalar.transpose());
//             INFOLOG("num: {}", info_local.num);
//           }
//           // double rs0 = 1.0;
//           // double rs1 = 1.0;
//           ceres::CostFunction *fr0 =
//               new EigenFactorRotAxis(&info_map, info_local.rot_mat.col(1),
//               rs0);
//           ceres::CostFunction *fr1 =
//               new EigenFactorRotAxis(&info_map, info_local.rot_mat.col(2),
//               rs1);
//           problem.AddResidualBlock(fr0, nullptr,
//                                    // new ceres::HuberLoss(0.1 * rs0),
//                                    rot_param.data());
//           problem.AddResidualBlock(fr1, nullptr,
//                                    // new ceres::HuberLoss(0.1 * rs1),
//                                    rot_param.data());
//         }
//       }
//     }
//     // INFOLOG("residual block");

//     ceres::Solver::Options options;
//     options.linear_solver_type = ceres::DENSE_QR;
//     options.max_num_iterations = 100;
//     options.minimizer_progress_to_stdout = true;
//     ceres::Solver::Summary summary;
//     ceres::Solve(options, &problem, &summary);

//     SE3 Twb_est(SO3::exp(rot_param), trans_param);
//     INFOLOG("pose before opt: {}",
//             frame_src->getTwb().translation().transpose());
//     INFOLOG("pose scan2map: {}", Twb_est.translation().transpose());
//     frame_src->setTwb(Twb_est);
//     INFOLOG("pose gt: {}", frame_src->Tgt_.translation().transpose());
//     // frame_src->setTwb(Twb_est);
//   }
// }

// void RegistrationBA::registration() {

//   double resolution = 0.4;

//   voxel_map_ = std::unique_ptr<VoxelMapOctree>(new
//   VoxelMapOctree(common::resolution));

//   for (size_t i = 0; i < frames_.size(); i++) {
//     auto frame = frames_[i];

//     vector<Segment *> segments;
//     voxel_map_->insertFrame(frame->cloud_, frame.get(), segments);

//     frame->segments_ = segments;
//   }

//   unordered_set<Cell *> active_cells;

//   for (auto &&frame : frames_) {
//     CHECK_NOTNULL(frame);
//     for (auto &&seg : frame->segments_) {
//       active_cells.insert(seg->cell_);
//     }
//   }
//   //   cout << "aaa" << endl;

//   // define optimization problem
//   ceres::Problem problem;
//   eigen_aligned_std_vector<Vector3d> rot_params(frames_.size());
//   eigen_aligned_std_vector<Vector3d> trans_params(frames_.size());

//   unordered_map<Frame *, size_t> frame_pose;
//   auto eval_cb = new SurfRegXXEvalCallBack;

//   eigen_aligned_std_vector<SE3> poses;

//   for (size_t i = 0; i < frames_.size(); i++) {
//     // CHECK_NOTNULL(frames_[i]);
//     auto &&frame = frames_[i];
//     ceres::LocalParameterization *parametrization = new ParamSO3();

//     rot_params[i] = frame->getTwb().so3().log();
//     trans_params[i] = frame->getTwb().translation();
//     poses.push_back(frame->getTwb());

//     problem.AddParameterBlock(rot_params[i].data(), 3, parametrization);
//     problem.AddParameterBlock(trans_params[i].data(), 3);

//     frame_pose[frame.get()] = i;

//     eval_cb->addFrame(frame.get(), rot_params[i].data(),
//                       trans_params[i].data());
//   }
//   problem.SetParameterBlockConstant(rot_params[0].data());
//   problem.SetParameterBlockConstant(trans_params[0].data());
//   //   cout << "a" << endl;

//   // STEP. plannar
//   for (auto &&cell : active_cells) {
//     // TODO: check optable
//     CHECK_NOTNULL(cell);
//     CHECK_EQ(cell->frames_.size(), cell->segments_.size());
//     CHECK(!cell->frames_.empty());

//     vector<Frame *> frames(cell->frames_.begin(), cell->frames_.end());
//     vector<Segment *> segments(cell->segments_.begin(),
//     cell->segments_.end()); vector<bool> discard(frames.size(), true);

//     // vector<bool> &outliers = cell->is_outlier_;
//     // outliers.resize(frames.size(), false);
//     // cout << "c" << endl;

//     // CHECK optable
//     // TODO: insert segment check
//     int num_non_trivial = 0;
//     // for (auto &&segment : segments)
//     for (size_t i = 0; i < segments.size(); i++) {
//       auto &&segment = segments[i];

//       bool c1 = segment->gaussian_.num >= 8;
//       bool c2 = segment->gaussian_.scalar.x() < 0.1;
//       // bool c2 = true;

//       if (c1 && c2) {
//         num_non_trivial++;
//         discard[i] = false;
//       }
//     }
//     if (num_non_trivial < 2) {
//       continue;
//     }
//     // cout << "d" << endl;

//     // TODO: necessary?
//     cell->is_outlier_ = discard;
//     cell->updateInfoAll(false);
//     // cout << "e" << endl;

//     // TODO:
//     // if (!cell->optable_) {
//     //   continue;
//     // }
//     Vector3d nhat = cell->info_all_.rot_mat.col(0);
//     // cout << nhat << endl;
//     // SE3 pose_tgt(cell->rot_, cell->mean_);

//     vector<EigenFactorRotAxis *> rfs;
//     vector<EigenFactorRotAxis *> rfs1;
//     vector<EigenFactorTrans2 *> tfs;

//     // for (auto &&frame : frames_)
//     for (size_t i = 0; i < frames.size(); i++) {
//       auto frame = frames[i];
//       if (!frame || discard[i]) {
//         continue;
//       }

//       auto param_idx = frame_pose[frame];
//       Vector3d &param_rot = rot_params[param_idx];
//       Vector3d &param_t = trans_params[param_idx];

//       auto segment = segments[i];
//       const Gaussian3d &info_map = segment->cell_->info_all_;
//       const Gaussian3d &info_local = segment->gaussian_;

//       double scalar = sqrt(info_local.num * 1.0 / info_map.num);
//       // double scalar = sqrt(info_local.num * 1.0);
//       // double scalar = 1.0;
//       ceres::CostFunction *ft =
//           new EigenFactorTrans2(&info_map, info_local.mean, scalar);
//       problem.AddResidualBlock(ft,
//                                nullptr,
//                               //  new ceres::HuberLoss(0.1 * scalar),
//                                param_rot.data(), param_t.data());

//       // if (info_local.scalar.y() / info_local.scalar.x() >=
//       //     mapping::scan2map::eigen_ratio_surf) {
//       //   if (info_local.scalar.y() / info_local.scalar.x() >= 5.0)
//       {
//         double rs0 =
//             sqrt(info_local.scalar.y() * info_local.num / info_map.num);
//         double rs1 =
//             sqrt(info_local.scalar.z() * info_local.num / info_map.num);
//         // double rs2 =
//         //     sqrt(info_local.scalar.x() * info_local.num / info_map.num);
//         // if (isnan(rs0)) {
//         //   WARNLOG("fuck");
//         //   WARNLOG("{}, {}, {}", info_local.scalar.y(), info_local.num,
//         //           info_map.num);
//         // }
//         // double rs0 = sqrt(info_local.scalar.y() * info_local.num);
//         // double rs1 = sqrt(info_local.scalar.z() * info_local.num);
//         // double rs0 = sqrt(info_local.scalar.y());
//         // double rs1 = sqrt(info_local.scalar.z());
//         // double rs2 = sqrt(info_local.scalar.z());
//         // double rs0 = 1.0;
//         // double rs1 = 1.0;
//         // double rs2 = 1.0;
//         // if (rs0 <= 1e-8) {
//         //   cout << rs0 << ' ' << info_local.scalar.y() << ' ' <<
//         //   info_local.num
//         //        << endl;
//         //   exit(0);
//         // }
//         // cout << rs0 << ' ' << rs1 << endl;
//         ceres::CostFunction *fr0 =
//             new EigenFactorRotAxis(&info_map, info_local.rot_mat.col(1),
//             rs0);
//         ceres::CostFunction *fr1 =
//             new EigenFactorRotAxis(&info_map, info_local.rot_mat.col(2),
//             rs1);
//         // ceres::CostFunction *fr3 =
//         //     new EigenFactorRotAxis(&info_map, info_local.rot_mat.col(0),
//         //     rs2);
//         problem.AddResidualBlock(fr0,
//                                  nullptr,
//                                 //  new ceres::HuberLoss(0.1 * rs0),
//                                  param_rot.data());
//         problem.AddResidualBlock(fr1,
//                                  nullptr,
//                                 //  new ceres::HuberLoss(0.1 * rs1),
//                                  param_rot.data());
//         // problem.AddResidualBlock(fr3, nullptr,
//         //                          //  new ceres::HuberLoss(0.1 * rs1),
//         //                          param_rot.data());
//       }
//     }
//     // cout << "e" << endl;

//     // eval_cb->addPlane(cell, rfs, rfs1, tfs);
//     // TODO:
//     eval_cb->addPlane(cell);
//   }

//   ceres::Solver::Options options;
//   // options.update_state_every_iteration = true;
//   options.linear_solver_type = ceres::DENSE_SCHUR;
//   options.minimizer_progress_to_stdout = true;
//   // options.num_threads = 2;
//   options.trust_region_strategy_type = ceres::LEVENBERG_MARQUARDT;
//   // options.trust_region_strategy_type = ceres::DOGLEG;
//   // options.minimizer_type = ceres::LINE_SEARCH;
//   options.max_num_iterations = 10000;
//   options.num_threads = common::num_threads;
//   // options.function_tolerance = 1e-8;
//   // options.check_gradients = true;
//   options.evaluation_callback = eval_cb;

//   ceres::Solver::Summary summary;
//   ceres::Solve(options, &problem, &summary);
//   cout << summary.FullReport() << endl;

//   SE3 Tb0w = frames_[0]->getTbw();
//   double avg_err = 0;
//   for (size_t i = 0; i < frames_.size(); i++) {
//     // frames_[i]->setTwc(
//     //     SE3(SO3::exp(poses_param[i].topRows(3)),
//     auto &&frame = frames_[i];

//     // INFOLOG("frame {}, before opt: {}", i,
//     poses[i].translation().transpose());
//     //     poses_param[i].bottomRows(3)));
//     frame->setTwb(Tb0w * SE3(SO3::exp(rot_params[i]), trans_params[i]));
//     // INFOLOG("frame {}, after opt: {}", i,
//     //         frame->getTwb().translation().transpose());
//     SE3 Terr = poses[i].inverse() * frame->getTwb();
//     avg_err += Terr.translation().norm();
//     // poses[i]);
//   }
//   avg_err /= frames_.size();
//   // INFOLOG("err: {}", avg_err);
// }

}  // namespace xreg