#include "xreg/estimation/registration_base.h"

#include <pcl/filters/extract_indices.h>
#include <pcl/filters/uniform_sampling.h>
#include <pcl/filters/voxel_grid.h>

#include <pcl/features/fpfh_omp.h>
#include <pcl/features/normal_3d_omp.h>

#include <pcl/registration/correspondence_rejection_sample_consensus.h>
#include <pcl/registration/ia_ransac.h>

#include "xreg/common/config.h"

namespace xreg {
using namespace std;

// constexpr double downsample_resolution = 0.1;
// constexpr double kps_resolution = downsample_resolution * 4;

void RegistrationBase::detectKeyPoints(xreg::PointCloudXYZ::Ptr cloud,
                                       xreg::PointCloudXYZ::Ptr &kps,
                                       pcl::PointIndices::Ptr &indices) {

  kps = xreg::PointCloudXYZ::Ptr(new pcl::PointCloud<pcl::PointXYZ>);

  pcl::UniformSampling<pcl::PointXYZ> sampler;
  sampler.setInputCloud(cloud);
  sampler.setRadiusSearch(feature::resolution_keypoints);
  sampler.filter(*kps);

  // pcl::PointIndices::Ptr indices(new pcl::PointIndices);

  pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> est_normal;
  pcl::search::KdTree<pcl::PointXYZ>::Ptr tree(
      new pcl::search::KdTree<pcl::PointXYZ>());

  // get indices
  indices = pcl::PointIndices::Ptr(new pcl::PointIndices);
  {
    tree->setInputCloud(cloud);
    for (pcl::PointXYZ &pt : kps->points) {
      std::vector<int> inds;
      vector<float> dists;
      tree->nearestKSearch(pt, 1, inds, dists);
      indices->indices.push_back(inds[0]);
    }
  }

  INFOLOG("#kps {}", kps->size());
}

void RegistrationBase::findCorrespondences(
    typename pcl::PointCloud<pcl::FPFHSignature33>::ConstPtr source,
    typename pcl::PointCloud<pcl::FPFHSignature33>::ConstPtr target,
    std::vector<std::pair<int, int>> &corr) {
  // std::cout <<  << std::flush;
  INFOLOG("correspondence assignment...");
  std::vector<int> corr_s2t, corr_t2s;
  {
    corr_s2t.resize(source->size());

    // Use a KdTree to search for the nearest matches in feature space
    pcl::KdTreeFLANN<pcl::FPFHSignature33> descriptor_kdtree;
    descriptor_kdtree.setInputCloud(target);

    // Find the index of the best match for each keypoint, and store it in
    // "correspondences_out"
    const int k = 1;
    std::vector<int> k_indices(k);
    std::vector<float> k_squared_distances(k);
    for (int i = 0; i < static_cast<int>(source->size()); ++i) {
      descriptor_kdtree.nearestKSearch(*source, i, k, k_indices,
                                       k_squared_distances);
      corr_s2t[i] = k_indices[0];
    }
  }
  {
    corr_t2s.resize(target->size());

    // Use a KdTree to search for the nearest matches in feature space
    pcl::KdTreeFLANN<pcl::FPFHSignature33> descriptor_kdtree;
    descriptor_kdtree.setInputCloud(source);

    // Find the index of the best match for each keypoint, and store it in
    // "correspondences_out"
    const int k = 1;
    std::vector<int> k_indices(k);
    std::vector<float> k_squared_distances(k);
    for (int i = 0; i < static_cast<int>(target->size()); ++i) {
      descriptor_kdtree.nearestKSearch(*target, i, k, k_indices,
                                       k_squared_distances);
      corr_t2s[i] = k_indices[0];
    }
  }

  for (std::size_t cIdx = 0; cIdx < corr_s2t.size(); ++cIdx)
    if (corr_t2s[corr_s2t[cIdx]] == static_cast<int>(cIdx))
      corr.push_back(std::make_pair(cIdx, corr_s2t[cIdx]));
  // std::cout << "OK" << std::endl;
  INFOLOG("done");
}

void RegistrationBase::extractDescriptors(
    pcl::PointCloud<pcl::PointXYZ>::ConstPtr input,
    const pcl::PointCloud<pcl::PointXYZ>::Ptr &keypoints,
    pcl::PointCloud<pcl::FPFHSignature33>::Ptr features) {

  typename pcl::PointCloud<pcl::PointXYZ>::Ptr kpts(
      new pcl::PointCloud<pcl::PointXYZ>);
  kpts->points.resize(keypoints->size());

  pcl::copyPointCloud(*keypoints, *kpts);

  // auto est_fpfh =
  //     pcl::FPFHEstimation<pcl::PointXYZ, pcl::Normal, pcl::FPFHSignature33>::
  //         Ptr(new pcl::FPFHEstimation<pcl::PointXYZ, pcl::Normal,
  //                                     pcl::FPFHSignature33>);
  auto est_fpfh =
      pcl::FPFHEstimationOMP<pcl::PointXYZ, pcl::Normal, pcl::FPFHSignature33>::
          Ptr(new pcl::FPFHEstimationOMP<pcl::PointXYZ, pcl::Normal,
                                         pcl::FPFHSignature33>);

  est_fpfh->setNumberOfThreads(common::num_threads);

  est_fpfh->setSearchSurface(input);
  est_fpfh->setInputCloud(kpts);

  // if (feature_from_normals)
  {
    // std::cout << "normal estimation..." << std::flush;
    INFOLOG("normal estimation");
    pcl::PointCloud<pcl::Normal>::Ptr normals(new pcl::PointCloud<pcl::Normal>);
    // pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> normal_estimation;
    pcl::NormalEstimationOMP<pcl::PointXYZ, pcl::Normal> normal_estimation;

    normal_estimation.setNumberOfThreads(common::num_threads);

    normal_estimation.setSearchMethod(pcl::search::Search<pcl::PointXYZ>::Ptr(
        new pcl::search::KdTree<pcl::PointXYZ>));

    // normal_estimation.setRadiusSearch(downsample_resolution * 4);
    normal_estimation.setRadiusSearch(feature::resolution_normal);
    normal_estimation.setInputCloud(input);
    normal_estimation.compute(*normals);

    est_fpfh->setInputNormals(normals);
    est_fpfh->setSearchMethod(pcl::search::Search<pcl::PointXYZ>::Ptr(
        new pcl::search::KdTree<pcl::PointXYZ>));
    // est_fpfh->setRadiusSearch(kps_resolution * 4);
    est_fpfh->setRadiusSearch(feature::resolution_fpfh);
    // std::cout << "OK" << std::endl;
    INFOLOG("done.");
  }

  // std::cout << "descriptor extraction..." << std::flush;

  INFOLOG("compute fpfh.");
  est_fpfh->compute(*features);
  // est_fpfh->
  // std::cout << "OK" << std::endl;
  INFOLOG("done.");
}

} // namespace xreg
