#include "xreg/estimation/registration_fgr.h"

#include "xreg/types/cell.h"

#include "xreg/optimization/factors.h"
#include "xreg/optimization/factors_pr.h"
#include "xreg/optimization/se3_param.hpp"

#include "xreg/common/config.h"

#include "fgr/fgr.h"

#include <omp.h>
#include <teaser/registration.h>

#include <Open3D/Registration/FastGlobalRegistration.h>

#include <pcl/filters/voxel_grid.h>
#include <pcl/registration/ia_ransac.h>

#include <fstream>

namespace xreg {

using namespace std;

RegistrationFGR::RegistrationFGR() : RegistrationBase() {
  // cout
}

void RegistrationFGR::optimizePoseGraph() {
  CHECK_EQ(relative_poses_.size(), olp_frames_.size());

  size_t num_frames = frames_.size();

  // initialize frame info
  eigen_aligned_std_vector<Vector3d> trans_params;
  eigen_aligned_std_vector<Vector3d> rot_params;

  ceres::LocalParameterization *parametrization = new ParamSO3();

  ceres::Problem problem;
  for (auto &&frame : frames_) {
    rot_params.push_back(frame->Twb_.so3().log());
    trans_params.push_back(frame->Twb_.translation());
    problem.AddParameterBlock(rot_params.back().data(), 3, parametrization);
    problem.AddParameterBlock(trans_params.back().data(), 3);
  }

  int num_outlier = 0;
  for (size_t i = 0; i < olp_frames_.size(); i++) {
    auto pair = olp_frames_[i];
    auto frame_tgt = pair.first;
    auto frame_src = pair.second;
    auto tgt_idx = frame_tgt->idx_;
    auto src_idx = frame_src->idx_;

    SE3 Tts = relative_poses_[i];

    // cheating
    // SE3 Tts_ini = frame_tgt->Twb_.inverse() * frame_src->Twb_;
    SE3 Tts_ini = frame_tgt->Tgt_.inverse() * frame_src->Tgt_;
    SE3 Terr = Tts_ini.inverse() * Tts;
    if (Terr.translation().norm() > 0.3 ||
        Terr.so3().log().norm() > M_PI / 180.0 * 10) {
      num_outlier++;
      continue;
    }

    ceres::CostFunction *rf = new RotationFactor(Tts.so3());
    problem.AddResidualBlock(rf, new ceres::HuberLoss(0.1),
                             rot_params[src_idx].data(),
                             rot_params[tgt_idx].data());

    ceres::CostFunction *tf = new TranslationFactor(Tts);
    problem.AddResidualBlock(
        tf, new ceres::HuberLoss(0.09), rot_params[src_idx].data(),
        trans_params[src_idx].data(), trans_params[tgt_idx].data());
  }
  WARNLOG("num_outlier: {}", num_outlier);

  problem.SetParameterBlockConstant(rot_params[0].data());
  problem.SetParameterBlockConstant(trans_params[0].data());

  ceres::Solver::Options options;
  options.linear_solver_type = ceres::DENSE_QR;
  options.max_num_iterations = 100;
  options.minimizer_progress_to_stdout = false;
  ceres::Solver::Summary summary;
  ceres::Solve(options, &problem, &summary);

  for (size_t i = 0; i < frames_.size(); i++) {
    SO3 roti = SO3::exp(rot_params[i]);
    Vector3d ti = trans_params[i];
    frames_[i]->setTwb(SE3(roti, ti));
  }

  cout << summary.FullReport() << endl;
}

// void RegistrationFGR::initialize() {
//   INFOLOG("initialize");

//   // Tlb_.push_back(SE3());

//   // for (size_t i = 1; i < frames_.size(); i++) {
//   //   auto &&f = frames_[i];
//   //   auto &&fl = frames_[i - 1];

//   //   // frame
//   //   SE3 Tlb = fl->getTbw() * f->getTwb();
//   //   Tlb_.push_back(Tlb);
//   // }

//   if (reg::b_downsample) {
//     INFOLOG("downsample");
//     pcl::VoxelGrid<pcl::PointXYZ> voxel_filer;
//     voxel_filer.setLeafSize(reg::resolution_downsample,
//                             reg::resolution_downsample,
//                             reg::resolution_downsample);
//     for (auto &&frame : frames_) {
//       frame->cloud_down_ = PointCloudXYZ::Ptr(new PointCloudXYZ);

//       voxel_filer.setInputCloud(frame->cloud_);
//       voxel_filer.filter(*frame->cloud_down_);
//     }
//   } else {
//     for (auto &&frame : frames_) {
//       frame->cloud_down_ = PointCloudXYZ::Ptr(new PointCloudXYZ);
//       *frame->cloud_down_ += *frame->cloud_;
//     }
//   }

//   for (auto &&frame : frames_) {
//     /* code */

//     // open3d::registration::FastGlobalRegistration()
//   }
// }

void RegistrationFGR::registration() {
  if (tpp::b_pgo) {
    // if (true) {

    for (size_t i = 0; i < frames_.size() - 1; i++) {
      auto frame_tgt = frames_[i];
      auto frame_src = frames_[i + 1];

      // frame_src->
      size_t j = 0;
      for (; j < olp_frames_.size(); j++) {
        auto pair = olp_frames_[j];
        if (pair.first == frame_tgt && pair.second == frame_src) {
          break;
        }
      }

      if (j == olp_frames_.size()) {
        throw std::runtime_error("no overlap between consequential frame.");
      }

      // const auto &matches = matches_[j];

      // SE3 Tts = registerByTeaser(frame_tgt->kpts_, frame_src->kpts_,
      // matches);
      fgr::CApp *app = new fgr::CApp;

      eigen_aligned_std_vector<Vector3f> pts_src, pts_tgt;
      std::vector<VectorXf> feat_src, feat_tgt;

      CHECK_EQ(frame_tgt->kpts_->size(), frame_tgt->features_->size());
      for (size_t ipt = 0; ipt < frame_tgt->kpts_->size(); ipt++) {
        pcl::PointXYZ ptp = frame_tgt->kpts_->points[ipt];
        Vector3f pt(ptp.x, ptp.y, ptp.z);
        pts_tgt.push_back(pt);

        VectorXf feat = Eigen::Map<VectorXf>(
            frame_tgt->features_->points[ipt].histogram, 33);
        feat_tgt.push_back(feat);
      }
      CHECK_EQ(frame_src->kpts_->size(), frame_src->features_->size());
      for (size_t ipt = 0; ipt < frame_src->kpts_->size(); ipt++) {
        pcl::PointXYZ ptp = frame_src->kpts_->points[ipt];
        Vector3f pt(ptp.x, ptp.y, ptp.z);
        pts_src.push_back(pt);

        VectorXf feat = Eigen::Map<VectorXf>(
            frame_src->features_->points[ipt].histogram, 33);
        feat_src.push_back(feat);
      }

      app->LoadFeature(pts_tgt, feat_tgt);
      app->LoadFeature(pts_src, feat_src);
      app->NormalizePoints();
      app->AdvancedMatching();
      app->OptimizePairwise(true);

      Eigen::Matrix4f T = app->GetOutputTrans();
      Matrix3f rot_mat = T.topLeftCorner(3, 3);
      Quaterniond q = Quaternionf(rot_mat).cast<double>();
      Vector3d t = T.topRightCorner(3, 1).cast<double>();

      // cout << T << endl << endl;
      // SE3 pose = frame_tgt->Tgt_.inverse() * frame_src->Tgt_;
      // cout << pose.translation().transpose() << endl;
      // getchar();

      // INFOLOG();
      SE3 Test(q, t);
      // frame_src->setTwb(frame_tgt->getTwb() * Test);

      delete app;

      frame_src->setTwb(frame_tgt->getTwb() * Test);
    }

    for (size_t i = 0; i < olp_frames_.size(); i++) {
      auto pair = olp_frames_[i];
      auto frame_tgt = pair.first;
      auto frame_src = pair.second;
      // const auto &matches = matches_[i];

      // SE3 Tts = registerByTeaser(frame_tgt->kpts_, frame_src->kpts_,
      // matches);
      fgr::CApp *app = new fgr::CApp;

      eigen_aligned_std_vector<Vector3f> pts_src, pts_tgt;
      std::vector<VectorXf> feat_src, feat_tgt;

      CHECK_EQ(frame_tgt->kpts_->size(), frame_tgt->features_->size());
      for (size_t ipt = 0; ipt < frame_tgt->kpts_->size(); ipt++) {
        pcl::PointXYZ ptp = frame_tgt->kpts_->points[ipt];
        Vector3f pt(ptp.x, ptp.y, ptp.z);
        pts_tgt.push_back(pt);

        VectorXf feat = Eigen::Map<VectorXf>(
            frame_tgt->features_->points[ipt].histogram, 33);
        feat_tgt.push_back(feat);
      }
      CHECK_EQ(frame_src->kpts_->size(), frame_src->features_->size());
      for (size_t ipt = 0; ipt < frame_src->kpts_->size(); ipt++) {
        pcl::PointXYZ ptp = frame_src->kpts_->points[ipt];
        Vector3f pt(ptp.x, ptp.y, ptp.z);
        pts_src.push_back(pt);

        VectorXf feat = Eigen::Map<VectorXf>(
            frame_src->features_->points[ipt].histogram, 33);
        feat_src.push_back(feat);
      }

      app->LoadFeature(pts_tgt, feat_tgt);
      app->LoadFeature(pts_src, feat_src);
      app->NormalizePoints();
      app->AdvancedMatching();
      app->OptimizePairwise(true);

      Eigen::Matrix4f T = app->GetOutputTrans();
      Matrix3f rot_mat = T.topLeftCorner(3, 3);
      Quaterniond q = Quaternionf(rot_mat).cast<double>();
      Vector3d t = T.topRightCorner(3, 1).cast<double>();

      // cout << T << endl << endl;
      // SE3 pose = frame_tgt->Tgt_.inverse() * frame_src->Tgt_;
      // cout << pose.translation().transpose() << endl;
      // getchar();

      // INFOLOG();
      SE3 Test(q, t);
      // frame_src->setTwb(frame_tgt->getTwb() * Test);

      delete app;
      relative_poses_.push_back(Test);
    }

    // STEP. write to g2o
    // ofstream ofs("/home/hyhuang/test.g2o");
    // ofs << fixed << setprecision(8);
    // for (size_t i = 0; i < frames_.size(); i++) {
    //   auto &&frame = frames_[i];
    //   auto idx = frame->idx_;
    //   SE3 Twb = frame->getTwb();

    //   Quaterniond q = Twb.unit_quaternion();
    //   Vector3d t = Twb.translation();

    //   ofs << "VERTEX_SE3:QUAT " << idx << ' ' << t.x() << ' ' << t.y() << ' '
    //       << t.z() << ' ' << q.x() << ' ' << q.y() << ' ' << q.z() << ' '
    //       << q.w() << endl;
    // }
    // for (size_t i = 0; i < olp_frames_.size(); i++) {
    //   auto pair = olp_frames_[i];
    //   auto frame_tgt = pair.first;
    //   auto frame_src = pair.second;
    //   auto tgt_idx = frame_tgt->idx_;
    //   auto src_idx = frame_src->idx_;

    //   SE3 Tts = relative_poses_[i];

    //   Quaterniond q = Tts.unit_quaternion();
    //   Vector3d t = Tts.translation();

    //   ofs << "EDGE_SE3:QUAT " << tgt_idx << ' ' << src_idx << ' ' << ' '
    //       << t.x() << ' ' << t.y() << ' ' << t.z() << ' ' << q.x() << ' '
    //       << q.y() << ' ' << q.z() << ' ' << q.w() << " 1 0 0 0 0 0 0 1 0 0 0
    //       0 0 0 1 0 0 0 0 0 0 1 0 0 0 0 0 0 1 0 0 0 "
    //          "0 0 0 1"
    //       << endl;
    // }

    optimizePoseGraph();

  } else {
    frames_[0]->setTwb(SE3());
    for (size_t i = 0; i < frames_.size() - 1; i++) {
      auto frame_tgt = frames_[i];
      auto frame_src = frames_[i + 1];

      // frame_src->
      size_t j = 0;
      for (; j < olp_frames_.size(); j++) {
        auto pair = olp_frames_[j];
        if (pair.first == frame_tgt && pair.second == frame_src) {
          break;
        }
      }

      if (j == olp_frames_.size()) {
        throw std::runtime_error("no overlap between consequential frame.");
      }

      // const auto &matches = matches_[j];
      fgr::CApp *app = new fgr::CApp;

      eigen_aligned_std_vector<Vector3f> pts_src, pts_tgt;
      std::vector<VectorXf> feat_src, feat_tgt;

      CHECK_EQ(frame_tgt->kpts_->size(), frame_tgt->features_->size());
      for (size_t ipt = 0; ipt < frame_tgt->kpts_->size(); ipt++) {
        pcl::PointXYZ ptp = frame_tgt->kpts_->points[ipt];
        Vector3f pt(ptp.x, ptp.y, ptp.z);
        pts_tgt.push_back(pt);

        VectorXf feat = Eigen::Map<VectorXf>(
            frame_tgt->features_->points[ipt].histogram, 33);
        feat_tgt.push_back(feat);
      }
      CHECK_EQ(frame_src->kpts_->size(), frame_src->features_->size());
      for (size_t ipt = 0; ipt < frame_src->kpts_->size(); ipt++) {
        pcl::PointXYZ ptp = frame_src->kpts_->points[ipt];
        Vector3f pt(ptp.x, ptp.y, ptp.z);
        pts_src.push_back(pt);

        VectorXf feat = Eigen::Map<VectorXf>(
            frame_src->features_->points[ipt].histogram, 33);
        feat_src.push_back(feat);
      }

      app->LoadFeature(pts_tgt, feat_tgt);
      app->LoadFeature(pts_src, feat_src);
      app->NormalizePoints();
      app->AdvancedMatching();
      app->OptimizePairwise(true);

      Eigen::Matrix4f T = app->GetOutputTrans();
      Matrix3f rot_mat = T.topLeftCorner(3, 3);
      Quaterniond q = Quaternionf(rot_mat).cast<double>();
      Vector3d t = T.topRightCorner(3, 1).cast<double>();

      // cout << T << endl << endl;
      // SE3 pose = frame_tgt->Tgt_.inverse() * frame_src->Tgt_;
      // cout << pose.translation().transpose() << endl;
      // getchar();

      // INFOLOG();
      SE3 Test(q, t);
      frame_src->setTwb(frame_tgt->getTwb() * Test);

      delete app;
    }
  }
}

}  // namespace xreg