#include "xreg/estimation/registration_icp_pt2pl.h"

#include <pcl/registration/icp.h>

#include <pcl/features/normal_3d.h>
#include <pcl/filters/voxel_grid.h>

#include "xreg/common/config.h"

namespace xreg {

RegistrationICPPlane::RegistrationICPPlane(/* args */) : RegistrationBase() {}

void RegistrationICPPlane::initialize() {
  INFOLOG("initialize");

  Tlb_.push_back(SE3());

  for (size_t i = 1; i < frames_.size(); i++) {
    auto &&f = frames_[i];
    auto &&fl = frames_[i - 1];

    // frame
    SE3 Tlb = fl->getTbw() * f->getTwb();
    Tlb_.push_back(Tlb);
  }

  // auto  = []() {

  // }
  auto concat_normal =
      [](pcl::PointCloud<pcl::PointXYZ>::Ptr cloud,
         pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr cloud_with_normals) {
        pcl::PointCloud<pcl::Normal>::Ptr normals(
            new pcl::PointCloud<pcl::Normal>);

        pcl::search::KdTree<pcl::PointXYZ>::Ptr searchTree(
            new pcl::search::KdTree<pcl::PointXYZ>);
        searchTree->setInputCloud(cloud);

        pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> normalEstimator;
        normalEstimator.setInputCloud(cloud);
        normalEstimator.setSearchMethod(searchTree);

        normalEstimator.setKSearch(common::num_knn);

        normalEstimator.compute(*normals);

        pcl::concatenateFields(*cloud, *normals, *cloud_with_normals);
      };

  if (reg::b_downsample) {
    INFOLOG("downsample");
    pcl::VoxelGrid<pcl::PointXYZ> voxel_filer;
    voxel_filer.setLeafSize(reg::resolution_downsample,
                            reg::resolution_downsample,
                            reg::resolution_downsample);
    for (auto &&frame : frames_) {
      frame->cloud_down_ = PointCloudXYZ::Ptr(new PointCloudXYZ);

      voxel_filer.setInputCloud(frame->cloud_);
      voxel_filer.filter(*frame->cloud_down_);
    }
  } else {
    for (auto &&frame : frames_) {
      frame->cloud_down_ = PointCloudXYZ::Ptr(new PointCloudXYZ);
      *frame->cloud_down_ += *frame->cloud_;
    }
  }

  // if (common::method == "icp_pt2pl")
  {
    for (size_t i = 0; i < frames_.size(); i++) {
      auto &frame = frames_[i];
      frame->cloud_wn_ = pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr(
          new pcl::PointCloud<pcl::PointXYZRGBNormal>);
      concat_normal(frame->cloud_down_, frame->cloud_wn_);
    }
  }
}

template <>
void RegistrationICPPlane::scan2map(
    pcl::Registration<pcl::PointXYZRGBNormal, pcl::PointXYZRGBNormal> *reg) {
  INFOLOG("scan2map.");
  double avg_err = 0;
  pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr cloud_map(
      new pcl::PointCloud<pcl::PointXYZRGBNormal>);
  (*cloud_map) += *frames_[0]->cloud_wn_;

  // initialize voxel filter
  pcl::VoxelGrid<pcl::PointXYZRGBNormal> voxel_filer;
  if (reg::b_downsample) {
    voxel_filer.setLeafSize(reg::resolution_downsample,
                            reg::resolution_downsample,
                            reg::resolution_downsample);
  }

  for (size_t i = 1; i < frames_.size(); i++) {
    INFOLOG("frame {}.", i);
    auto frame_src = frames_[i];
    auto frame_tgt = frames_[i - 1];

    auto cloud_src = frame_src->cloud_wn_;

    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr cloud_aligned(
        new pcl::PointCloud<pcl::PointXYZRGBNormal>);

    SE3 Tpred = frame_tgt->getTwb() * Tlb_[i];

    reg->setInputTarget(cloud_map);
    reg->setInputSource(cloud_src);

    reg->align(*cloud_aligned, Tpred.matrix().cast<float>());

    Matrix4d result = reg->getFinalTransformation().cast<double>();

    double fitness_score = reg->getFitnessScore();

    Matrix3d rot_mat_est = result.topLeftCorner(3, 3);
    Quaterniond rot_est(rot_mat_est);
    rot_est.normalize();
    SE3 Test(rot_est, result.topRightCorner(3, 1));

    frame_src->setTwb(Test);

    // aggregate map
    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr cloud_tmp(
        new pcl::PointCloud<pcl::PointXYZRGBNormal>);
    *cloud_tmp += *cloud_map;
    *cloud_tmp += *cloud_aligned;
    if (reg::b_downsample) {
      voxel_filer.setInputCloud(cloud_tmp);
      voxel_filer.filter(*cloud_map);
    }
  }
}

// template <typename PT, type>
template <>
void RegistrationICPPlane::scan2scan(
    pcl::Registration<pcl::PointXYZRGBNormal, pcl::PointXYZRGBNormal> *reg) {
  double avg_err = 0;
  for (size_t i = 1; i < frames_.size(); i++) {
    auto frame_src = frames_[i];
    auto frame_tgt = frames_[i - 1];

    auto cloud_src = frame_src->cloud_wn_;
    auto cloud_tgt = frame_tgt->cloud_wn_;

    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr cloud_aligned(
        new pcl::PointCloud<pcl::PointXYZRGBNormal>);

    auto Tsrc = frame_src->Tgt_;
    auto Ttgt = frame_tgt->Tgt_;

    SE3 Tpred = Tlb_[i];

    reg->setInputTarget(cloud_tgt);
    reg->setInputSource(cloud_src);

    reg->align(*cloud_aligned, Tpred.matrix().cast<float>());

    Matrix4d result = reg->getFinalTransformation().cast<double>();

    double fitness_score = reg->getFitnessScore();

    Matrix3d rot_mat_est = result.topLeftCorner(3, 3);
    Quaterniond rot_est(rot_mat_est);
    rot_est.normalize();
    SE3 Test(rot_est, result.topRightCorner(3, 1));

    SE3 Tgt = frame_tgt->Tgt_.inverse() * frame_src->Tgt_;
    SE3 Terr = Tgt.inverse() * Test;

    frame_src->setTwb(frame_tgt->getTwb() * Test);

    SE3 Terr_gt = Tsrc.inverse() * frame_src->getTwb();
    // INFOLOG("err2gt: {}", Terr_gt.translation().norm());
    avg_err += Terr_gt.translation().norm();
  }

  avg_err /= frames_.size();
}

void RegistrationICPPlane::registration() {
  // gicp::
  INFOLOG("pt2pl registration, method set to.");

  pcl::IterativeClosestPointWithNormals<pcl::PointXYZRGBNormal,
                                        pcl::PointXYZRGBNormal>::Ptr
      reg(new pcl::IterativeClosestPointWithNormals<pcl::PointXYZRGBNormal,
                                                    pcl::PointXYZRGBNormal>);

  reg->setMaxCorrespondenceDistance(common::corr_dist);

  // reg = icp;
  // }
  // else {
  //   throw std::runtime_error("unsupported registration method");
  // }

  if (reg::b_scan2map) {
    scan2map(reg.get());
  } else {
    scan2scan(reg.get());
  }
}

}  // namespace xreg