#include "xreg/estimation/registration_ndt.h"

#include "pcl/registration/icp.h"

#include "xreg/common/config.h"

#include "xreg/estimation/fast_gicp/fast_vgicp.h"
#include "xreg/estimation/fast_gicp/gicp_omp.h"
#include "xreg/estimation/fast_gicp/ndt_omp.h"

namespace xreg {

using namespace std;

RegistrationNDT::RegistrationNDT(const std::string &method, NDTOptions options)
    : RegistrationBase(), options_(options) {
  // frame reg
}

void RegistrationNDT::initialize() {
  INFOLOG("initialize");

  Tlb_.push_back(SE3());

  for (size_t i = 1; i < frames_.size(); i++) {
    auto &&f = frames_[i];
    auto &&fl = frames_[i - 1];

    // frame
    SE3 Tlb = fl->getTbw() * f->getTwb();
    Tlb_.push_back(Tlb);
  }

  // auto  = []() {

  // }
  auto concat_normal =
      [](pcl::PointCloud<pcl::PointXYZ>::Ptr cloud,
         pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr cloud_with_normals) {
        pcl::PointCloud<pcl::Normal>::Ptr normals(
            new pcl::PointCloud<pcl::Normal>);

        pcl::search::KdTree<pcl::PointXYZ>::Ptr searchTree(
            new pcl::search::KdTree<pcl::PointXYZ>);
        searchTree->setInputCloud(cloud);

        pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> normalEstimator;
        normalEstimator.setInputCloud(cloud);
        normalEstimator.setSearchMethod(searchTree);

        normalEstimator.setKSearch(common::num_knn);

        normalEstimator.compute(*normals);

        pcl::concatenateFields(*cloud, *normals, *cloud_with_normals);
      };

  if (reg::b_downsample) {
    INFOLOG("downsample");
    pcl::VoxelGrid<pcl::PointXYZ> voxel_filer;
    voxel_filer.setLeafSize(reg::resolution_downsample,
                            reg::resolution_downsample,
                            reg::resolution_downsample);
    for (auto &&frame : frames_) {
      frame->cloud_down_ = PointCloudXYZ::Ptr(new PointCloudXYZ);

      voxel_filer.setInputCloud(frame->cloud_);
      voxel_filer.filter(*frame->cloud_down_);
    }
  } else {
    for (auto &&frame : frames_) {
      frame->cloud_down_ = PointCloudXYZ::Ptr(new PointCloudXYZ);
      *frame->cloud_down_ += *frame->cloud_;
    }
  }

  if (common::method == "icp_pt2pl") {
    for (size_t i = 0; i < frames_.size(); i++) {
      auto &frame = frames_[i];
      frame->cloud_wn_ = pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr(
          new pcl::PointCloud<pcl::PointXYZRGBNormal>);
      concat_normal(frame->cloud_down_, frame->cloud_wn_);
    }
  }
}

template <>
void RegistrationNDT::scan2map(pcl::Registration<PointXYZ, PointXYZ> *reg) {
  INFOLOG("scan2map.");
  double avg_err = 0;
  PointCloudXYZ::Ptr cloud_map(new PointCloudXYZ);
  (*cloud_map) += *frames_[0]->cloud_down_;

  // initialize voxel filter
  pcl::VoxelGrid<pcl::PointXYZ> voxel_filer;
  if (reg::b_downsample) {
    voxel_filer.setLeafSize(reg::resolution_downsample,
                            reg::resolution_downsample,
                            reg::resolution_downsample);
  }

  for (size_t i = 1; i < frames_.size(); i++) {
    INFOLOG("frame {}.", i);
    auto frame_src = frames_[i];
    auto frame_tgt = frames_[i - 1];

    auto cloud_src = frame_src->cloud_down_;

    PointCloudXYZ::Ptr cloud_aligned(new PointCloudXYZ);

    SE3 Tpred = frame_tgt->getTwb() * Tlb_[i];

    reg->setInputTarget(cloud_map);
    reg->setInputSource(cloud_src);

    reg->align(*cloud_aligned, Tpred.matrix().cast<float>());

    Matrix4d result = reg->getFinalTransformation().cast<double>();

    double fitness_score = reg->getFitnessScore();

    Matrix3d rot_mat_est = result.topLeftCorner(3, 3);
    Quaterniond rot_est(rot_mat_est);
    rot_est.normalize();
    SE3 Test(rot_est, result.topRightCorner(3, 1));

    frame_src->setTwb(Test);

    // aggregate map
    PointCloudXYZ::Ptr cloud_tmp(new PointCloudXYZ);
    *cloud_tmp += *cloud_map;
    *cloud_tmp += *cloud_aligned;
    if (reg::b_downsample) {
      voxel_filer.setInputCloud(cloud_tmp);
      voxel_filer.filter(*cloud_map);
    }
  }
}

// template <typename PT, type>
template <>
void RegistrationNDT::scan2scan(pcl::Registration<PointXYZ, PointXYZ> *reg) {
  double avg_err = 0;
  for (size_t i = 1; i < frames_.size(); i++) {
    auto frame_src = frames_[i];
    auto frame_tgt = frames_[i - 1];

    auto cloud_src = frame_src->cloud_down_;
    auto cloud_tgt = frame_tgt->cloud_down_;

    PointCloudXYZ::Ptr cloud_aligned(new PointCloudXYZ);

    auto Tsrc = frame_src->Tgt_;
    auto Ttgt = frame_tgt->Tgt_;

    SE3 Tpred = Tlb_[i];

    reg->setInputTarget(cloud_tgt);
    reg->setInputSource(cloud_src);

    reg->align(*cloud_aligned, Tpred.matrix().cast<float>());

    Matrix4d result = reg->getFinalTransformation().cast<double>();

    double fitness_score = reg->getFitnessScore();

    Matrix3d rot_mat_est = result.topLeftCorner(3, 3);
    Quaterniond rot_est(rot_mat_est);
    rot_est.normalize();
    SE3 Test(rot_est, result.topRightCorner(3, 1));

    SE3 Tgt = frame_tgt->Tgt_.inverse() * frame_src->Tgt_;
    SE3 Terr = Tgt.inverse() * Test;

    frame_src->setTwb(frame_tgt->getTwb() * Test);

    SE3 Terr_gt = Tsrc.inverse() * frame_src->getTwb();
    // INFOLOG("err2gt: {}", Terr_gt.translation().norm());
    avg_err += Terr_gt.translation().norm();
  }

  avg_err /= frames_.size();
}

void RegistrationNDT::registration() {
  // gicp::
  INFOLOG("registration, method set to: {}", options_.method);

  pcl::Registration<PointXYZ, PointXYZ>::Ptr reg;

  if (options_.method == "ndt") {
    gicp::NormalDistributionsTransform<PointXYZ, PointXYZ>::Ptr _reg(
        new gicp::NormalDistributionsTransform<PointXYZ, PointXYZ>());

    _reg->setResolution(common::resolution);
    if (common::search_method == 0) {
      _reg->setNeighborhoodSearchMethod(gicp::KDTREE);
    } else if (common::search_method == 1) {
      _reg->setNeighborhoodSearchMethod(gicp::DIRECT7);
    } else if (common::search_method == 2) {
      _reg->setNeighborhoodSearchMethod(gicp::DIRECT1);
    }

    _reg->setNumThreads(common::num_threads);
    _reg->setMaximumIterations(reg::num_iteration);

    reg = _reg;
  } else if (options_.method == "gicp") {
    gicp::GeneralizedIterativeClosestPoint<PointXYZ, PointXYZ>::Ptr _reg(
        new gicp::GeneralizedIterativeClosestPoint<PointXYZ, PointXYZ>());
    reg = _reg;
  } else if (options_.method == "vgicp") {
    gicp::FastVGICP<PointXYZ, PointXYZ>::Ptr _reg(
        new gicp::FastVGICP<PointXYZ, PointXYZ>());

    _reg->setResolution(common::resolution);
    if (common::search_method == 0) {
      _reg->setNeighborSearchMethod(gicp::KDTREE);
    } else if (common::search_method == 1) {
      _reg->setNeighborSearchMethod(gicp::DIRECT7);
    } else if (common::search_method == 2) {
      _reg->setNeighborSearchMethod(gicp::DIRECT1);
    }
    _reg->setNumThreads(common::num_threads);
    reg = _reg;
  } else if (options_.method == "icp_pt2pt") {
    pcl::IterativeClosestPoint<PointXYZ, PointXYZ>::Ptr icp(
        new pcl::IterativeClosestPoint<PointXYZ, PointXYZ>);

    // icp->setMaxCorrespondenceDistance(0.1);
    icp->setMaxCorrespondenceDistance(common::corr_dist);

    reg = icp;
  } else {
    throw std::runtime_error("unsupported registration method");
  }

  if (options_.b_scan2map) {
    scan2map(reg.get());
  } else {
    scan2scan(reg.get());
  }
}

}  // namespace xreg