#include "xreg/estimation/registration_teaser.h"

// #include "pcl/registration/icp.h"

#include "xreg/types/cell.h"

#include "xreg/optimization/factors.h"
#include "xreg/optimization/factors_pr.h"
#include "xreg/optimization/se3_param.hpp"

#include "xreg/common/config.h"

#include <omp.h>
#include <teaser/registration.h>

#include <fstream>

namespace xreg {

using namespace std;

RegistrationTeaser::RegistrationTeaser(TeaserOptions options)
    : RegistrationBase(), options_(options) {
  // cout
}

void RegistrationTeaser::optimizePoseGraph() {
  CHECK_EQ(relative_poses_.size(), olp_frames_.size());

  size_t num_frames = frames_.size();

  // initialize frame info
  eigen_aligned_std_vector<Vector3d> trans_params;
  eigen_aligned_std_vector<Vector3d> rot_params;

  ceres::LocalParameterization *parametrization = new ParamSO3();

  ceres::Problem problem;
  for (auto &&frame : frames_) {
    rot_params.push_back(frame->Twb_.so3().log());
    trans_params.push_back(frame->Twb_.translation());
    problem.AddParameterBlock(rot_params.back().data(), 3, parametrization);
    problem.AddParameterBlock(trans_params.back().data(), 3);
  }

  int num_outlier = 0;
  for (size_t i = 0; i < olp_frames_.size(); i++) {
    auto pair = olp_frames_[i];
    auto frame_tgt = pair.first;
    auto frame_src = pair.second;
    auto tgt_idx = frame_tgt->idx_;
    auto src_idx = frame_src->idx_;

    SE3 Tts = relative_poses_[i];

    // cheating
    SE3 Tts_ini = frame_tgt->Twb_.inverse() * frame_src->Twb_;
    SE3 Terr = Tts_ini.inverse() * Tts;
    if (Terr.translation().norm() > 0.6 ||
        Terr.so3().log().norm() > M_PI / 180.0 * 10) {
      num_outlier++;
      continue;
    }

    ceres::CostFunction *rf = new RotationFactor(Tts.so3());
    problem.AddResidualBlock(rf, new ceres::HuberLoss(0.1),
                             rot_params[src_idx].data(),
                             rot_params[tgt_idx].data());

    ceres::CostFunction *tf = new TranslationFactor(Tts);
    problem.AddResidualBlock(
        tf, new ceres::HuberLoss(0.09), rot_params[src_idx].data(),
        trans_params[src_idx].data(), trans_params[tgt_idx].data());
  }
  WARNLOG("num_outlier: {}", num_outlier);

  problem.SetParameterBlockConstant(rot_params[0].data());
  problem.SetParameterBlockConstant(trans_params[0].data());

  ceres::Solver::Options options;
  options.linear_solver_type = ceres::DENSE_QR;
  options.max_num_iterations = 100;
  options.minimizer_progress_to_stdout = false;
  ceres::Solver::Summary summary;
  ceres::Solve(options, &problem, &summary);

  for (size_t i = 0; i < frames_.size(); i++) {
    SO3 roti = SO3::exp(rot_params[i]);
    Vector3d ti = trans_params[i];
    frames_[i]->setTwb(SE3(roti, ti));
  }

  cout << summary.FullReport() << endl;

  // SE3 Twb_est(SO3::exp(rot_param), trans_param);
  // INFOLOG("pose before opt: {}",
  // frame_src->getTwb().translation().transpose()); INFOLOG("pose scan2map:
  // {}", Twb_est.translation().transpose()); frame_src->setTwb(Twb_est);
  // INFOLOG("pose gt: {}", frame_src->Tgt_.translation().transpose());

  // prob

  // ceres::Solver::Options opt_cfg;
  // ceres::Solve(opt_cfg, &problem, )
  // ceres::

  // eigen_aligned_std_vector
}

void RegistrationTeaser::registration() {
  if (tpp::b_pgo) {
    // if (true) {

    for (size_t i = 0; i < frames_.size() - 1; i++) {
      auto frame_tgt = frames_[i];
      auto frame_src = frames_[i + 1];

      // frame_src->
      size_t j = 0;
      for (; j < olp_frames_.size(); j++) {
        auto pair = olp_frames_[j];
        if (pair.first == frame_tgt && pair.second == frame_src) {
          break;
        }
      }

      if (j == olp_frames_.size()) {
        WARNLOG("no overlap between consequential frame: {}.", i);
        throw std::runtime_error("no overlap between consequential frame.");
      }

      const auto &matches = matches_[j];

      SE3 Tts = registerByTeaser(frame_tgt->kpts_, frame_src->kpts_, matches);

      frame_src->setTwb(frame_tgt->getTwb() * Tts);
    }

    // INFOLOG("#olp_frames: {}", olp_frames_.size());
    // getchar();

    for (size_t i = 0; i < olp_frames_.size(); i++) {
      auto pair = olp_frames_[i];
      auto frame_tgt = pair.first;
      auto frame_src = pair.second;
      const auto &matches = matches_[i];

      SE3 Tts = registerByTeaser(frame_tgt->kpts_, frame_src->kpts_, matches);
      relative_poses_.push_back(Tts);
    }

    // STEP. write to g2o
    ofstream ofs("/home/hyhuang/test.g2o");
    ofs << fixed << setprecision(8);
    for (size_t i = 0; i < frames_.size(); i++) {
      auto &&frame = frames_[i];
      auto idx = frame->idx_;
      SE3 Twb = frame->getTwb();

      Quaterniond q = Twb.unit_quaternion();
      Vector3d t = Twb.translation();

      ofs << "VERTEX_SE3:QUAT " << idx << ' ' << t.x() << ' ' << t.y() << ' '
          << t.z() << ' ' << q.x() << ' ' << q.y() << ' ' << q.z() << ' '
          << q.w() << endl;
    }
    for (size_t i = 0; i < olp_frames_.size(); i++) {
      auto pair = olp_frames_[i];
      auto frame_tgt = pair.first;
      auto frame_src = pair.second;
      auto tgt_idx = frame_tgt->idx_;
      auto src_idx = frame_src->idx_;

      SE3 Tts = relative_poses_[i];

      Quaterniond q = Tts.unit_quaternion();
      Vector3d t = Tts.translation();

      ofs << "EDGE_SE3:QUAT " << tgt_idx << ' ' << src_idx << ' ' << ' '
          << t.x() << ' ' << t.y() << ' ' << t.z() << ' ' << q.x() << ' '
          << q.y() << ' ' << q.z() << ' ' << q.w()
          << " 1 0 0 0 0 0 0 1 0 0 0 0 0 0 1 0 0 0 0 0 0 1 0 0 0 0 0 0 1 0 0 0 "
             "0 0 0 1"
          << endl;
    }

    optimizePoseGraph();

  } else {
    for (size_t i = 0; i < frames_.size() - 1; i++) {
      auto frame_tgt = frames_[i];
      auto frame_src = frames_[i + 1];

      // frame_src->
      size_t j = 0;
      for (; j < olp_frames_.size(); j++) {
        auto pair = olp_frames_[j];
        if (pair.first == frame_tgt && pair.second == frame_src) {
          break;
        }
      }

      if (j == olp_frames_.size()) {
        throw std::runtime_error("no overlap between consequential frame.");
      }

      const auto &matches = matches_[j];

      SE3 Tts = registerByTeaser(frame_tgt->kpts_, frame_src->kpts_, matches);

      frame_src->setTwb(frame_tgt->getTwb() * Tts);
    }
  }
}

SE3 RegistrationTeaser::registerByTeaser(
    pcl::PointCloud<pcl::PointXYZ>::Ptr &kps0,
    pcl::PointCloud<pcl::PointXYZ>::Ptr &kps1,
    // pcl::CorrespondencesPtr &corr
    const std::vector<std::pair<int, int>> &corr) {
  auto cloud_src = std::shared_ptr<teaser::PointCloud>(new teaser::PointCloud);
  for (auto &&pt : kps1->points) {
    teaser::PointXYZ ptt;
    ptt.x = pt.x;
    ptt.y = pt.y;
    ptt.z = pt.z;
    cloud_src->push_back(ptt);
  }

  auto cloud_tgt = std::shared_ptr<teaser::PointCloud>(new teaser::PointCloud);
  for (auto &&pt : kps0->points) {
    teaser::PointXYZ ptt;
    ptt.x = pt.x;
    ptt.y = pt.y;
    ptt.z = pt.z;
    cloud_tgt->push_back(ptt);
  }
  // teaser::PointCloud

  omp_set_num_threads(10);
  teaser::RobustRegistrationSolver::Params params;
  params.noise_bound = tpp::noise_bound;
  params.cbar2 = 1;
  params.estimate_scaling = false;
  params.rotation_max_iterations = 100;
  params.rotation_gnc_factor = 1.4;
  params.rotation_estimation_algorithm =
      teaser::RobustRegistrationSolver::ROTATION_ESTIMATION_ALGORITHM::GNC_TLS;
  params.rotation_cost_threshold = 0.005;

  // Solve with TEASER++
  teaser::RobustRegistrationSolver solver(params);
  std::chrono::steady_clock::time_point begin =
      std::chrono::steady_clock::now();

  solver.solve(*cloud_src, *cloud_tgt, corr);

  auto solution = solver.getSolution();
  INFOLOG("estimated: {}", solution.translation.transpose());
  WARNLOG("valid: {}", solution.valid);

  // solution.rvoido
  return xreg::SE3(solution.rotation, solution.translation);
}

}  // namespace xreg