#include "xreg/io/dataloader.h"

#include <fstream>
#include <iostream>

#include <glog/logging.h>

#include <sys/stat.h>

namespace xreg {

// DataloaderKITTI::DataloaderKITTI

using namespace std;

bool isPathExist(const std::string &s) {
  struct stat buffer;
  return (stat(s.c_str(), &buffer) == 0);
}

Dataloader::Dataloader(const std::string &str, DataType cfg)
    : cfg_(cfg), base_path_(str) {}

bool Dataloader::getTrajectory(eigen_aligned_std_vector<Quaterniond> &rot,
                               eigen_aligned_std_vector<Vector3d> &trans) {
  rot = rot_;
  trans = trans_;

  return true;
}

DataloaderEuRoC::DataloaderEuRoC(const std::string &str,
                                 const std::string &traj_file, DataType cfg)
    : Dataloader(str, cfg) {

  LOG(WARNING) << "loading data from: " << base_path_;
  if (!isPathExist(base_path_))
    throw std::runtime_error("base path not exists: " + base_path_);

  if (cfg_ & DataType::GT) {
    if (!isPathExist(traj_file))
      throw std::runtime_error("traj_file not exists: " + traj_file);

    loadTrajectory(traj_file);
    LOG(WARNING) << "load trajectory: " << traj_file;
    num_ = rot_.size();
  }

  if (cfg_ & DataType::Mono) {
    loadImages(base_path_);
    num_ = mono_file_.size();
  }

  if (cfg_ & DataType::IMU) {
    loadIMUData(base_path_ + "/imu0/data.csv");
    // num_ = mono_file_.size();
  }
}

DataFrame::Ptr DataloaderEuRoC::getNextFrame() {
  if (idx_ >= num_)
    return nullptr;

  auto ptr = getFrameByIndex(idx_);

  idx_++;
  return ptr;
}

DataFrame::Ptr DataloaderEuRoC::getFrameByIndex(size_t idx) {
  if (idx >= num_)
    return nullptr;

  auto ptr = DataFrame::Ptr(new DataFrame());

  if (cfg_ & DataType::GT) {
    ptr->rot_w_c = rot_[idx];
    ptr->t_w_c = trans_[idx];
  }

  if (cfg_ & DataType::Mono) {
    if (isPathExist(mono_file_[idx])) {
      ptr->mono = cv::imread(mono_file_[idx]);
    } else {
      ptr->mono = cv::Mat();
    }

    ptr->timestamp = time_stamp_[idx];
  }
  if (cfg_ & DataType::Depth) {
    if (isPathExist(depth_file_[idx])) {
      ptr->depth = cv::imread(depth_file_[idx]);
    } else {
      ptr->depth = cv::Mat();
    }
  }
  //   if (cfg_ & DataType::IMU && idx != 0) {
  //     double last_stamp = time_stamp_[idx - 1];
  //     double curr_stamp = time_stamp_[idx];

  //     size_t imu_iter = 0;
  //     while (imu_iter < imu_stamps.size() &&
  //            imu_stamps[imu_iter++] < last_stamp) {
  //     }

  //     if (imu_iter != imu_stamps.size()) {
  //       while (imu_iter < imu_stamps.size() &&
  //              imu_stamps[imu_iter] <= curr_stamp) {

  //         ptr->imu_stamps.push_back(imu_stamps[imu_iter]);
  //         ptr->acc.push_back(acc[imu_iter]);
  //         ptr->gyro.push_back(gyro[imu_iter]);

  //         imu_iter++;
  //       }
  //       // imu_iter
  //     }
  //   }

  //   ptr->sub_str = sub_str_[idx];

  ptr->idx = idx;
  return ptr;
}

void DataloaderEuRoC::loadImages(const string &base_path) {
  ifstream fTimes;
  string strPathTimeFile = base_path + "/cam0/data.csv";
  string strPrefixLeft = base_path + "/cam0/data/";
  string strPrefixRight = base_path + "/cam1/data/";

  fTimes.open(strPathTimeFile.c_str());
  string s;
  getline(fTimes, s);
  while (!fTimes.eof()) {
    string s;
    getline(fTimes, s);
    if (!s.empty()) {
      int index = s.find_first_of(",");
      string t = s.substr(0, index);

      time_stamp_.push_back(stod(t) / 10.0e8);
      mono_file_.push_back(strPrefixLeft + t + ".png");
      depth_file_.push_back(strPrefixRight + t + ".png");

      sub_str_.push_back(t);
    }
  }
}

void DataloaderEuRoC::loadTrajectory(const std::string &traj_file) {
  ifstream fs_mean(traj_file.c_str());
  string str_line;
  while (getline(fs_mean, str_line) && !fs_mean.eof()) {
    stringstream ss(str_line);
    double time, x, y, z, qx, qy, qz, qw;
    ss >> time;
    ss >> x;
    ss >> y;
    ss >> z;
    ss >> qx;
    ss >> qy;
    ss >> qz;
    ss >> qw;

    Eigen::Vector3d trans_tmp(x, y, z);
    Eigen::Quaterniond q_tmp(qw, qx, qy, qz);
    trans_.push_back(trans_tmp);
    rot_.push_back(q_tmp);
  }
}

void DataloaderEuRoC::loadIMUData(const std::string &imu_file) {
  ifstream ifs(imu_file.c_str());
  string str_line;

  while (getline(ifs, str_line) && !ifs.eof()) {
    if (str_line[0] == '#') {
      continue;
    }

    if (!str_line.empty()) {
      string item;
      size_t pos = 0;
      double data[7];
      int count = 0;
      while ((pos = str_line.find(',')) != string::npos) {
        item = str_line.substr(0, pos);
        data[count++] = stod(item);
        str_line.erase(0, pos + 1);
      }
      item = str_line.substr(0, pos);
      data[6] = stod(item);

      imu_stamps.push_back(data[0] / 1e9);
      acc.emplace_back(data[4], data[5], data[6]);
      gyro.emplace_back(data[1], data[2], data[3]);
      // gyro.push_back(cv::Point3f(data[1], data[2], data[3]));
    }
  }
} // namespace geoloc

void DataloaderKITTI::loadTimeStamps(const std::string &stamp_file) {}

DataloaderKITTI::DataloaderKITTI(const std::string &str, const std::string &seq,
                                 DataType cfg)
    : Dataloader(str, cfg) {

  LOG(WARNING) << "loading data from: " << base_path_;
  if (!isPathExist(base_path_))
    throw std::runtime_error("base path not exists: " + base_path_);

  ifstream fTimes;
  string strPathTimeFile = str + "/color/sequences/" + seq + "/times.txt";
  fTimes.open(strPathTimeFile.c_str());
  while (!fTimes.eof()) {
    string s;
    getline(fTimes, s);
    if (!s.empty()) {
      stringstream ss;
      ss << s;
      double t;
      ss >> t;
      time_stamp_.push_back(t);
    }
  }

  string strPrefixLeft = str + "/gray/sequences/" + seq + "/image_0/";
  string strPrefixRight = str + "/gray/sequences/" + seq + "/image_1/";
  string prefix_laser = str + "/velodyne/sequences/" + seq + "/velodyne/";

  num_ = time_stamp_.size();
  mono_file_.resize(num_);
  depth_file_.resize(num_);
  laser_file_.resize(num_);

  for (int i = 0; i < num_; i++) {
    stringstream ss;
    ss << setfill('0') << setw(6) << i;
    mono_file_[i] = strPrefixLeft + ss.str() + ".png";
    depth_file_[i] = strPrefixRight + ss.str() + ".png";
    laser_file_[i] = prefix_laser + ss.str() + ".bin";
  }

  if (cfg_ & DataType::GT) {
    // if (!isPathExist(traj_file))
    //   throw std::runtime_error("traj_file not exists: " + traj_file);

    // loadTrajectory(traj_file);
    // LOG(WARNING) << "load trajectory: " << traj_file;
    // num_ = rot_.size();
  }

  if (cfg_ & DataType::Mono) {
    // loadImages(base_path_);
    // num_ = mono_file_.size();
  }

  //   if (cfg_ & DataType::IMU) {
  //     loadIMUData(base_path_ + "/imu0/data.csv");
  //     // num_ = mono_file_.size();
  //   }
  if (cfg_ & DataType::Lidar) {
    // loadLaserData(base_path_ + "/velodyne/sequences/" + seq + '/');
    // num_ = mono_file_.size();
  }
}

DataFrame::Ptr DataloaderKITTI::getNextFrame() {
  if (idx_ >= num_)
    return nullptr;

  auto ptr = getFrameByIndex(idx_);

  idx_++;
  return ptr;
}

DataFrame::Ptr DataloaderKITTI::getFrameByIndex(size_t idx) {
  if (idx >= num_)
    return nullptr;

  auto ptr = DataFrame::Ptr(new DataFrame());

  if (cfg_ & DataType::GT) {
    // ptr->rot_w_c = rot_[idx];
    // ptr->t_w_c = trans_[idx];
  }

  if (cfg_ & DataType::Mono) {
    // if (isPathExist(mono_file_[idx])) {
    //   ptr->mono = cv::imread(mono_file_[idx]);
    // } else {
    //   ptr->mono = cv::Mat();
    // }

    // ptr->timestamp = time_stamp_[idx];
  }
  if (cfg_ & DataType::Depth) {
    // if (isPathExist(depth_file_[idx])) {
    //   ptr->depth = cv::imread(depth_file_[idx]);
    // } else {
    //   ptr->depth = cv::Mat();
    // }
  }

  if (cfg_ & DataType::Lidar) {
    // cout << laser_file_[idx] << endl;
    auto cloud = loadLaserBinFile(laser_file_[idx]);
    if (!cloud) {
      ERRLOG("empty pointcloud input");
    }

    ptr->cloud = cloud;
  }

  //   ptr->sub_str = sub_str_[idx];

  ptr->idx = idx;
  return ptr;
}

// void
PointCloudXYZI::Ptr
DataloaderKITTI::loadLaserBinFile(const std::string &infile) {
  // load point cloud
  fstream input(infile.c_str(), ios::in | ios::binary);
  if (!input.good()) {
    cerr << "Could not read file: " << infile << endl;
    return nullptr;
  }
  input.seekg(0, ios::beg);

  PointCloudXYZI::Ptr points(new PointCloudXYZI);

  int i;
  for (i = 0; input.good() && !input.eof(); i++) {
    pcl::PointXYZI point;
    input.read((char *)&point.x, 3 * sizeof(float));
    input.read((char *)&point.intensity, sizeof(float));
    points->push_back(point);
  }
  input.close();

  return points;
}

// void DataloaderKITTI::loadImages(const string &base_path) {}

void DataloaderKITTI::loadImages(const string &base_path) {
  //   ifstream fTimes;
  //   string strPathTimeFile = base_path + "/cam0/data.csv";
  //   string strPrefixLeft = base_path + "/cam0/data/";
  //   string strPrefixRight = base_path + "/cam1/data/";

  //   fTimes.open(strPathTimeFile.c_str());
  //   string s;
  //   getline(fTimes, s);
  //   while (!fTimes.eof()) {
  //     string s;
  //     getline(fTimes, s);
  //     if (!s.empty()) {
  //       int index = s.find_first_of(",");
  //       string t = s.substr(0, index);

  //       time_stamp_.push_back(stod(t) / 10.0e8);
  //       mono_file_.push_back(strPrefixLeft + t + ".png");
  //       depth_file_.push_back(strPrefixRight + t + ".png");

  //       sub_str_.push_back(t);
  //     }
  //   }
}

void DataloaderKITTI::loadTrajectory(const std::string &traj_file) {
  //   ifstream fs_mean(traj_file.c_str());
  //   string str_line;
  //   while (getline(fs_mean, str_line) && !fs_mean.eof()) {
  //     stringstream ss(str_line);
  //     double time, x, y, z, qx, qy, qz, qw;
  //     ss >> time;
  //     ss >> x;
  //     ss >> y;
  //     ss >> z;
  //     ss >> qx;
  //     ss >> qy;
  //     ss >> qz;
  //     ss >> qw;

  //     Eigen::Vector3d trans_tmp(x, y, z);
  //     Eigen::Quaterniond q_tmp(qw, qx, qy, qz);
  //     trans_.push_back(trans_tmp);
  //     rot_.push_back(q_tmp);
  //   }
}

} // namespace xreg
