#include "xreg/odometry/odometry.h"

// #include <ceres/ceres.h>

#include <pcl/common/distances.h>

#include "xreg/optimization/factors_loam.h"
#include "xreg/optimization/se3_param.hpp"

#include "xreg/utils/math_utils.h"
#include "xreg/utils/timing.h"

namespace xreg {
using namespace std;

Odometry::Odometry(const Odometry::Param &param) : param_(param) {
  kdtree_corner_ = pcl::KdTreeFLANN<pcl::PointXYZI>::Ptr(
      new pcl::KdTreeFLANN<pcl::PointXYZI>);
  kdtree_surf_ = pcl::KdTreeFLANN<pcl::PointXYZI>::Ptr(
      new pcl::KdTreeFLANN<pcl::PointXYZI>);

  param_rot_last_curr_.setZero();
  param_t_last_curr_.setZero();
}

Odometry::Result Odometry::estimate(PointCloudXYZI::Ptr feats_flat,
                                    PointCloudXYZI::Ptr feats_less_flat,
                                    PointCloudXYZI::Ptr feats_sharp,
                                    PointCloudXYZI::Ptr feats_less_sharp) {
  Odometry::Result res;

  if (!init_) {
    init_ = true;

  } else {

    timing::Timer timer_total("odom/total");
    int num_match_corner = 0, num_match_surf = 0;
    Vector3d param_rot_last_curr, param_t_last_curr;

    for (size_t opti_counter = 0; opti_counter < 2; ++opti_counter) {
      num_match_corner = 0;
      num_match_surf = 0;

      // ceres::LossFunction *loss_function = NULL;
      ceres::LossFunction *loss_function = new ceres::HuberLoss(0.1);

      ceres::Problem::Options problem_options;

      ceres::Problem problem(problem_options);

      // add residual blocks
      ceres::LocalParameterization *rot_param = new ParamSO3();
      problem.AddParameterBlock(param_rot_last_curr.data(), 3, rot_param);
      problem.AddParameterBlock(param_t_last_curr.data(), 3);

      // add factors
      num_match_corner = addEdgeFactors(feats_sharp, problem);
      num_match_surf = addSurfFactors(feats_flat, problem);
      INFOLOG("#match_surf: {}, #match_edge: {}", num_match_surf,
              num_match_corner);

      if ((num_match_corner + num_match_surf) < 10) {
        ERRLOG("########### few matches...");
      }

      ceres::Solver::Options options;
      options.linear_solver_type = ceres::DENSE_QR;
      options.max_num_iterations = 4;
      options.minimizer_progress_to_stdout = false;
      ceres::Solver::Summary summary;
      ceres::Solve(options, &problem, &summary);
      // printf("solver time %f ms \n", t_solver.toc());

      // cout << summary.FullReport() << endl;
      Tlc_ = SE3(SO3::exp(param_rot_last_curr_), param_t_last_curr_);
    }
    //   printf("optimization twice time %f \n", t_opt.toc());
    timer_total.Stop();
  }
  cout << Tlc_.translation().transpose() << endl;

  feats_surf_last_ = feats_less_flat;
  feats_corner_last_ = feats_less_sharp;

  kdtree_surf_->setInputCloud(feats_surf_last_);
  kdtree_corner_->setInputCloud(feats_corner_last_);

  // Tlc_ = SE3(SO3::exp(param_rot_last_curr_), param_t_last_curr_);
  // cout << "b" << endl;
  res.Tlc = Tlc_;
  // cout << "b" << endl;

  return res;
}

int Odometry::addEdgeFactors(PointCloudXYZI::ConstPtr feats,
                             ceres::Problem &problem) {

  int num_corner_sharp = feats->size();
  int num_match = 0;

  pcl::PointXYZI point_sel;
  std::vector<int> point_search_indices;
  std::vector<float> point_search_dist;

  // find correspondence for corner features
  for (int i = 0; i < num_corner_sharp; ++i) {
    transformPoint(&(feats->points[i]), &point_sel);

    kdtree_corner_->nearestKSearch(point_sel, 1, point_search_indices,
                                   point_search_dist);

    int idx0 = -1, idx1 = -1;
    if (point_search_dist[0] < param_.th_dist_sqr) {
      idx0 = point_search_indices[0];
      int scan_id = int(feats_corner_last_->points[idx0].intensity);

      double min_dist2 = param_.th_dist_sqr;
      // search in the direction of increasing scan line
      for (int j = idx0 + 1; j < (int)feats_corner_last_->points.size(); ++j) {
        // if in the same scan line, continue
        if (int(feats_corner_last_->points[j].intensity) <= scan_id)
          continue;

        // if not in nearby scans, end the loop
        if (int(feats_corner_last_->points[j].intensity) >
            (scan_id + param_.num_scan_nearby))
          break;

        float dist2 = pcl::squaredEuclideanDistance(
            feats_corner_last_->points[j], point_sel);

        if (dist2 < min_dist2) {
          // find nearer point
          min_dist2 = dist2;
          idx1 = j;
        }
      }

      // search in the direction of decreasing scan line
      for (int j = idx0 - 1; j >= 0; --j) {
        // if in the same scan line, continue
        if (int(feats_corner_last_->points[j].intensity) >= scan_id)
          continue;

        // if not in nearby scans, end the loop
        if (int(feats_corner_last_->points[j].intensity) <
            (scan_id - param_.num_scan_nearby))
          break;

        float dist2 = pcl::squaredEuclideanDistance(
            feats_corner_last_->points[j], point_sel);

        if (dist2 < min_dist2) {
          // find nearer point
          min_dist2 = dist2;
          idx1 = j;
        }
      }
    }
    if (idx1 >= 0) // both idx0 and idx1 is valid
    {
      Vector3d curr_point = MathUtils::pcl2Eigen(feats->points[i]);
      Vector3d pta = MathUtils::pcl2Eigen(feats_corner_last_->points[idx0]);
      Vector3d ptb = MathUtils::pcl2Eigen(feats_corner_last_->points[idx1]);

      // TODO:
      // double s;
      // if (DISTORTION)
      //   s = (feats_sharp->points[i].intensity -
      //        int(feats_sharp->points[i].intensity)) /
      //       SCAN_PERIOD;
      // else
      //   s = 1.0;
      //   ceres::CostFunction *cost_function =
      //       LidarEdgeFactor::Create(curr_point, last_point_a, last_point_b,
      //       s);

      ceres::CostFunction *factor = new EdgeFactor(curr_point, pta, ptb);
      problem.AddResidualBlock(factor, new ceres::HuberLoss(0.1),
                               param_rot_last_curr_.data(),
                               param_t_last_curr_.data());
      num_match++;
    }
  }

  return num_match;
}

int Odometry::addSurfFactors(PointCloudXYZI::ConstPtr feats,
                             ceres::Problem &problem) {
  int num_surf_flat = feats->size();
  int num_match = 0;
  pcl::PointXYZI point_sel;
  std::vector<int> point_search_indices;
  std::vector<float> point_search_dist;

  INFOLOG("#surf_feats: {}", num_surf_flat);
  // find correspondence for plane features
  for (int i = 0; i < num_surf_flat; ++i) {
    // FIXME:
    // TransformToStart(&(surfPointsFlat->points[i]), &point_sel);
    transformPoint(&(feats->points[i]), &point_sel);

    int num_searched = kdtree_surf_->nearestKSearch(
        point_sel, 1, point_search_indices, point_search_dist);
    // cout << num_searched << endl;
    // cout << point_search_dist[0] << ' ' << param_.th_dist_sqr << endl;

    int idx0 = -1, idx1 = -1, idx2 = -1;
    if (point_search_dist[0] < param_.th_dist_sqr) {
      idx0 = point_search_indices[0];

      // get closest point's scan ID
      int scan_id = int(feats_surf_last_->points[idx0].intensity);
      double min_dist2 = param_.th_dist_sqr, min_dist3 = param_.th_dist_sqr;

      // search in the direction of increasing scan line
      for (int j = idx0 + 1; j < (int)feats_surf_last_->points.size(); ++j) {
        // if not in nearby scans, end the loop
        if (int(feats_surf_last_->points[j].intensity) >
            (scan_id + param_.num_scan_nearby))
          break;

        float dist2 = pcl::squaredEuclideanDistance(feats_surf_last_->points[j],
                                                    point_sel);

        // if in the same or lower scan line
        if (int(feats_surf_last_->points[j].intensity) <= scan_id &&
            dist2 < min_dist2) {
          min_dist2 = dist2;
          idx1 = j;
        }

        // if in the higher scan line
        else if (int(feats_surf_last_->points[j].intensity) > scan_id &&
                 dist2 < min_dist3) {
          min_dist3 = dist2;
          idx2 = j;
        }
      }

      // search in the direction of decreasing scan line
      for (int j = idx0 - 1; j >= 0; --j) {
        // if not in nearby scans, end the loop
        if (int(feats_surf_last_->points[j].intensity) <
            (scan_id - param_.num_scan_nearby))
          break;

        float dist2 = pcl::squaredEuclideanDistance(feats_surf_last_->points[j],
                                                    point_sel);

        // if in the same or higher scan line
        if (int(feats_surf_last_->points[j].intensity) >= scan_id &&
            dist2 < min_dist2) {
          min_dist2 = dist2;
          idx1 = j;
        } else if (int(feats_surf_last_->points[j].intensity) < scan_id &&
                   dist2 < min_dist3) {
          // find nearer point
          min_dist3 = dist2;
          idx2 = j;
        }
      }

      if (idx1 >= 0 && idx2 >= 0) {
        Vector3d pt = MathUtils::pcl2Eigen(feats->points[i]);
        Vector3d pta = MathUtils::pcl2Eigen(feats_surf_last_->points[idx0]);
        Vector3d ptb = MathUtils::pcl2Eigen(feats_surf_last_->points[idx1]);
        Vector3d ptc = MathUtils::pcl2Eigen(feats_surf_last_->points[idx2]);

        // FIXME: undistortion
        //   double s;
        //   if (DISTORTION)
        //     s = (surfPointsFlat->points[i].intensity -
        //          int(surfPointsFlat->points[i].intensity)) /
        //         SCAN_PERIOD;
        //   else
        //     s = 1.0;

        // TODO: undistortion
        ceres::CostFunction *cost_function =
            new SurfPointFactor(pt, pta, ptb, ptc);

        problem.AddResidualBlock(cost_function, new ceres::HuberLoss(0.1),
                                 param_rot_last_curr_.data(),
                                 param_t_last_curr_.data());

        num_match++;
      }
    }
  }

  return num_match;
}

// TODO(undistortion)
void Odometry::transformPoint(PointXYZI const *const pi, PointXYZI *const po) {
  //   Eigen::Vector3d t_point_last = s * t_last_curr;
  Eigen::Vector3d point(pi->x, pi->y, pi->z);
  Eigen::Vector3d un_point = Tlc_ * point;

  po->x = un_point.x();
  po->y = un_point.y();
  po->z = un_point.z();
  po->intensity = pi->intensity;
}

} // namespace xreg
