#include "xreg/optimization/factors.h"

#include "xreg/common/common.h"

#include "xreg/utils/math_utils.h"

namespace xreg {

using namespace std;

EigenFactorTrans::EigenFactorTrans(const SE3 &tgt, const SE3 &src,
                                   const double s)
    : scalar_(s), T0_(tgt), T1_(src),
      ez0_skew_(MathUtils::skew(tgt.rotationMatrix().col(0)).topRows(2)) {}

void EigenFactorTrans::setTargetPose(const SE3 &tgt) {
  T0_ = tgt;
  ez0_skew_ = MathUtils::skew(tgt.rotationMatrix().col(0)).topRows(2);
}

EigenFactorRot::EigenFactorRot(const SE3 &tgt, const SE3 &src, const double s)
    : scalar_(s), T0_(tgt), T1_(src),
      ez0_skew_(MathUtils::skew(tgt.rotationMatrix().col(0)).topRows(2)) {}

void EigenFactorRot::setTargetPose(const SE3 &tgt) {
  T0_ = tgt;
  ez0_skew_ = MathUtils::skew(tgt.rotationMatrix().col(0)).topRows(2);
}

bool EigenFactorRot::Evaluate(double const *const *parameters,
                              double *residuals, double **jacobians) const {

  Eigen::Map<Vector2d> res(residuals);
  Eigen::Map<const Vector3d> rot_w_c_log(parameters[0]);
  // Eigen::Map<const Vector3d> t_w_c(parameters[1]);

  const auto &rot_l = SO3::exp(rot_w_c_log);
  // const auto &t_l = t_w_c;

  const Vector3d &ez0 = T0_.rotationMatrix().col(0);
  const Vector3d &ezk = T1_.rotationMatrix().col(0);
  const Vector3d &pk_bar = T1_.translation();

  if (residuals) {
    res = ez0_skew_ * (rot_l * ezk);
  }

  if (jacobians) {

    if (jacobians[0]) {
      // cout << "f" << endl;
      Eigen::Map<Eigen::Matrix<double, 2, 3, Eigen::RowMajor>> jacobian_pose(
          jacobians[0]);
      jacobian_pose.setZero();

      // rotation
      jacobian_pose = -ez0_skew_ * rot_l.matrix() * MathUtils::skew(ezk);

      jacobian_pose *= scalar_;
    }
  }

  return true;
}

bool EigenFactorTrans::Evaluate(double const *const *parameters,
                                double *residuals, double **jacobians) const {

  // Eigen::Map<Vector3d> res(residuals);
  Eigen::Map<const Vector3d> rot_w_c_log(parameters[0]);
  Eigen::Map<const Vector3d> t_w_c(parameters[1]);

  const auto &rot_l = SO3::exp(rot_w_c_log);
  const auto &t_l = t_w_c;

  const Vector3d &ez0 = T0_.rotationMatrix().col(0);
  const Vector3d &ezk = T1_.rotationMatrix().col(0);
  const Vector3d &pk_bar = T1_.translation();

  if (residuals) {
    Vector3d dpk0 = (rot_l * pk_bar + t_l) - T0_.translation();
    residuals[0] = scalar_ * ez0.dot(dpk0);
  }

  if (jacobians) {

    if (jacobians[0]) {
      Eigen::Map<Eigen::RowVector3d> jacobian_pose(jacobians[0]);
      jacobian_pose.setZero();

      // rotation
      // jacobian_pose.topLeftCorner(2, 3) =
      //     -ez0_skew_ * rot_l.matrix() * skew(ezk);
      // jacobian_pose.bottomLeftCorner(1, 3) =
      //     -ez0.transpose() * rot_l.matrix() * skew(pk_bar);
      jacobian_pose =
          -ez0.transpose() * rot_l.matrix() * MathUtils::skew(pk_bar);

      jacobian_pose *= scalar_;

      // jacobian_pose.bottomRightCorner(1, 3) = ez0.transpose();
      // INFOLOG("jrot:\n{}", jacobian_pose);
    }

    if (jacobians[1]) {
      // cout << "g" << endl;
      Eigen::Map<Eigen::RowVector3d> jacobian_pose(jacobians[1]);
      jacobian_pose = ez0.transpose();

      jacobian_pose *= scalar_;
      // cout << jacobian_pose << endl;
      // INFOLOG("jtrans:\n{}", jacobian_pose);
      // jacobian_pose.bottomRightCorner(1, 3) = ez0.transpose();
      // cout << "h" << endl;
    }
  }

  return true;
}

void SurfRegEvalCallBackDecoupled::PrepareForEvaluation(
    bool evaluate_jacobians, bool new_evaluation_point) {
  // prepare for evaluation
  // INFOLOG("callback for eval, eval_jacob: {}, eva_point: {}...",
  //         evaluate_jacobians, new_evaluation_point);

  if (new_evaluation_point) {
    for (size_t i = 0; i < frames_.size(); i++) {
      auto &&frame = frames_[i];
      double *pr = param_rot_[i];
      double *pt = param_trans_[i];

      Eigen::Map<const Vector3d> rot_log(pr);
      Eigen::Map<const Vector3d> t(pt);
      frame->setTwb(SE3(SO3::exp(rot_log), t));
    }

    // for (auto &&info : record_)
    for (size_t i = 0; i < cells_.size(); i++) {
      // auto &&plane = info.first;
      auto &&plane = cells_[i];
      plane->solveXX();
      SE3 tgt_pose(plane->rot_, plane->mean_);

      {
        for (auto &&f : rot_factors_[i]) {
          f->setTargetPose(tgt_pose);
        }
        for (auto &&f : trans_factors_[i]) {
          f->setTargetPose(tgt_pose);
        }
      }
    }
  }
  // INFOLOG("callback exit..");
}

void SurfRegEvalCallBackDecoupled::addPlane(
    Cell *plane, const std::vector<EigenFactorRot *> &rfs,
    const std::vector<EigenFactorTrans *> &tfs) {

  // #
  // addPlane();
  // record_[plane] = factor;
  cells_.push_back(plane);
  rot_factors_.push_back(rfs);
  trans_factors_.push_back(tfs);
}

void SurfRegEvalCallBackDecoupled::addFrame(Frame *f, double *rot,
                                            double *trans) {
  // if (x.count(f)) {
  //   return;
  // }

  // x[f] = param;
  frames_.push_back(f);
  param_rot_.push_back(rot);
  param_trans_.push_back(trans);
}

} // namespace xreg
