#include "xreg/optimization/factors_ef.h"

namespace xreg {

EigenFactorSQR::EigenFactorSQR(const Matrix4d &LT, const Cell *cell)
    : l_(LT), cell_(cell) {}

bool EigenFactorSQR::Evaluate(double const *const *parameters,
                              double *residuals, double **jacobians) const {

  Eigen::Map<Vector4d> res(residuals);
  Eigen::Map<const Vector6d> pose_log(parameters[0]);

  const auto &pose_w_c = SE3::exp(pose_log);

  const Matrix4d QT = (pose_w_c.matrix() * l_).transpose();

  const Vector4d pi_ = cell_->pi_;

  if (residuals) {
    res = (pi_.transpose() * pose_w_c.matrix() * l_).transpose();
  }

  if (jacobians) {

    if (jacobians[0]) {
      Eigen::Map<Eigen::Matrix<double, 4, 6, Eigen::RowMajor>> jacobian_pose(
          jacobians[0]);

      jacobian_pose.col(0) = pi_.x() * QT.col(3);
      jacobian_pose.col(1) = pi_.y() * QT.col(3);
      jacobian_pose.col(2) = pi_.z() * QT.col(3);

      jacobian_pose.col(3) = pi_(2) * QT.col(1) - pi_(1) * QT.col(2);
      jacobian_pose.col(4) = pi_(0) * QT.col(2) - pi_(2) * QT.col(0);
      jacobian_pose.col(5) = pi_(1) * QT.col(0) - pi_(0) * QT.col(1);
    }
  }

  return true;
}

void EigenFactorSQREvalCallBack::PrepareForEvaluation(
    bool evaluate_jacobians, bool new_evaluation_point) {
  // prepare for evaluation
  // INFOLOG("callback for eval, eval_jacob: {}, eva_point: {}...",
  //         evaluate_jacobians, new_evaluation_point);

  if (new_evaluation_point) {
    for (size_t i = 0; i < frames_.size(); i++) {
      auto &&frame = frames_[i];
      double *param = params_[i];

      Eigen::Map<const Vector6d> se3_log(param);
      frame->setTwb(SE3::exp(se3_log));
    }

    for (size_t i = 0; i < cells_.size(); i++) {
      auto &&plane = cells_[i];
      plane->updateInfoAllEigenFactor(false);
      // factors_[i]->setPlaneParam(plane->pi_);
    }
    // INFOLOG("callback exit..");
  }
}

void EigenFactorSQREvalCallBack::addPlane(Cell *plane) {
  cells_.push_back(plane);
  // factors_.push_back(factor);
}

void EigenFactorSQREvalCallBack::addFrame(Frame *f, double *param) {
  frames_.push_back(f);
  params_.push_back(param);
  //   param_trans_.push_back(trans);
}

// void EigenFactorSQREvalCallBack::PrepareForEvaluation(
//     bool evaluate_jacobians, bool new_evaluation_point) {
//   // prepare for evaluation
//   // INFOLOG("callback for eval, eval_jacob: {}, eva_point: {}...",
//   //         evaluate_jacobians, new_evaluation_point);

//   if (new_evaluation_point) {
//     for (auto &&info : x) {
//       Vector6d vec = Eigen::Map<Vector6d>(info.second);
//       info.first->setTwb(SE3::exp(vec));
//     }

//     for (auto &&info : record_) {
//       auto &&plane = info.first;
//       plane->solveEF();
//       // SE3 tgt_pose(plane->rot_, plane->mean_);

//       auto &factors = info.second;
//       for (auto &&f : factors) {
//         f->setPlaneParam(plane->pi_);
//       }
//     }
//   }
//   // INFOLOG("callback exit..");
// }

// void EigenFactorSQREvalCallBack::addPlane(
//     Cell *plane, const std::vector<EigenFactorSQR *> &factor) {
//   record_[plane] = factor;
// }

// void EigenFactorSQREvalCallBack::addFrame(Frame *f, double *param) {
//   if (x.count(f)) {
//     return;
//   }

//   x[f] = param;
// }

} // namespace xreg