#include "xreg/optimization/factors_loam.h"

#include "xreg/common/common.h"

#include "xreg/utils/math_utils.h"

namespace xreg {
EdgeFactor::EdgeFactor(Eigen::Vector3d curr_point, Eigen::Vector3d last_point_a,
                       Eigen::Vector3d last_point_b)
    : curr_point_(curr_point), last_point_a_(last_point_a),
      last_point_b_(last_point_b),
      de_norm_((last_point_a_ - last_point_b_).norm()) {}

bool EdgeFactor::Evaluate(double const *const *params, double *residuals,
                          double **jacobians) const {
  Eigen::Map<const Vector3d> rot_log(params[0]);
  Eigen::Map<const Vector3d> t(params[1]);

  const SO3 rot = SO3::exp(rot_log);

  Eigen::Vector3d lp;
  lp = rot * curr_point_ + t; // new point
  Eigen::Vector3d nu = (lp - last_point_a_).cross(lp - last_point_b_);
  // Eigen::Vector3d de = last_point_a_ - last_point_b_;

  if (residuals) {
    Eigen::Map<Vector3d> res(residuals);
    res = nu / de_norm_;
  }

  if (jacobians) {
    Eigen::Vector3d re = last_point_b_ - last_point_a_;
    Eigen::Matrix3d skew_re = MathUtils::skew(re / de_norm_);
    if (jacobians[0]) {
      // Eigen::Matrix3d skew_lp = skew(lp);
      // Eigen::Matrix<double, 3, 6> dp_by_so3;
      // dp_by_so3.block<3, 3>(0, 0) = -skew_lp;
      // (dp_by_so3.block<3, 3>(0, 3)).setIdentity();
      // Eigen::Map<Eigen::Matrix<double, 3, 7, Eigen::RowMajor>> J_se3(
      //     jacobians[0]);
      // J_se3.setZero();

      // J_se3.block<3, 6>(0, 0) = skew_re * dp_by_so3 / de.norm();
      Eigen::Map<Eigen::Matrix<double, 3, 3, Eigen::RowMajor>> jacob_rot(
          jacobians[0]);

      jacob_rot = -skew_re * rot.matrix() * MathUtils::skew(curr_point_);
    }
    if (jacobians[1]) {
      Eigen::Map<Eigen::Matrix<double, 3, 3, Eigen::RowMajor>> jacob_trans(
          jacobians[1]);

      jacob_trans = skew_re;
    }
  }

  return true;
}

SurfPointFactor::SurfPointFactor(Eigen::Vector3d curr_point,
                                 Eigen::Vector3d pt0, Eigen::Vector3d pt1,
                                 Eigen::Vector3d pt2)
    : pt_(curr_point), pt0_(pt0), pt1_(pt1), pt2_(pt2) {
  norm_ = (pt0_ - pt1_).cross(pt0_ - pt2_);
  norm_.normalize();
}

bool SurfPointFactor::Evaluate(double const *const *params, double *residuals,
                               double **jacobians) const {

  Eigen::Map<const Vector3d> rot_log(params[0]);
  Eigen::Map<const Vector3d> t(params[1]);

  SO3 rot = SO3::exp(rot_log);

  if (residuals) {
    Vector3d lp = rot * pt_ + t;
    residuals[0] = norm_.dot(lp - pt0_);
  }

  if (jacobians) {

    if (jacobians[0]) {
      Eigen::Map<RowVector3d> jacob_rot(jacobians[0]);
      // jacob_rot
      jacob_rot = -norm_.transpose() * rot.matrix() * MathUtils::skew(pt_);
    }

    if (jacobians[1]) {
      Eigen::Map<RowVector3d> jacob_trans(jacobians[1]);
      jacob_trans = norm_.transpose();
    }
  }

  return true;
}

SurfNormFactorLOAM::SurfNormFactorLOAM(const Eigen::Vector3d &pt,
                                       const Eigen::Vector3d &normal,
                                       const double &d)
    : pt_(pt), n_(normal), d_(d) {}

bool SurfNormFactorLOAM::Evaluate(double const *const *params,
                                  double *residuals, double **jacobians) const {

  Eigen::Map<const Vector3d> rot_log(params[0]);
  Eigen::Map<const Vector3d> t(params[1]);

  SO3 rot = SO3::exp(rot_log);

  Eigen::Vector3d ptw = rot * pt_ + t;

  if (residuals) {
    residuals[0] = n_.dot(ptw) + d_;
  }

  if (jacobians) {
    if (jacobians[0]) {
      Eigen::Map<Eigen::RowVector3d> jacob_rot(jacobians[0]);
      jacob_rot = -n_.transpose() * (rot.matrix() * MathUtils::skew(pt_));
    }
    if (jacobians[1]) {
      Eigen::Map<Eigen::RowVector3d> jacob_trans(jacobians[1]);
      jacob_trans = n_.transpose();
    }
  }

  return true;
}

// surf norm cost
// SurfNormFactor::SurfNormFactor(Eigen::Vector3d curr_point,
//                                Eigen::Vector3d plane_unit_norm,
//                                double negative_OA_dot_norm)
//     : curr_point_(curr_point), plane_unit_norm_(plane_unit_norm),
//       negative_OA_dot_norm_(negative_OA_dot_norm) {}
SurfNormFactor::SurfNormFactor(const Eigen::Vector3d &pt,
                               const Eigen::Vector3d &normal,
                               const Eigen::Vector3d &mean)
    : pt_(pt), m_(mean), n_(normal) {}

bool SurfNormFactor::Evaluate(double const *const *params, double *residuals,
                              double **jacobians) const {

  Eigen::Map<const Vector3d> rot_log(params[0]);
  Eigen::Map<const Vector3d> t(params[1]);

  SO3 rot = SO3::exp(rot_log);

  Eigen::Vector3d ptw = rot * pt_ + t;

  if (residuals) {
    residuals[0] = n_.dot(ptw - m_);
  }

  if (jacobians) {
    if (jacobians[0]) {
      Eigen::Map<Eigen::RowVector3d> jacob_rot(jacobians[0]);
      jacob_rot = -n_.transpose() * (rot.matrix() * MathUtils::skew(pt_));
    }
    if (jacobians[1]) {
      Eigen::Map<Eigen::RowVector3d> jacob_trans(jacobians[1]);
      jacob_trans = n_.transpose();
    }
  }

  return true;
}
} // namespace xreg