#include "xreg/system.h"

#include <chrono>
#include <fstream>

#include <pcl/common/transforms.h>
#include <pcl/filters/voxel_grid.h>
#include <pcl/io/pcd_io.h>

#include "xreg/estimation/registration_ba.h"
#include "xreg/estimation/registration_ba_ef.h"
#include "xreg/estimation/registration_base.h"
#include "xreg/estimation/registration_fpfh.h"
// #include "xreg/estimation/registration_gicp.h"
#include "xreg/estimation/registration_eigenfactor.h"
#include "xreg/estimation/registration_icp_pt2pl.h"
// #include "xreg/estimation/registration_icp_pt2pt.h"
#include "xreg/estimation/registration_balm.h"
#include "xreg/estimation/registration_fgr.h"
#include "xreg/estimation/registration_ndt.h"
#include "xreg/estimation/registration_ransac.h"
#include "xreg/estimation/registration_teaser.h"
// #include "xreg/estimation/registration_vgicp.h"

#include "xreg/common/config.h"

// #include <filesystem>

namespace xreg {

using namespace std;

System::System(ros::NodeHandle nh) : nh_(nh) {
  // viewer_ = std::unique_ptr<Visualizer>(new Visualizer(nh_));
  // thread_viz_ =
  //     unique_ptr<thread>(new std::thread(&Visualizer::spin, viewer_.get()));

  initParameters(nh_);

  loadFrames(common::data_path + common::dataset);
}

void System::spin() {
  // visualizeIdx();

  // return;

  if (common::b_init) {
    if (common::init_method == "teaser") {
      TeaserOptions options;
      auto init_ = RegistrationBase::Ptr(new RegistrationTeaser(options));
      init_->setInputFrames(frames_);
      init_->initialize();

      init_->registration();
    } else if (common::init_method == "preload") {
      CHECK(!common::init_traj_path.empty());
      eigen_aligned_std_vector<Quaterniond> rot;
      eigen_aligned_std_vector<Vector3d> trans;
      readTrajectory(common::init_traj_path + '/' + common::dataset, rot,
                     trans);

      CHECK_EQ(frames_.size(), trans.size());

      for (size_t i = 0; i < frames_.size(); i++) {
        Quaterniond q = rot[i];
        Vector3d t = trans[i];
        frames_[i]->setTwb(SE3(q, t));
      }

      SE3 Tb0w = frames_[0]->getTbw();
      for (size_t i = 0; i < frames_.size(); i++) {
        // Quaterniond q = rot[i];
        // Vector3d t = trans[i];
        frames_[i]->setTwb(Tb0w * frames_[i]->getTwb());
      }
    } else {
      // TODO:
      throw std::runtime_error("not implemented initializer");
    }
  }

  // if (common::method == "icp_pt2pt") {
  //   registration_ = RegistrationBase::Ptr(new RegistrationICP);
  // }
  if (common::method == "ndt" || common::method == "gicp" ||
      common::method == "vgicp" || common::method == "icp_pt2pt") {
    NDTOptions options;
    options.b_scan2map = reg::b_scan2map;
    options.method = common::method;
    registration_ =
        RegistrationBase::Ptr(new RegistrationNDT(common::method, options));
  } else if (common::method == "icp_pt2pl") {
    registration_ = RegistrationBase::Ptr(new RegistrationICPPlane);
  }
  // else if (common::method == "gicp") {
  //   registration_ = RegistrationBase::Ptr(new RegistrationGICP);
  // } else if (common::method == "vgicp") {
  //   registration_ = RegistrationBase::Ptr(new RegistrationVGICP);
  // }
  else if (common::method == "balm") {
    registration_ = RegistrationBase::Ptr(new RegistrationBALM);
  } else if (common::method == "ba") {
    registration_ = RegistrationBase::Ptr(new RegistrationBA);
  } else if (common::method == "ef") {
    registration_ = RegistrationBase::Ptr(new RegistrationEFSGD);
  } else if (common::method == "ba_ef") {
    registration_ = RegistrationBase::Ptr(new RegistrationEigenFactor);
  } else if (common::method == "teaser") {
    TeaserOptions options;
    registration_ = RegistrationBase::Ptr(new RegistrationTeaser(options));
    registration_->setOverlap(overlap_matrix_);
  } else if (common::method == "fgr") {
    // TeaserOptions options;
    registration_ = RegistrationBase::Ptr(new RegistrationFGR());
    registration_->setOverlap(overlap_matrix_);
  } else if (common::method == "ransac") {
    // TeaserOptions options;
    // registration_ = RegistrationBase::Ptr(new RegistrationFGR());
    // registration_->setOverlap(overlap_matrix_);
    registration_ = RegistrationBase::Ptr(new RegistrationRANSAC());
    registration_->setOverlap(overlap_matrix_);
  } else if (common::method == "fpfh") {
    // registration_ = RegistrationBase::Ptr(new RegistrationFPFH);
    throw std::runtime_error("fpfh");
  }

  // registration_ = RegistrationBase::Ptr(new RegistrationGICP);

  registration_->setInputFrames(frames_);
  registration_->initialize();

  auto start = std::chrono::steady_clock::now();
  registration_->registration();
  auto end = std::chrono::steady_clock::now();
  auto duration =
      std::chrono::duration_cast<std::chrono::milliseconds>(end - start);
  // std::cout << "elapsed time: " << duration.count() << " ms\n";
  INFOLOG("elapsed time: {} ms", duration.count());

  summarize(duration.count());
}

void System::summarize(double time) {
  // STEP. output trajectory
  INFOLOG("write traj to {}", common::log_path + "/traj");
  ofstream ofs(common::log_path + "/traj");

  ofstream ofs_time(common::log_path + "/est/timing", ios::out | ios::app);
  ofs_time << common::dataset << " " << time / frames_.size() << endl;
  ofs_time.close();

  ofs << std::setprecision(8) << std::fixed;
  for (int i = 0; i < frames_.size(); i++) {
    // auto
    auto frame = frames_[i];
    SE3 Twb = frame->getTwb();
    Quaterniond q = Twb.unit_quaternion();
    Vector3d t = Twb.translation();
    ofs << i << ' ' << t.x() << ' ' << t.y() << ' ' << t.z() << ' ' << q.x()
        << ' ' << q.y() << ' ' << q.z() << ' ' << q.w() << endl;
  }
  ofs.close();

  // x
  double err_rot = 0, err_trans = 0, err_rel = 0;
  for (size_t i = 1; i < frames_.size(); i++) {
    auto frame_src = frames_[i];
    auto frame_tgt = frames_[i - 1];

    SE3 deltaT_gt = frame_tgt->Tgt_.inverse() * frame_src->Tgt_;
    SE3 deltaT = frame_tgt->getTbw() * frame_src->getTwb();
    SE3 deltaT_err = (deltaT_gt.inverse() * deltaT);
    INFOLOG("frame {}, delta: {}", i, deltaT_err.translation().norm());

    double err_rel_frame = deltaT_err.translation().norm();
    err_rel += err_rel_frame;

    SE3 Terr = frame_tgt->Tgt_.inverse() * frame_tgt->getTwb();

    double err_trans_frame = Terr.translation().norm();
    err_trans += err_trans_frame;

    double err_rot_frame = Terr.so3().log().norm() / M_PI * 180.0;
    err_rot += err_rot_frame;
  }

  double time_frame = time;
  err_rel /= (frames_.size() - 1);
  time_frame /= (frames_.size() - 1);
  err_trans /= (frames_.size() - 1);
  err_rot /= (frames_.size() - 1);

  WARNLOG("err_rel: {} , err_t: {}, err_rot: {}, time: {}", err_rel, err_trans,
          err_rot, time_frame);
  // WARNLOG("{}, {}, {}, {}", err_rel, err_trans, err_rot, time_frame);
}

void System::shutdown() {
  // stop viewer
  if (viewer_) {
    viewer_->stop();

    if (thread_viz_) {
      thread_viz_->join();
    }
  }
}

void System::loadFrames(const std::string &file_path) {
  // load csv
  {
    std::string csv_file = file_path + "/icpList.csv";
    ifstream ifs(csv_file);

    if (!ifs.is_open()) {
      ERRLOG("file not opened: {}", csv_file);
    }

    string line;
    // header
    getline(ifs, line);

    // data
    size_t frame_idx = 0;
    while (!ifs.eof() && std::getline(ifs, line)) {
      std::stringstream ss(line);
      string fname;
      ss >> fname;
      int colIdx = 0;
      double val;
      vector<double> vals;
      while (ss >> val) {
        // If the next token is a comma, ignore it and move on
        if (ss.peek() == ',') ss.ignore();

        // INFOLOG("val: {}", val);
        vals.push_back(val);
      }
      //   exit(0);
      Matrix3d rot_mat;
      rot_mat << vals[0], vals[1], vals[2], vals[4], vals[5], vals[6], vals[8],
          vals[9], vals[10];

      Quaterniond rot(rot_mat);
      rot.normalize();

      Vector3d trans;
      trans << vals[3], vals[7], vals[11];

      auto frame = Frame::Ptr(new Frame);
      frames_.push_back(frame);
      frame->idx_ = frame_idx;
      frame_idx++;
      frame->Tgt_ = SE3(rot, trans);
    }
  }
  INFOLOG("num_frames: {}", frames_.size());
  // INFOLOG("last t: {}", frames_.back()->Tgt_.translation().transpose());

  for (size_t i = 0; i < frames_.size(); i++) {
    string pcd_file = file_path + "/pcd/" + to_string(i) + ".pcd";

    auto cloud = PointCloudXYZ::Ptr(new PointCloudXYZ);
    pcl::io::loadPCDFile(pcd_file, *cloud);
    frames_[i]->cloud_ = cloud;
  }

  // load overlap matrix
  {
    size_t num_frames = frames_.size();
    overlap_matrix_.resize(num_frames, num_frames);
    std::string csv_file = file_path + "/overlap.csv";

    ifstream ifs(csv_file);

    if (!ifs.is_open()) {
      ERRLOG("file not opened: {}", csv_file);
    }

    // data
    std::string line;
    size_t frame_idx = 0;
    // while (!ifs.eof() && )
    for (size_t i = 0; i < num_frames; i++) {
      std::getline(ifs, line);
      std::stringstream ss(line);

      for (size_t j = 0; j < num_frames; j++) {
        double v;
        ss >> v;
        if (ss.peek() == ',') ss.ignore();

        overlap_matrix_(i, j) = v;
      }
    }

    // cout << overlap_matrix_ << endl;
    // exit(0);
  }

  //   visualize();
}

inline Vector3i makeJet3B(float id) {
  if (id <= 0) return Vector3i(128, 0, 0);
  if (id >= 1) return Vector3i(0, 0, 128);

  int icP = (id * 8);
  float ifP = (id * 8) - icP;

  if (icP == 0) return Vector3i(255 * (0.5 + 0.5 * ifP), 0, 0);
  if (icP == 1) return Vector3i(255, 255 * (0.5 * ifP), 0);
  if (icP == 2) return Vector3i(255, 255 * (0.5 + 0.5 * ifP), 0);
  if (icP == 3) return Vector3i(255 * (1 - 0.5 * ifP), 255, 255 * (0.5 * ifP));
  if (icP == 4)
    return Vector3i(255 * (0.5 - 0.5 * ifP), 255, 255 * (0.5 + 0.5 * ifP));
  if (icP == 5) return Vector3i(0, 255 * (1 - 0.5 * ifP), 255);
  if (icP == 6) return Vector3i(0, 255 * (0.5 - 0.5 * ifP), 255);
  if (icP == 7) return Vector3i(0, 0, 255 * (1 - 0.5 * ifP));
  return Vector3i(255, 255, 255);
}

void System::visualizeIdx() {
  {
    // PointCloudXYZ::Ptr cloud_map_raw(new PointCloudXYZ);
    // downsample(cloud_map_raw, cloud_map, 0.05);

    pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_map(
        new pcl::PointCloud<pcl::PointXYZRGB>);
    for (size_t i = 0; i < frames_.size(); i++) {
      auto frame = frames_[i];
      // viewer_->addCloud("cloud" + to_string(i), "frame" + to_string(i),
      //                   frame->cloud_);

      // viewer_->setTransform(frame->Tgt_, "frame" + to_string(i));

      Vector3i c = makeJet3B((float)i / frames_.size());

      PointCloudXYZ::Ptr cloud_tmp(new PointCloudXYZ);
      pcl::transformPointCloud(*frame->cloud_, *cloud_tmp,
                               frame->Tgt_.matrix().cast<float>());
      pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_color(
          new pcl::PointCloud<pcl::PointXYZRGB>);

      for (auto &&pt : cloud_tmp->points) {
        pcl::PointXYZRGB ptc(c.x(), c.y(), c.z());
        ptc.x = pt.x;
        ptc.y = pt.y;
        ptc.z = pt.z;

        cloud_color->push_back(ptc);
      }

      (*cloud_map) += *cloud_color;
    }

    viewer_->addCloud("cloud_map_idx", "map", *cloud_map);

    // PointCloudXYZ::Ptr cloud_map(new PointCloudXYZ);
    // downsample(cloud_map_raw, cloud_map, 0.05);

    // viewer_->addCloud("cloud_map_gt", "map", cloud_map);
  }
}

void System::visualize() {
  // {
  //   PointCloudXYZ::Ptr cloud_map_raw(new PointCloudXYZ);
  //   for (size_t i = 0; i < frames_.size(); i++) {
  //     auto frame = frames_[i];
  //     // viewer_->addCloud("cloud" + to_string(i), "frame" + to_string(i),
  //     //                   frame->cloud_);

  //     // viewer_->setTransform(frame->Tgt_, "frame" + to_string(i));

  //     PointCloudXYZ::Ptr cloud_tmp(new PointCloudXYZ);
  //     pcl::transformPointCloud(*frame->cloud_, *cloud_tmp,
  //                              frame->Tgt_.matrix().cast<float>());

  //     (*cloud_map_raw) += *cloud_tmp;
  //   }

  //   viewer_->addCloud("cloud_map_gt", "map", *cloud_map_raw);

  //   // PointCloudXYZ::Ptr cloud_map(new PointCloudXYZ);
  //   // downsample(cloud_map_raw, cloud_map, 0.05);

  //   // viewer_->addCloud("cloud_map_gt", "map", cloud_map);
  // }

  // {
  //   PointCloudXYZ::Ptr cloud_map_raw(new PointCloudXYZ);
  //   for (size_t i = 0; i < frames_.size(); i++) {
  //     auto frame = frames_[i];

  //     PointCloudXYZ::Ptr cloud_tmp(new PointCloudXYZ);
  //     pcl::transformPointCloud(*frame->cloud_, *cloud_tmp,
  //                              frame->getTwb().matrix().cast<float>());

  //     (*cloud_map_raw) += *cloud_tmp;
  //   }
  //   viewer_->addCloud("cloud_map_est", "map", *cloud_map_raw);

  //   // PointCloudXYZ::Ptr cloud_map(new PointCloudXYZ);
  //   // downsample(cloud_map_raw, cloud_map, 0.05);

  //   // viewer_->addCloud("cloud_map_est", "map", cloud_map);
  // }
}

void System::downsample(PointCloudXYZ::Ptr cloud,
                        PointCloudXYZ::Ptr cloud_filtered, double size) {
  pcl::VoxelGrid<PointXYZ> filter;
  filter.setLeafSize(size, size, size);
  filter.setInputCloud(cloud);
  filter.filter(*cloud_filtered);
}

void System::readTrajectory(const std::string &traj_file,
                            eigen_aligned_std_vector<Quaterniond> &rot,
                            eigen_aligned_std_vector<Vector3d> &trans) {
  rot.clear();
  trans.clear();

  ifstream fs_mean(traj_file.c_str());
  string str_line;
  while (getline(fs_mean, str_line) && !fs_mean.eof()) {
    stringstream ss(str_line);
    double time, x, y, z, qx, qy, qz, qw;
    ss >> time;
    ss >> x;
    ss >> y;
    ss >> z;
    ss >> qx;
    ss >> qy;
    ss >> qz;
    ss >> qw;

    Eigen::Vector3d trans_tmp(x, y, z);
    Eigen::Quaterniond q_tmp(qw, qx, qy, qz);
    trans.push_back(trans_tmp);
    rot.push_back(q_tmp);
  }

  // CHECK_EQ(rot.size(), frames_);
}

}  // namespace xreg
