#include "xreg/types/cell.h"

namespace xreg {

using namespace std;

// Segment::Segment() {

// }
#define FIX_OUTLIER 1

Gaussian3d::Gaussian3d(eigen_aligned_std_vector<Vector3d> &pts) {
  num = pts.size();

  // compute mean
  mean.setZero();
  for (int i = 0; i < pts.size(); i++) {
    mean = ((double)i) / (i + 1) * mean + 1.0 / (i + 1) * pts[i];
  }

  // compute cov
  cov.setZero();
  for (int i = 0; i < pts.size(); i++) {
    Vector3d delta = pts[i] - mean;
    cov = (i * 1.0) / (i + 1) * cov + 1.0 / (i + 1) * delta * delta.transpose();
  }

  decompose();
}

Gaussian3d::Gaussian3d(int n, const Vector3d &mu, const Matrix3d &cov_mat)
    : num(n), mean(mu), cov(cov_mat) {}

// void Gaussian3d::setParam(int n, const Vector3d &mu, const Matrix3d &c) {
//   num = n;
//   mean = mu;
//   cov = c;
// }

// TODO: which is faster? JacobiSVD or SelfAdjoint
void Gaussian3d::decompose() {
  auto eigen_solver = Eigen::SelfAdjointEigenSolver<Eigen::Matrix3d>(cov);

  Matrix3d axis_ = eigen_solver.eigenvectors();

  axis_.col(2) = axis_.col(0).cross(axis_.col(1));
  rot = SO3(axis_);
  rot_mat = rot.matrix();
  scalar = eigen_solver.eigenvalues();
  // cout << scalar.transpose() << endl;
  // exit(0);

  // {
  //   auto eigen_solver =
  //       Eigen::JacobiSVD<Eigen::Matrix3d>(cov_, Eigen::ComputeFullU);

  //   // if (eigen_solver.info() == Eigen::Success)
  //   {

  //     Matrix3d axis_ = eigen_solver.matrixU();
  //     // axis_.col(2) = axis_.col(0).cross(axis_.col(1));
  //     Matrix3d ax;
  //     ax << axis_.col(2), axis_.col(1), axis_.col(2).cross(axis_.col(1));
  //     rot_ = SO3(ax);

  //     Eigen::Vector3d scalar = eigen_solver.singularValues();
  //     scalar_.x() = scalar.z();
  //     scalar_.y() = scalar.y();
  //     scalar_.z() = scalar.x();

  //     // cout << scalar_.transpose() << endl;

  //     optable_ = true;

  //     nhat_t_ = ax.col(0).transpose();
  //   }
}

Gaussian3d Gaussian3d::merge(const Gaussian3d &g) {
  // throw std::runtime_error("todo");
  int _num = this->num + g.num;

  double f0 = (this->num * 1.0) / _num;
  double f1 = (g.num * 1.0) / _num;

  Vector3d _mean = f0 * this->mean + f1 * g.mean;

  Vector3d d0 = this->mean - _mean;
  Vector3d d1 = g.mean - _mean;

  Matrix3d _cov = f0 * this->cov + f1 * g.cov + f0 * d0 * d0.transpose() +
                  f1 * d1 * d1.transpose();

  return Gaussian3d(_num, _mean, _cov);
}

void Gaussian3d::addPoint(const Vector3d &pt) {
  throw std::runtime_error("todo");
}

Segment::Segment(const eigen_aligned_std_vector<Vector3d> &pts, bool cache_pts)
    : pts_(pts) {
  num_ = pts_.size();

  estimateParam();

  estimateParamHomo();

  if (!cache_pts) {
    clearCache();
  }
}

// Matrix4d Segment::llt() { return L_; }

Matrix4d Segment::getQ() {
  CHECK_NOTNULL(frame_);
  Matrix4d T = frame_->getTwb().matrix();
  Q_ = T * S_ * T.transpose();

  return Q_;
}

void Segment::estimateParam() {
  num_ = pts_.size();

  gaussian_ = Gaussian3d(pts_);
}

void Segment::estimateParamHomo() {
  // ADD(hhy): x
  S_.setZero();
  for (const Vector3d &p : pts_) {
    CHECK_LE(p.norm(), 10000);
    Vector4d pHomog;
    pHomog << p, 1.0;
    S_ += pHomog * pHomog.transpose();
  }

  L_ = S_.llt().matrixL();
}

int Cell::countTotal() {
  int num = info_map_.num;

  for (auto &&p : segments_) {
    num += p->num_;
  }
}

Gaussian3d Cell::merge(std::vector<Segment *> &segments) {
  Vector3d mean = Vector3d::Zero();
  Matrix3d cov = Matrix3d::Zero();

  int num = 0;
  for (auto &&p : segments) {
    num += p->num_;
  }

  for (size_t i = 0; i < segments.size(); i++) {
    // num_total_ += ff->planes_[idx_]->pts_.size();
    Segment *segment = segments[i];
    Frame *frame = segment->frame_;

    mean += (segment->num_ * 1.0 / num) *
            (frame->getTwb() * segment->gaussian_.mean);
  }

  // for (size_t i = idx)
  for (size_t i = 0; i < segments.size(); i++) {
    Segment *segment = segments[i];
    Frame *frame = segment->frame_;
    SE3 Twb = frame->getTwb();

    Matrix3d rot_w_c = Twb.rotationMatrix();
    cov += (segment->num_ * 1.0 / num) *
           (rot_w_c * segment->gaussian_.cov * rot_w_c.transpose());
    Vector3d delta = (Twb * segment->gaussian_.mean - mean);
    cov += (segment->num_ * 1.0 / num) * delta * delta.transpose();
  }

  return Gaussian3d(num, mean, cov);
}

Gaussian3d Cell::merge2(const Gaussian3d &gaussian,
                        const std::vector<Segment *> &segments) {

  int num = gaussian.num;
  // CHECK_EQ(segments.size(), 1);
  for (auto &&p : segments) {
    num += p->num_;
  }
  // INFOLOG("num_total: {}, num_raw: {}", num, gaussian.num);
  // getchar();

  // Vector3d mean = (gaussian.num * 1.0 / num) * gaussian.mean;
  Vector3d mean = gaussian.mean;
  for (size_t i = 0; i < segments.size(); i++) {
    // num_total_ += ff->planes_[idx_]->pts_.size();
    Segment *segment = segments[i];
    Frame *frame = segment->frame_;
    // CHECK_EQ(frame, segments[0]->frame_);
    // CHECK_NOTNULL(frame);

    mean += (segment->num_ * 1.0 / num) *
            (frame->getTwb() * segment->gaussian_.mean);
    // mean += (frame->getTwb() * segment->gaussian_.mean);
  }
  // if (gaussian.num != 0) {

  //   mean /= segments.size() + 1;
  // } else {
  //   mean /= segments.size();
  // }

  // for (size_t i = idx)
  Matrix3d cov = (gaussian.num * 1.0 / num) * gaussian.cov;
  {
    Vector3d delta = (gaussian.mean - mean);
    cov += (gaussian.num * 1.0 / num) * delta * delta.transpose();
  }

  for (size_t i = 0; i < segments.size(); i++) {
    Segment *segment = segments[i];
    Frame *frame = segment->frame_;
    SE3 Twb = frame->getTwb();

    Matrix3d rot_w_c = Twb.rotationMatrix();
    cov += (segment->num_ * 1.0 / num) *
           (rot_w_c * segment->gaussian_.cov * rot_w_c.transpose());
    Vector3d delta = (Twb * segment->gaussian_.mean - mean);
    cov += (segment->num_ * 1.0 / num) * delta * delta.transpose();
  }

  auto res = Gaussian3d(num, mean, cov);
  res.decompose();
  return res;
}

Gaussian3d Cell::merge(const Gaussian3d &gaussian,
                       const std::vector<Segment *> &segments) {

  int num = gaussian.num;
  // CHECK_EQ(segments.size(), 1);
  for (auto &&p : segments) {
    num += p->num_;
  }
  // INFOLOG("num_total: {}, num_raw: {}", num, gaussian.num);
  // getchar();

  // Vector3d mean = (gaussian.num * 1.0 / num) * gaussian.mean;
  Vector3d mean = gaussian.mean;
  for (size_t i = 0; i < segments.size(); i++) {
    // num_total_ += ff->planes_[idx_]->pts_.size();
    Segment *segment = segments[i];
    Frame *frame = segment->frame_;
    mean += (frame->getTwb() * segment->gaussian_.mean);
  }
  if (gaussian.num != 0) {
    mean /= segments.size() + 1;
  } else {
    mean /= segments.size();
  }

  // for (size_t i = idx)
  // TODO: size of segments;
  Matrix3d cov = gaussian.cov;
  if (gaussian.num != 0) {
    Vector3d delta = (gaussian.mean - mean);
    cov += delta * delta.transpose();
  }

  for (size_t i = 0; i < segments.size(); i++) {
    Segment *segment = segments[i];
    Frame *frame = segment->frame_;
    SE3 Twb = frame->getTwb();

    Matrix3d rot_w_c = Twb.rotationMatrix();
    cov += (rot_w_c * segment->gaussian_.cov * rot_w_c.transpose());
    Vector3d delta = (Twb * segment->gaussian_.mean - mean);
    cov += delta * delta.transpose();
  }
  if (gaussian.num != 0) {
    cov /= segments.size() + 1;
  } else {
    cov /= segments.size();
  }

  auto res = Gaussian3d(num, mean, cov);
  res.decompose();
  return res;
}

Cell::Cell() {
  qmat_all_.setZero();
  qmat_sw_.setZero();
  qmat_map_.setZero();
}

double Cell::updateInfoAllEigenFactor(bool ignore_back) {
  if (segments_.size() < 1 || (segments_.size() == 1 && ignore_back)) {
    qmat_all_ = qmat_map_;
    return 0.0;
  }

  vector<Segment *> segments;
  size_t size = segments_.size();
  if (ignore_back) {
    size--;
  }

  qmat_sw_.setZero();
  for (size_t i = 0; i < size; i++) {
    if (!is_outlier_[i]) {
      // segments.push_back(segments_[i]);
      Matrix4d Qi = segments_[i]->getQ();
      qmat_sw_ += Qi;
    }
  }

  qmat_all_ = qmat_sw_ + qmat_map_;

  Eigen::JacobiSVD<Matrix4d> svd(qmat_all_, Eigen::ComputeFullU);
  // INFOLOG("lambda: {}", svd.singularValues().transpose());
  pi_ = svd.matrixU().col(3);
  // exit(0);

  double lambda = svd.singularValues()(3);

  return lambda;

  // info_all_ = merge2(info_map_, segments);
}

void Cell::updateInfoAll(bool ignore_back) {
  if (segments_.size() < 1 || (segments_.size() == 1 && ignore_back)) {
    info_all_ = info_map_;
    return;
  }
  // auto end_iter = segments_.end();
  // if (ignore_back) {
  //   end_iter--;
  // }
  // vector<Segment *> segments(segments_.begin(), end_iter);
  vector<Segment *> segments;
  // CHECK_EQ(segments_.size(), is_outlier_.size());
  size_t size = segments_.size();
  if (ignore_back) {
    size--;
  }
  for (size_t i = 0; i < size; i++) {
    if (!is_outlier_[i]) {
      segments.push_back(segments_[i]);
    }
  }

  info_all_ = merge2(info_map_, segments);
  // info_all_ = merge(info_map_, segments);
}

void Cell::marginalize(int num_frames) {
  vector<Segment *> segments;
  for (int i = 0; i != num_frames; i++) {
    // Frame *frame = frames_.front();
    Segment *segment = segments_.front();

    // frame->
    frames_.pop_front();
    segments_.pop_front();
    segments.push_back(segment);
  }

  info_map_ = merge(info_map_, segments);

  for (auto &&segment : segments) {
    delete segment;
  }
}

// Segment *Cell::getSegment(Frame *frame) {
//   if (segments_.count(frame)) {
//     return segments_[frame];
//   } else {
//     auto segment = new Segment;
//     segments_[frame] = segment;

//     frames_.push_back(frame);
//     segs_.push_back(segment);

//     return segment;
//   }
// }

// bool Cell::checkX() {
//   // cout << scalar_.y() / scalar_.x() << endl;
//   // cout << frames_.size() << endl;
//   // cout << scalar_.x() << ' ' << scalar_.y() << endl;
//   return num_total_ >= 15 && scalar_.y() / scalar_.x() > 9.0;
// }
bool Cell::checkFeature() {
  // return countTotal() >= 15 && scalar_.y() / scalar_.x() > 9.0;
}

void Cell::solveParam() {
  // mean_.setZero();
  // cov_.setZero();
  // optable_ = false;

  // num_total_ = 0;
  int num;
  Vector3d mean;
  Matrix3d cov;

  for (size_t i = 0; i < frames_.size(); i++) {
#ifdef FIX_OUTLIER
    if (is_outlier_[i]) {
      continue;
    }
#endif
    num += segments_[i]->pts_.size();
  }

  // TODO: min_num
  // if (num_total_ < 15)
  //   return;

  for (size_t i = 0; i < frames_.size(); i++) {
    // num_total_ += ff->planes_[idx_]->pts_.size();
#ifdef FIX_OUTLIER
    if (is_outlier_[i]) {
      continue;
    }
#endif
    auto &&ff = frames_[i];

    auto plocal = segments_[i];

    mean +=
        (plocal->num_ * 1.0 / num) * (ff->getTwb() * plocal->gaussian_.mean);
  }

  // for (size_t i = idx)
  for (size_t i = 0; i < frames_.size(); i++) {
    // num_total_ += ff->planes_[idx_]->pts_.size();
#ifdef FIX_OUTLIER
    if (is_outlier_[i]) {
      continue;
    }
#endif
    auto ff = frames_[i];
    SE3 Twb = ff->getTwb();

    Matrix3d rot_w_c = Twb.rotationMatrix();
    auto plocal = segments_[i];
    cov += (plocal->num_ * 1.0 / num) *
           (rot_w_c * plocal->gaussian_.cov * rot_w_c.transpose());
    Vector3d delta = (Twb * plocal->gaussian_.mean - mean);
    cov += (plocal->num_ * 1.0 / num) * delta * delta.transpose();
  }

  info_sw_ = Gaussian3d(num, mean, cov);
  info_all_ = info_map_;

  info_all_.merge(info_sw_);

  // {
  //   auto eigen_solver = Eigen::SelfAdjointEigenSolver<Eigen::Matrix3d>(cov_);

  //   if (eigen_solver.info() == Eigen::Success) {

  //     Matrix3d axis_ = eigen_solver.eigenvectors();
  //     axis_.col(2) = axis_.col(0).cross(axis_.col(1));
  //     rot_ = SO3(axis_);
  //     scalar_ = eigen_solver.eigenvalues();

  //     // cout << scalar_.transpose() << endl;

  //     optable_ = true;

  //     nhat_t_ = axis_.col(0);
  //   } else {
  //     /* code */
  //   }
  // }
  // {
  //   auto eigen_solver =
  //       Eigen::JacobiSVD<Eigen::Matrix3d>(cov_, Eigen::ComputeFullU);

  //   // if (eigen_solver.info() == Eigen::Success)
  //   {

  //     Matrix3d axis_ = eigen_solver.matrixU();
  //     // axis_.col(2) = axis_.col(0).cross(axis_.col(1));
  //     Matrix3d ax;
  //     ax << axis_.col(2), axis_.col(1), axis_.col(2).cross(axis_.col(1));
  //     rot_ = SO3(ax);

  //     Eigen::Vector3d scalar = eigen_solver.singularValues();
  //     scalar_.x() = scalar.z();
  //     scalar_.y() = scalar.y();
  //     scalar_.z() = scalar.x();

  //     // cout << scalar_.transpose() << endl;

  //     optable_ = true;

  //     nhat_t_ = ax.col(0).transpose();
  //   }

  //   // else {
  //   //   /* code */
  //   // }
  // }

  // cout << eigen_solver.eigenvalues().transpose() << endl;

  // TODO:
}

double Cell::solveParamHomo() {
  qmat_sw_.setZero();

  for (size_t i = 0; i < segments_.size(); i++) {

#ifdef FIX_OUTLIER
    if (is_outlier_[i]) {
      continue;
    }
#endif

    auto &&seg = segments_[i];
    CHECK_NOTNULL(seg);
    // FIXME:
    Matrix4d Qi = seg->getQ();
    qmat_sw_ += Qi;
  }

  qmat_all_ = qmat_sw_ + qmat_map_;

  Eigen::JacobiSVD<Matrix4d> svd(qmat_all_, Eigen::ComputeFullU);
  pi_ = svd.matrixU().col(3);
  lambda_ = svd.singularValues()(3);

  return lambda_;
}

} // namespace xreg
