#include "xreg/utils/math_utils.h"

namespace xreg {

Matrix3d MathUtils::skew(const Vector3d &v) {
  Matrix3d m;
  m.fill(0.);
  m(0, 1) = -v(2);
  m(0, 2) = v(1);
  m(1, 2) = -v(0);
  m(1, 0) = v(2);
  m(2, 0) = -v(1);
  m(2, 1) = v(0);
  return m;
}

// cv::Mat W = (cv::Mat_<float>(3, 3) << 0, -z, y, z, 0, -x, -y, x, 0);
// if (d < eps_inte_rot) {

//   deltaR = I + W;
//   rightJ = cv::Mat::eye(3, 3, CV_32F);

// } else {
//   deltaR = I + W * sin(d) / d + W * W * (1.0f - cos(d)) / d2;
//   rightJ = I - W * (1.0f - cos(d)) / d2 + W * W * (d - sin(d)) / (d2 * d);
// }

Matrix3d MathUtils::rightJacobian(const Vector3d &rot_se3) {
  // Vector3d rot_se3 = rot.log();
  Matrix3d w = skew(rot_se3);
  // .norm()
  double theta_sq = rot_se3.squaredNorm();
  double theta = sqrt(theta_sq);
  if (theta < Sophus::Constants<double>::epsilon()) {
    return Matrix3d::Identity() + w;

  } else {
    // deltaR = I + W * sin(d) / d + W * W * (1.0f - cos(d)) / d2;
    Matrix3d rightJ = Matrix3d::Identity() -
                      w * (1.0f - cos(theta)) / theta_sq +
                      w * w * (theta - sin(theta)) / (theta_sq * theta);
    return rightJ;
  }
}

Matrix3d MathUtils::rightJacobian(const SO3 &rot) {
  Vector3d rot_se3 = rot.log();
  return rightJacobian(rot_se3);
  // Matrix3d w = skew(rot_se3);
  // // .norm()
  // double theta_sq = rot_se3.squaredNorm();
  // double theta = sqrt(theta_sq);
  // if (theta < Sophus::Constants<double>::epsilon()) {
  //   return Matrix3d::Identity() + w;

  // } else {
  //   // deltaR = I + W * sin(d) / d + W * W * (1.0f - cos(d)) / d2;
  //   Matrix3d rightJ = Matrix3d::Identity() -
  //                     w * (1.0f - cos(theta)) / theta_sq +
  //                     w * w * (theta - sin(theta)) / (theta_sq * theta);
  //   return rightJ;
  // }
}

Matrix3d MathUtils::inverseRightJacobian(const Vector3d &rot_se3) {
  // Vector3d rot_se3 = rot.log();
  Matrix3d w = skew(rot_se3);
  // .norm()
  double theta_sq = rot_se3.squaredNorm();
  double theta = sqrt(theta_sq);
  Matrix3d J = Matrix3d::Identity();
  if (theta >= Sophus::Constants<double>::epsilon()) {
    J += 0.5 * w +
         w * w *
             (1.0 / theta_sq - (1.0 + cos(theta)) / (2.0 * theta * sin(theta)));
  }

  return J;
}

// Matrix3d MathUtils::computeEssentialMatrix(const SE3Quat &Tcw1,
//                                            const SE3Quat &Tcw2

// ) {

//   Quaterniond rot_c1_w = Tcw1.rotation();
//   Vector3d t_c1_w = Tcw1.translation();
//   Quaterniond rot_c2_w = Tcw2.rotation();
//   Vector3d t_c2_w = Tcw2.translation();

//   Quaterniond rot_c1_c2 = rot_c1_w * rot_c2_w.conjugate();
//   Vector3d t_c1_c2 = -(rot_c1_c2 * t_c2_w) + t_c1_w;

//   Matrix3d t12_skew = skew(t_c1_c2);

//   return t12_skew * rot_c1_c2.toRotationMatrix();
// }

// Matrix3d MathUtils::computeFundamentalMatrix(const SE3Quat &Tcw1,
//                                              const Matrix3d &K1,
//                                              const SE3Quat &Tcw2,
//                                              const Matrix3d &K2

// ) {
//   Matrix3d essential_mat = computeEssentialMatrix(Tcw1, Tcw2);

//   return K1.transpose().inverse() * essential_mat * K2.inverse();
// }
Eigen::MatrixXd MathUtils::marginalize(const Eigen::MatrixXd &H,
                                       const int &start, const int &end) {
  // Goal
  // a  | ab | ac       a*  | 0 | ac*
  // ba | b  | bc  -->  0   | 0 | 0
  // ca | cb | c        ca* | 0 | c*

  // Size of block before block to marginalize
  const int a = start;
  // Size of block to marginalize
  const int b = end - start + 1;
  // Size of block after block to marginalize
  const int c = H.cols() - (end + 1);

  // Reorder as follows:
  // a  | ab | ac       a  | ac | ab
  // ba | b  | bc  -->  ca | c  | cb
  // ca | cb | c        ba | bc | b

  Eigen::MatrixXd Hn = Eigen::MatrixXd::Zero(H.rows(), H.cols());
  if (a > 0) {
    Hn.block(0, 0, a, a) = H.block(0, 0, a, a);
    Hn.block(0, a + c, a, b) = H.block(0, a, a, b);
    Hn.block(a + c, 0, b, a) = H.block(a, 0, b, a);
  }
  if (a > 0 && c > 0) {
    Hn.block(0, a, a, c) = H.block(0, a + b, a, c);
    Hn.block(a, 0, c, a) = H.block(a + b, 0, c, a);
  }
  if (c > 0) {
    Hn.block(a, a, c, c) = H.block(a + b, a + b, c, c);
    Hn.block(a, a + c, c, b) = H.block(a + b, a, c, b);
    Hn.block(a + c, a, b, c) = H.block(a, a + b, b, c);
  }
  Hn.block(a + c, a + c, b, b) = H.block(a, a, b, b);

  // Perform marginalization (Schur complement)
  Eigen::JacobiSVD<Eigen::MatrixXd> svd(
      Hn.block(a + c, a + c, b, b), Eigen::ComputeThinU | Eigen::ComputeThinV);
  Eigen::JacobiSVD<Eigen::MatrixXd>::SingularValuesType singularValues_inv =
      svd.singularValues();
  for (int i = 0; i < b; ++i) {
    if (singularValues_inv(i) > 1e-6)
      singularValues_inv(i) = 1.0 / singularValues_inv(i);
    else
      singularValues_inv(i) = 0;
  }
  Eigen::MatrixXd invHb = svd.matrixV() * singularValues_inv.asDiagonal() *
                          svd.matrixU().transpose();
  Hn.block(0, 0, a + c, a + c) =
      Hn.block(0, 0, a + c, a + c) -
      Hn.block(0, a + c, a + c, b) * invHb * Hn.block(a + c, 0, b, a + c);
  Hn.block(a + c, a + c, b, b) = Eigen::MatrixXd::Zero(b, b);
  Hn.block(0, a + c, a + c, b) = Eigen::MatrixXd::Zero(a + c, b);
  Hn.block(a + c, 0, b, a + c) = Eigen::MatrixXd::Zero(b, a + c);

  // Inverse reorder
  // a*  | ac* | 0       a*  | 0 | ac*
  // ca* | c*  | 0  -->  0   | 0 | 0
  // 0   | 0   | 0       ca* | 0 | c*
  Eigen::MatrixXd res = Eigen::MatrixXd::Zero(H.rows(), H.cols());
  if (a > 0) {
    res.block(0, 0, a, a) = Hn.block(0, 0, a, a);
    res.block(0, a, a, b) = Hn.block(0, a + c, a, b);
    res.block(a, 0, b, a) = Hn.block(a + c, 0, b, a);
  }
  if (a > 0 && c > 0) {
    res.block(0, a + b, a, c) = Hn.block(0, a, a, c);
    res.block(a + b, 0, c, a) = Hn.block(a, 0, c, a);
  }
  if (c > 0) {
    res.block(a + b, a + b, c, c) = Hn.block(a, a, c, c);
    res.block(a + b, a, c, b) = Hn.block(a, a + c, c, b);
    res.block(a, a + b, b, c) = Hn.block(a + c, a, b, c);
  }

  res.block(a, a, b, b) = Hn.block(a + c, a + c, b, b);

  return res;
}

} // namespace xreg
