#include "xreg/visualization/visualizer.h"

#include <tf2_ros/transform_broadcaster.h>

#include <pcl_conversions/pcl_conversions.h>

namespace xreg {

using namespace std;

CloudVisualizer::CloudVisualizer(const std::string &name,
                                 const std::string &frame_id,
                                 ros::NodeHandle &nh)
    : name_(name) {

  header_.seq = 0;
  header_.frame_id = frame_id;

  pub_ = nh.advertise<sensor_msgs::PointCloud2>("cloud_" + name_, 10, true);

  msg_ = sensor_msgs::PointCloud2::Ptr(new sensor_msgs::PointCloud2);
}

void CloudVisualizer::visualize() {
  if (!msg_) {
    return;
  }

  //   msg_->header.frame_id = frame_id_;
  //   msg_->header.stamp = ros::Time::now();
  msg_->header = header_;
  msg_->header.stamp = ros::Time::now();
  // msg_->header.seq
  // msg_->header.stamp = ros::Time::now();

  pub_.publish(msg_);
}

template <typename PointT>
void CloudVisualizer::setPointCloud(const pcl::PointCloud<PointT> &cloud) {
  // pcl::toros
  pcl::toROSMsg(cloud, *msg_);
}

Visualizer::Visualizer(ros::NodeHandle &nh) : nh_(nh), stop_(false) {}

template <typename PointT>
void Visualizer::addCloud(const std::string &name, const std::string &frame_id,
                          const pcl::PointCloud<PointT> &cloud,
                          Vector3d offset) {
  unique_lock<mutex> lock(mutex_cloud_);
  // cout << "?" << endl;

  if (!cloud_server_.count(name)) {
    cloud_server_[name] =
        CloudVisualizer::Ptr(new CloudVisualizer(name, frame_id, nh_));
  }
  typename pcl::PointCloud<PointT>::Ptr cloud_tt(new pcl::PointCloud<PointT>);
  // clout_tt = typename pcl::PointCloud<PointT>::Ptr(new
  // pcl::PointCloud<PointT>);
  for (auto &&pt : cloud.points) {
    // pt
    // pcl::PointXYZ
    PointT ptt;
    ptt.x = pt.x + offset.x();
    ptt.y = pt.y + offset.y();
    ptt.z = pt.z + offset.z();

    cloud_tt->points.push_back(ptt);
  }

  cloud_server_[name]->setPointCloud(*cloud_tt);
}

void Visualizer::visualizeCloud() {
  unique_lock<mutex> lock(mutex_cloud_);
  for (auto &&info : cloud_server_) {
    // info.first;
    info.second->visualize();
  }
}

void Visualizer::spin() {
  ros::Rate rate(20);

  while (ros::ok() && !stop_) {
    broadcastTF();
    visualizeCloud();

    rate.sleep();
  }
}

void Visualizer::broadcastTF() {
  // cout
  static tf2_ros::TransformBroadcaster br;

  unique_lock<std::mutex> lock(mutex_pose_);
  for (auto &&info : tf_records_) {
    geometry_msgs::TransformStamped transformStamped;

    transformStamped.header.stamp = ros::Time::now();
    transformStamped.header.frame_id = "map";
    transformStamped.child_frame_id = info.first;

    Vector3d t_w_c = info.second.translation();
    Quaterniond rot_w_c = info.second.so3().unit_quaternion();
    transformStamped.transform.translation.x = t_w_c.x();
    transformStamped.transform.translation.y = t_w_c.y();
    transformStamped.transform.translation.z = t_w_c.z();

    transformStamped.transform.rotation.x = rot_w_c.x();
    transformStamped.transform.rotation.y = rot_w_c.y();
    transformStamped.transform.rotation.z = rot_w_c.z();
    transformStamped.transform.rotation.w = rot_w_c.w();
    br.sendTransform(transformStamped);
  }
}

void Visualizer::setTransform(const SE3 &trans, const std::string &name) {

  std::unique_lock<mutex> lock(mutex_pose_);
  // auto pose = SE3(rot, trans);
  tf_records_[name] = trans;
}

void Visualizer::setTransform(const Quaterniond &rot, const Vector3d &trans,
                              const std::string &name) {

  std::unique_lock<mutex> lock(mutex_pose_);
  auto pose = SE3(rot, trans);
  tf_records_[name] = pose;

  // cout << "what" << endl;

  // if (name == "camera") {
  //   if (!traj_records_.count("camera")) {
  //     traj_records_["camera"] =
  //         TrajectoryViewer::Ptr(new TrajectoryViewer("camera", nh_));
  //   }

  //   auto &ptr = traj_records_["camera"];

  //   ptr->traj_.push_back(pose);
  // }
}

void Visualizer::spinOnce() { visualizeCloud(); }

// template void CloudVisualizer::setPointCloud<PointXYZ>(const
// pcl::PointCloud<PointXYZ> &cloud); template void
// CloudVisualizer::setPointCloud(const pcl::PointCloud<pcl::PointXYZRGB>
// &cloud);

// template <typename PointT>
template void Visualizer::addCloud(const std::string &name,
                                   const std::string &frame_id,
                                   const pcl::PointCloud<PointXYZ> &cloud,
                                   Vector3d offset);

template void
Visualizer::addCloud(const std::string &name, const std::string &frame_id,
                     const pcl::PointCloud<pcl::PointXYZRGB> &cloud,
                     Vector3d offset);

template void Visualizer::addCloud(const std::string &name,
                                   const std::string &frame_id,
                                   const pcl::PointCloud<pcl::PointXYZI> &cloud,
                                   Vector3d offset);
} // namespace xreg
