#include <pcl/io/pcd_io.h>
#include <pcl/keypoints/harris_3d.h>
#include <pcl/keypoints/iss_3d.h>

#include <pcl/filters/extract_indices.h>
#include <pcl/filters/uniform_sampling.h>
#include <pcl/filters/voxel_grid.h>

#include <pcl/registration/correspondence_rejection_sample_consensus.h>
#include <pcl/registration/ia_ransac.h>

#include <teaser/registration.h>

#include "xreg/types/frame.hpp"
#include "xreg/visualization/visualizer.h"

using namespace std;

double downsample_resolution = 0.1;
double kps_resolution = downsample_resolution * 4;

using namespace xreg;

std::vector<xreg::Frame::Ptr> frames;

void loadFrames(const std::string &file_path) {
  // load csv
  {
    std::string csv_file = file_path + "/icpList.csv";
    ifstream ifs(csv_file);

    if (!ifs.is_open()) {
      ERRLOG("file not opened: {}", csv_file);
    }

    string line;
    // header
    getline(ifs, line);

    // data
    while (!ifs.eof() && std::getline(ifs, line)) {
      std::stringstream ss(line);
      string fname;
      ss >> fname;
      int colIdx = 0;
      double val;
      vector<double> vals;
      while (ss >> val) {
        // If the next token is a comma, ignore it and move on
        if (ss.peek() == ',')
          ss.ignore();

        // INFOLOG("val: {}", val);
        vals.push_back(val);
      }
      //   exit(0);
      Matrix3d rot_mat;
      rot_mat << vals[0], vals[1], vals[2], vals[4], vals[5], vals[6], vals[8],
          vals[9], vals[10];

      Quaterniond rot(rot_mat);
      rot.normalize();

      Vector3d trans;
      trans << vals[3], vals[7], vals[11];

      auto frame = Frame::Ptr(new Frame);
      frames.push_back(frame);
      frame->Tgt_ = SE3(rot, trans);
    }
  }
  INFOLOG("num_frames: {}", frames.size());
  INFOLOG("last t: {}", frames.back()->Tgt_.translation().transpose());

  for (size_t i = 0; i < frames.size(); i++) {
    string pcd_file = file_path + "/pcd/" + to_string(i) + ".pcd";

    auto cloud = PointCloudXYZ::Ptr(new PointCloudXYZ);
    pcl::io::loadPCDFile(pcd_file, *cloud);
    frames[i]->cloud_ = cloud;
  }

  //   visualize();
}

static void
findCorrespondences(typename pcl::PointCloud<pcl::FPFHSignature33>::Ptr source,
                    typename pcl::PointCloud<pcl::FPFHSignature33>::Ptr target,
                    std::vector<int> &correspondences) {
  // std::cout <<  << std::flush;
  INFOLOG("correspondence assignment...");

  correspondences.resize(source->size());

  // Use a KdTree to search for the nearest matches in feature space
  pcl::KdTreeFLANN<pcl::FPFHSignature33> descriptor_kdtree;
  descriptor_kdtree.setInputCloud(target);

  // Find the index of the best match for each keypoint, and store it in
  // "correspondences_out"
  // cout << " ? " << endl;
  const int k = 1;
  std::vector<int> k_indices(k);
  std::vector<float> k_squared_distances(k);
  for (int i = 0; i < static_cast<int>(source->size()); ++i) {
    descriptor_kdtree.nearestKSearch(*source, i, k, k_indices,
                                     k_squared_distances);
    correspondences[i] = k_indices[0];
  }
  // std::cout << "OK" << std::endl;
  INFOLOG("done");
}

xreg::SE3 registerByTeaser(pcl::PointCloud<pcl::PointXYZ>::Ptr &kps0,
                           pcl::PointCloud<pcl::PointXYZ>::Ptr &kps1,
                           // pcl::CorrespondencesPtr &corr
                           const std::vector<std::pair<int, int>> &corr) {

  auto cloud_src = std::shared_ptr<teaser::PointCloud>(new teaser::PointCloud);
  for (auto &&pt : kps1->points) {
    teaser::PointXYZ ptt;
    ptt.x = pt.x;
    ptt.y = pt.y;
    ptt.z = pt.z;
    cloud_src->push_back(ptt);
  }

  auto cloud_tgt = std::shared_ptr<teaser::PointCloud>(new teaser::PointCloud);
  for (auto &&pt : kps0->points) {
    teaser::PointXYZ ptt;
    ptt.x = pt.x;
    ptt.y = pt.y;
    ptt.z = pt.z;
    cloud_tgt->push_back(ptt);
  }
  // teaser::PointCloud

  teaser::RobustRegistrationSolver::Params params;
  params.noise_bound = 0.1;
  params.cbar2 = 1;
  params.estimate_scaling = false;
  params.rotation_max_iterations = 100;
  params.rotation_gnc_factor = 1.4;
  params.rotation_estimation_algorithm =
      teaser::RobustRegistrationSolver::ROTATION_ESTIMATION_ALGORITHM::GNC_TLS;
  params.rotation_cost_threshold = 0.005;

  // Solve with TEASER++
  teaser::RobustRegistrationSolver solver(params);
  std::chrono::steady_clock::time_point begin =
      std::chrono::steady_clock::now();

  solver.solve(*cloud_src, *cloud_tgt, corr);

  auto solution = solver.getSolution();
  INFOLOG("estimated: {}", solution.translation.transpose());

  // solution.rvoido
  return xreg::SE3(solution.rotation, solution.translation);
}

void extractDescriptors(pcl::PointCloud<pcl::PointXYZ>::ConstPtr input,
                        const pcl::PointCloud<pcl::PointXYZ>::Ptr &keypoints,
                        pcl::PointCloud<pcl::FPFHSignature33>::Ptr features) {
  typename pcl::PointCloud<pcl::PointXYZ>::Ptr kpts(
      new pcl::PointCloud<pcl::PointXYZ>);
  kpts->points.resize(keypoints->size());

  pcl::copyPointCloud(*keypoints, *kpts);

  auto est_fpfh =
      pcl::FPFHEstimation<pcl::PointXYZ, pcl::Normal, pcl::FPFHSignature33>::
          Ptr(new pcl::FPFHEstimation<pcl::PointXYZ, pcl::Normal,
                                      pcl::FPFHSignature33>);

  est_fpfh->setSearchSurface(input);
  est_fpfh->setInputCloud(kpts);

  // if (feature_from_normals)
  {
    // std::cout << "normal estimation..." << std::flush;
    INFOLOG("normal estimation");
    pcl::PointCloud<pcl::Normal>::Ptr normals(new pcl::PointCloud<pcl::Normal>);
    pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> normal_estimation;
    normal_estimation.setSearchMethod(pcl::search::Search<pcl::PointXYZ>::Ptr(
        new pcl::search::KdTree<pcl::PointXYZ>));

    normal_estimation.setRadiusSearch(downsample_resolution * 4);
    normal_estimation.setInputCloud(input);
    normal_estimation.compute(*normals);

    est_fpfh->setInputNormals(normals);
    est_fpfh->setSearchMethod(pcl::search::Search<pcl::PointXYZ>::Ptr(
        new pcl::search::KdTree<pcl::PointXYZ>));
    est_fpfh->setRadiusSearch(kps_resolution * 4);
    // std::cout << "OK" << std::endl;
    INFOLOG("done.");
  }

  // std::cout << "descriptor extraction..." << std::flush;

  INFOLOG("compute fpfh.");
  est_fpfh->compute(*features);
  // est_fpfh->
  // std::cout << "OK" << std::endl;
  INFOLOG("done.");
}

static void detectKeyPoints(xreg::PointCloudXYZ::Ptr cloud,
                            xreg::PointCloudXYZ::Ptr &kps,
                            pcl::PointIndices::Ptr &indices) {

  kps = xreg::PointCloudXYZ::Ptr(new pcl::PointCloud<pcl::PointXYZ>);

  pcl::UniformSampling<pcl::PointXYZ> sampler;
  sampler.setInputCloud(cloud);
  sampler.setRadiusSearch(kps_resolution);
  sampler.filter(*kps);

  // pcl::PointIndices::Ptr indices(new pcl::PointIndices);

  pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> est_normal;
  pcl::search::KdTree<pcl::PointXYZ>::Ptr tree(
      new pcl::search::KdTree<pcl::PointXYZ>());

  // get indices
  indices = pcl::PointIndices::Ptr(new pcl::PointIndices);
  {
    tree->setInputCloud(cloud);
    for (pcl::PointXYZ &pt : kps->points) {
      std::vector<int> inds;
      vector<float> dists;
      tree->nearestKSearch(pt, 1, inds, dists);
      indices->indices.push_back(inds[0]);
    }
  }

  INFOLOG("#kps {}", kps->size());
}

void filterCorrespondences(const std::vector<int> &s2t,
                           const std::vector<int> &t2s,
                           pcl::PointCloud<pcl::PointXYZ>::Ptr kps0,
                           pcl::PointCloud<pcl::PointXYZ>::Ptr kps1,
                           pcl::CorrespondencesPtr &corr) {
  // std::cout << "" << std::flush;
  INFOLOG("correspondence rejection...")
  std::vector<std::pair<unsigned, unsigned>> correspondences;
  for (std::size_t cIdx = 0; cIdx < s2t.size(); ++cIdx)
    if (t2s[s2t[cIdx]] == static_cast<int>(cIdx))
      correspondences.push_back(std::make_pair(cIdx, s2t[cIdx]));

  corr->resize(correspondences.size());
  for (std::size_t cIdx = 0; cIdx < correspondences.size(); ++cIdx) {
    (*corr)[cIdx].index_query = correspondences[cIdx].first;
    (*corr)[cIdx].index_match = correspondences[cIdx].second;
  }

  pcl::registration::CorrespondenceRejectorSampleConsensus<pcl::PointXYZ>
      rejector;
  rejector.setMaxIterations(1000);
  rejector.setMaximumIterations(1000);
  rejector.setInputSource(kps0);
  rejector.setInputTarget(kps1);
  rejector.setInputCorrespondences(corr);
  rejector.getCorrespondences(*corr);
  // std::cout << "OK" << std::endl;
  INFOLOG("done corr.");
}

int main(int argc, char **argv) {
  ros::init(argc, argv, "test0");

  if (argc != 4) {
    WARNLOG("input frame.");
    exit(0);
  }

  int idx0 = atoi(argv[2]);
  int idx1 = atoi(argv[3]);

  ros::NodeHandle nh;

  xreg::Visualizer visualizer(nh);
  auto thread_viz_ =
      unique_ptr<thread>(new std::thread(&xreg::Visualizer::spin, &visualizer));

  loadFrames("/home/hyhuang/Data/ethz_registration/" + std::string(argv[1]) +
             '/');

  pcl::VoxelGrid<pcl::PointXYZ> voxel_filer;
  voxel_filer.setLeafSize(downsample_resolution, downsample_resolution,
                          downsample_resolution);
  auto frame0 = frames[idx0];
  auto frame1 = frames[idx1];
  SE3 Tgt = frame0->Tgt_.inverse() * frame1->Tgt_;
  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud0(
      new pcl::PointCloud<pcl::PointXYZ>);
  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud1(
      new pcl::PointCloud<pcl::PointXYZ>);
  {
    voxel_filer.setInputCloud(frame0->cloud_);
    voxel_filer.filter(*cloud0);
  }
  {
    // pcl::io::loadPCDFile(pcd_file, *cloud);
    voxel_filer.setInputCloud(frame1->cloud_);
    voxel_filer.filter(*cloud1);
  }

  // pcl::PointCloud<pcl::PointXYZ>::Ptr cloud0(
  //     new pcl::PointCloud<pcl::PointXYZ>);
  // pcl::PointCloud<pcl::PointXYZ>::Ptr cloud1(
  //     new pcl::PointCloud<pcl::PointXYZ>);
  // {
  //   pcl::PointCloud<pcl::PointXYZ>::Ptr cloud(
  //       new pcl::PointCloud<pcl::PointXYZ>);
  //   std::string pcd_file = "/home/hyhuang/Downloads/stairs/pcd/24.pcd";
  //   pcl::io::loadPCDFile(pcd_file, *cloud);
  //   voxel_filer.setInputCloud(cloud);
  //   voxel_filer.filter(*cloud0);
  // }
  // {
  //   pcl::PointCloud<pcl::PointXYZ>::Ptr cloud(
  //       new pcl::PointCloud<pcl::PointXYZ>);
  //   std::string pcd_file = "/home/hyhuang/Downloads/stairs/pcd/25.pcd";
  //   pcl::io::loadPCDFile(pcd_file, *cloud);
  //   voxel_filer.setInputCloud(cloud);
  //   voxel_filer.filter(*cloud1);
  // }

  for (size_t i = 0; i < cloud1->size(); i++) {
    // cloud1->points[i].x += 20;
    // cloud1->points[i].y += 10;
  }

  visualizer.addCloud("raw0", "map", *cloud0);
  visualizer.addCloud("raw1", "map", *cloud1, Eigen::Vector3d(20, 0, 0));

  xreg::PointCloudXYZ::Ptr kps0;
  pcl::PointIndices::Ptr indices0;
  xreg::CloudFPFH::Ptr fpfh0(new xreg::CloudFPFH);
  xreg::PointCloudXYZ::Ptr kps1;
  pcl::PointIndices::Ptr indices1;
  xreg::CloudFPFH::Ptr fpfh1(new xreg::CloudFPFH);
  detectKeyPoints(cloud0, kps0, indices0);
  visualizer.addCloud("kps0", "map", *kps0);
  detectKeyPoints(cloud1, kps1, indices1);
  visualizer.addCloud("kps1", "map", *kps1, Eigen::Vector3d(20, 0, 0));

  extractDescriptors(cloud0, kps0, fpfh0);
  extractDescriptors(cloud1, kps1, fpfh1);

  std::vector<int> corr_src2tgt, corr_tgt2src;
  findCorrespondences(fpfh1, fpfh0, corr_src2tgt);
  findCorrespondences(fpfh0, fpfh1, corr_tgt2src);

  // correspondence.resize(fpfh0->size(), -1);
  // for (int i = 0; i < correspondence.size(); i++) {
  //   if (corr_tgt2src[corr_src2tgt[i]] == i) {
  //     correspondence[i] = corr_src2tgt[i];
  //   }
  // }
  // auto correspondence = pcl::CorrespondencesPtr(new pcl::Correspondences);
  std::vector<std::pair<int, int>> correspondences;
  for (std::size_t cIdx = 0; cIdx < corr_src2tgt.size(); ++cIdx)
    if (corr_tgt2src[corr_src2tgt[cIdx]] == static_cast<int>(cIdx))
      correspondences.push_back(std::make_pair(cIdx, corr_src2tgt[cIdx]));
  xreg::SE3 Test = registerByTeaser(kps0, kps1, correspondences);
  INFOLOG("est rotation: {}", Test.so3().log().transpose());
  INFOLOG("gt rotation: {}", Tgt.so3().log().transpose());
  INFOLOG("est translation: {}", Test.translation().transpose());
  INFOLOG("gt translation: {}", Tgt.translation().transpose());
  // filterCorrespondences(corr_src2tgt, corr_tgt2src, kps0, kps1,
  // correspondence);

  // INFOLOG("#pts: {}, #keypoints: {}", cloud->size(),
  // indices->indices.size());
  auto checkInlier = [](const pcl::PointXYZ &p0, const pcl::PointXYZ &p1,
                        const xreg::SE3 &T01,
                        const double dist = 0.05) -> bool {
    Eigen::Vector3d delta = T01 * Eigen::Vector3d(p1.x, p1.y, p1.z);
    delta = delta - Eigen::Vector3d(p0.x, p0.y, p0.z);
    // cout << delta.norm() << endl;
    // T01.
    if (delta.norm() < dist) {
      return true;
    } else {
      return false;
    }
  };

  auto pub_corr =
      nh.advertise<visualization_msgs::MarkerArray>("correspondences", 1, true);
  auto msg_corr =
      visualization_msgs::MarkerArray::Ptr(new visualization_msgs::MarkerArray);

  auto stamp = ros::Time::now();
  for (int i = 0; i < correspondences.size(); i++) {
    auto corr = correspondences.at(i);
    // if (idx < 0 || idx >= kps1->size()) {
    //   continue;
    // }

    /* code */
    auto marker = visualization_msgs::Marker();
    marker.header.frame_id = "map";
    marker.header.seq = 0;
    marker.header.stamp = stamp;

    marker.id = i++;
    marker.ns = "corr_line";
    marker.type = visualization_msgs::Marker::LINE_LIST;
    marker.action = visualization_msgs::Marker::ADD;

    marker.pose.position.x = 0.0;
    marker.pose.position.y = 0.0;
    marker.pose.position.z = 0.0;
    marker.pose.orientation.x = 0.0;
    marker.pose.orientation.y = 0.0;
    marker.pose.orientation.z = 0.0;
    marker.pose.orientation.w = 1.0;

    geometry_msgs::Point point0, point1, point2;
    std_msgs::ColorRGBA color_line;
    // color_line.a = 0.8;
    // color_line.r = color(1);
    // color_line.g = color(2);
    // color_line.b = color(3);
    pcl::PointXYZ &p0 = kps0->points[corr.second];
    pcl::PointXYZ &p1 = kps1->points[corr.first];
    point0.x = p0.x;
    point0.y = p0.y;
    point0.z = p0.z;

    point1.x = p1.x + 20;
    point1.y = p1.y;
    point1.z = p1.z;
    marker.points.push_back(point0);
    marker.points.push_back(point1);

    auto check_res = checkInlier(p0, p1, Test, 0.1);

    if (check_res) {
      marker.scale.x = 0.04;
      marker.color.a = 0.8;
      marker.color.r = 0;
      marker.color.g = 1.0;
      marker.color.b = 0;
    } else {
      marker.scale.x = 0.01;
      marker.color.a = 0.6;
      marker.color.r = 1.0f;
      marker.color.g = 0;
      marker.color.b = 0;
    }

    msg_corr->markers.push_back(marker);
  }

  pub_corr.publish(msg_corr);

  // visualization_msgs
  visualizer.addCloud("align0", "map", *cloud0, Eigen::Vector3d(0, 30, 0));

  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud1_transformed(
      new pcl::PointCloud<pcl::PointXYZ>);
  pcl::transformPointCloud(*cloud1, *cloud1_transformed,
                           Test.matrix().cast<float>());

  visualizer.addCloud("align1", "map", *cloud1_transformed,
                      Eigen::Vector3d(0, 30, 0));

  ros::spin();

  thread_viz_->join();

  //   cout <<
  // extractor->setKSearch();
  //   pcl::extractor->compute();
  // pcl::HarrisKeypoint3D()
}